Package: profExtrema
Type: Package
Title: Compute and Visualize Profile Extrema Functions
Version: 0.2.2
Authors@R: person("Dario","Azzimonti",role=c("aut","cre","cph"),email= "dario.azzimonti@gmail.com", comment = c(ORCID = "0000-0001-5080-3061"))
Maintainer: Dario Azzimonti <dario.azzimonti@gmail.com>
Date: 2026-07-21
Description: Computes profile extrema functions for arbitrary functions. If the function is expensive-to-evaluate it computes profile extrema by emulating the function with a Gaussian process (using package 'DiceKriging'). In this case uncertainty quantification on the profile extrema can also be computed. The different plotting functions for profile extrema give the user a tool to better locate excursion sets. 
License: GPL-3
Encoding: UTF-8
LazyData: true
Depends: R (>= 2.10), DiceKriging, KrigInv, pGPx
Imports: microbenchmark, quantreg, lhs, splines, methods, RColorBrewer,
        utils, MASS, rcdd
URL: https://github.com/dazzimonti/profExtrema
BugReports: https://github.com/dazzimonti/profExtrema/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-21 10:55:47 UTC; dario.azzimonti
Author: Dario Azzimonti [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-5080-3061>)
Repository: CRAN
Date/Publication: 2026-07-21 11:50:02 UTC

# Generated by roxygen2: do not edit by hand

export(approxMaxMin)
export(approxProfileExtrema)
export(bound_profiles)
export(cleanProfileResults)
export(coordProf_UQ)
export(coordinateProfiles)
export(getAllMaxMin)
export(getChangePoints)
export(getPointProportion)
export(getProfileExtrema)
export(getProfileInf_optim)
export(getProfileSup_optim)
export(getSegments)
export(gradKm_dnewdata)
export(obliqueProf_UQ)
export(obliqueProfiles)
export(plotBivariateProfiles)
export(plotMaxMin)
export(plotOblique)
export(plotOneBivProfile)
export(plot_univariate_profiles_UQ)
export(prof_mean_var_Delta)
export(setPlotOptions)
import(microbenchmark)
importFrom(DiceKriging,covMat1Mat2)
importFrom(DiceKriging,covMatrix)
importFrom(DiceKriging,covVector.dx)
importFrom(DiceKriging,km)
importFrom(DiceKriging,predict.km)
importFrom(DiceKriging,simulate)
importFrom(KrigInv,integration_design)
importFrom(MASS,Null)
importFrom(RColorBrewer,brewer.pal)
importFrom(grDevices,adjustcolor)
importFrom(grDevices,cairo_pdf)
importFrom(grDevices,dev.off)
importFrom(grDevices,gray.colors)
importFrom(grDevices,heat.colors)
importFrom(grDevices,pdf)
importFrom(grDevices,terrain.colors)
importFrom(graphics,abline)
importFrom(graphics,axis)
importFrom(graphics,contour)
importFrom(graphics,hist)
importFrom(graphics,image)
importFrom(graphics,layout)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(graphics,polygon)
importFrom(lhs,maximinLHS)
importFrom(methods,is)
importFrom(pGPx,grad_kweights)
importFrom(pGPx,krig_weight_GPsimu)
importFrom(pGPx,optim_dist_measure)
importFrom(quantreg,predict.rq)
importFrom(quantreg,rq)
importFrom(rcdd,d2q)
importFrom(rcdd,lpcdd)
importFrom(rcdd,makeH)
importFrom(rcdd,q2d)
importFrom(rcdd,scdd)
importFrom(splines,bs)
importFrom(stats,as.formula)
importFrom(stats,deriv)
importFrom(stats,model.matrix)
importFrom(stats,optim)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,smooth.spline)
importFrom(utils,combn)
importFrom(utils,modifyList)

<!-- README.md is generated from README.Rmd. Please edit that file -->
profExtrema
===========

Computation and plots of profile extrema functions. The package main functions are:

### Computation:

-   `coordinateProfiles`: Given a km objects computes the coordinate profile extrema function for the posterior mean and its quantiles.

-   `coordProf_UQ`: UQ part of coordinateProfiles.

-   `obliqueProfiles`: Given a km objects computes the profile extrema functions for a generic list of matrices Psi for the posterior mean and its quantiles.

-   `obliqueProf_UQ`: The UQ part of obliqueProfiles.

-   `getAllMaxMin`: computes coordinate profile extrema with full optimization for a deterministic function.

-   `approxMaxMin`: approximates coordinate profile extrema for a deterministic function.

-   `getProfileExtrema`: computes profile extrema given a list of matrices Psi for a deterministic function.

-   `approxProfileExtrema`: approximates profile extrema given a list of matrices Psi for a deterministic function.

### Plotting:

-   `plot_univariate_profiles_UQ`: plots for the results of coordProf\_UQ or obliqueProf\_UQ. Note that this function only works for univariate profiles.

-   `plotBivariateProfiles`: plots the bivariate maps results of a call to obliqueProfiles with a two dimensional projection matrix Psi.

-   `plotMaxMin`: simple plotting function for univariate profile extrema.

-   `plotOneBivProfile`: simple plotting function for bivariate profile extrema.

Note
----

This work was supported in part the Hasler Foundation, grant number 16065 and by the Swiss National Science Foundation, grant number 167199. The author warmly thanks David Ginsbourger, Jérémy Rohmer and Déborah Idier for fruitful discussions and accurate, thought provoking suggestions.

References
----------

Azzimonti, D., Bect, J., Chevalier, C., and Ginsbourger, D. (2016). Quantifying uncertainties on excursion sets under a Gaussian random field prior. SIAM/ASA Journal on Uncertainty Quantification, 4(1):850–874.

Azzimonti, D., Ginsbourger, D., Rohmer, J. and Idier, D. (2017+). Profile extrema for visualizing and quantifying uncertainties on excursion regions. Application to coastal flooding. arXiv:1710.00688.

Chevalier, C. (2013). Fast uncertainty reduction strategies relying on Gaussian process models. PhD thesis, University of Bern.

Chevalier, C., Picheny, V., Ginsbourger, D. (2014). An efficient and user-friendly implementation of batch-sequential inversion strategies based on kriging. Computational Statistics & Data Analysis, 71: 1021-1034.

Johnson, S. G. The NLopt nonlinear-optimization package, <http://ab-initio.mit.edu/nlopt>

Koenker, R. (2017). quantreg: Quantile Regression. R package version 5.33.

Nocedal, J. and Wright, S. J. (2006). Numerical Optimization, second edition. Springer- Verlag, New York.

Neuwirth, E. (2014). RColorBrewer: ColorBrewer Palettes. R package version 1.1-2.

Roustant, O., Ginsbourger, D., Deville, Y. (2012). DiceKriging, DiceOptim: Two R Packages for the Analysis of Computer Experiments by Kriging-Based Metamodeling and Optimization. Journal of Statistical Software, 51(1): 1-55.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# profExtrema

<!-- badges: start -->

[![R-CMD-check](https://github.com/dazzimonti/profExtrema/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/dazzimonti/profExtrema/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/profExtrema)](https://CRAN.R-project.org/package=profExtrema)
[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
<!-- badges: end -->

Computation and plots of profile extrema functions. The package main
functions are:

### Computation:

- `coordinateProfiles`: Given a km objects computes the coordinate
  profile extrema function for the posterior mean and its quantiles.

- `coordProf_UQ`: UQ part of coordinateProfiles.

- `obliqueProfiles`: Given a km objects computes the profile extrema
  functions for a generic list of matrices Psi for the posterior mean
  and its quantiles.

- `obliqueProf_UQ`: The UQ part of obliqueProfiles.

- `getAllMaxMin`: computes coordinate profile extrema with full
  optimization for a deterministic function.

- `approxMaxMin`: approximates coordinate profile extrema for a
  deterministic function.

- `getProfileExtrema`: computes profile extrema given a list of matrices
  Psi for a deterministic function.

- `approxProfileExtrema`: approximates profile extrema given a list of
  matrices Psi for a deterministic function.

### Plotting:

- `plot_univariate_profiles_UQ`: plots for the results of coordProf_UQ
  or obliqueProf_UQ. Note that this function only works for univariate
  profiles.

- `plotBivariateProfiles`: plots the bivariate maps results of a call to
  obliqueProfiles with a two dimensional projection matrix Psi.

- `plotMaxMin`: simple plotting function for univariate profile extrema.

- `plotOneBivProfile`: simple plotting function for bivariate profile
  extrema.

## Note

This work was supported in part the Hasler Foundation, grant number
16065 and by the Swiss National Science Foundation, grant number 167199.
The author warmly thanks David Ginsbourger, Jérémy Rohmer and Déborah
Idier for fruitful discussions and accurate, thought provoking
suggestions.

## References

Azzimonti, D., Bect, J., Chevalier, C., and Ginsbourger, D. (2016).
Quantifying uncertainties on excursion sets under a Gaussian random
field prior. SIAM/ASA Journal on Uncertainty Quantification,
4(1):850–874.

Azzimonti, D., Ginsbourger, D., Rohmer, J. and Idier, D. (2017+).
Profile extrema for visualizing and quantifying uncertainties on
excursion regions. Application to coastal flooding. arXiv:1710.00688.

Chevalier, C. (2013). Fast uncertainty reduction strategies relying on
Gaussian process models. PhD thesis, University of Bern.

Chevalier, C., Picheny, V., Ginsbourger, D. (2014). An efficient and
user-friendly implementation of batch-sequential inversion strategies
based on kriging. Computational Statistics & Data Analysis, 71:
1021-1034.

Johnson, S. G. The NLopt nonlinear-optimization package,
<https://nlopt.readthedocs.io/en/latest/>

Koenker, R. (2017). quantreg: Quantile Regression. R package version
5.33.

Nocedal, J. and Wright, S. J. (2006). Numerical Optimization, second
edition. Springer- Verlag, New York.

Neuwirth, E. (2014). RColorBrewer: ColorBrewer Palettes. R package
version 1.1-2.

Roustant, O., Ginsbourger, D., Deville, Y. (2012). DiceKriging,
DiceOptim: Two R Packages for the Analysis of Computer Experiments by
Kriging-Based Metamodeling and Optimization. Journal of Statistical
Software, 51(1): 1-55.

#' @author Dario Azzimonti
#' @name getProfileExtrema
#' @title Profile extrema with BFGS optimization
#' @description Evaluate profile extrema for a set of matrices allPsi with full optimization.
#' @param f the function to be evaluated
#' @param fprime derivative of the function
#' @param d dimension of the input domain
#' @param allPsi a list containing the matrices Psi (dim \eqn{pxd}) for which to compute the profile extrema
#' @param opts a list containing the options for this function and the subfunctions \link{getProfileSup_optim}, \link{getProfileInf_optim}. The options only for getProfileExtrema are
#' \itemize{
#' \item \code{limits:} an optional list containing \code{lower} and \code{upper}, two vectors with the limits of the input space. If NULL then \code{limits=list(upper=rep(1,d),lower=rep(0,d))}
#' \item \code{discretization:} an optional integer representing the discretization size for the profile computation for each dimension of eta. Pay attention that this leads to a grid of size \code{discretization^p}.
#' \item \code{heavyReturn:} If TRUE returns also all minimizers, default is FALSE.
#' \item \code{plts:} If TRUE and p==1 for all Psi in allPsi, plots the profile functions at each Psi, default is FALSE.
#' \item \code{verb:} If TRUE, outputs intermediate results, default is FALSE.
#' }
#' @return a list of two data frames (min, max) of the evaluations of \eqn{P^sup_Psi f(eta) = sup_{Psi x = \eta} f(x) } and \eqn{P^inf_Psi f(eta) = inf_{Psi x = \eta} f(x) }
#' discretized over 50 equally spaced points for each dimension for each Psi in \code{allPsi}. This number can be changed by defining it in options$discretization.
#' @examples
#' # Compute the oblique profile extrema with full optimization on 2d example
#'
#' # Define the function
#' testF <- function(x,params,v1=c(1,0),v2=c(0,1)){
#' return(sin(crossprod(v1,x)*params[1]+params[2])+cos(crossprod(v2,x)*params[3]+params[4])-1.5)
#' }
#'
#' testFprime <- function(x,params,v1=c(1,0),v2=c(0,1)){
#'   return(matrix(c(params[1]*v1[1]*cos(crossprod(v1,x)*params[1]+params[2])-
#'                   params[3]*v2[1]*sin(crossprod(v2,x)*params[3]+params[4]),
#'                  params[1]*v1[2]*cos(crossprod(v1,x)*params[1]+params[2])-
#'                   params[3]*v2[2]*sin(crossprod(v2,x)*params[3]+params[4])),ncol=1))
#' }
#'
#'
#' # Define the main directions of the function
#' theta=pi/6
#' pparams<-c(1,0,10,0)
#' vv1<-c(cos(theta),sin(theta))
#' vv2<-c(cos(theta+pi/2),sin(theta+pi/2))
#'
#' # Define optimizer friendly function
#' f <-function(x){
#' return(testF(x,pparams,vv1,vv2))
#' }
#' fprime <- function(x){
#'  return(testFprime(x,pparams,vv1,vv2))
#' }
#'
#' # Define list of directions where to evaluate the profile extrema
#' all_Psi <- list(Psi1=vv1,Psi2=vv2)
#'
#'
#' \donttest{
#' # Evaluate profile extrema along directions of all_Psi
#' allOblique<-getProfileExtrema(f=f,fprime = fprime,d = 2,allPsi = all_Psi,
#'                               opts = list(plts=FALSE,discretization=100,multistart=8))
#'
#'
#' # Consider threshold=0
#' threshold <- 0
#'
#' # Plot oblique profile extrema functions
#' plotMaxMin(allOblique,allOblique$Design,threshold = threshold)
#'
#' ## Since the example is two dimensional we can visualize the regions excluded by the profile extrema
#' # evaluate the function at a grid for plots
#' inDes<-seq(0,1,,100)
#' inputs<-expand.grid(inDes,inDes)
#' outs<-apply(X = inputs,MARGIN = 1,function(x){return(testF(x,pparams,v1=vv1,v2=vv2))})
#'
#' # obtain the points where the profiles take the threshold value
#' cccObl<-getChangePoints(threshold = threshold,allRes = allOblique,Design = allOblique$Design)
#'
#' # visualize the functions and the regions excluded
#'
#' image(inDes,inDes,matrix(outs,ncol=100),col=grey.colors(20),main="Example and oblique profiles")
#' contour(inDes,inDes,matrix(outs,ncol=100),add=TRUE,nlevels = 20)
#' contour(inDes,inDes,matrix(outs,ncol=100),add=TRUE,levels = c(threshold),col=4,lwd=1.5)
#' plotOblique(cccObl$alwaysEx$`0`[[1]],all_Psi[[1]],col=3)
#' plotOblique(cccObl$alwaysEx$`0`[[2]],all_Psi[[2]],col=3)
#' plotOblique(cccObl$neverEx$`0`[[1]],all_Psi[[1]],col=2)
#' plotOblique(cccObl$neverEx$`0`[[2]],all_Psi[[2]],col=2)
#'
#' }
#' @export
getProfileExtrema<-function(f,fprime=NULL,d,allPsi,opts=NULL){

  num_Psi<-length(allPsi)

  # initialize opts
  if(is.null(opts$limits)){
    ll_b = rep(0,d)
    uu_b = rep(1,d)
  }else{
    ll_b = opts$limits$lower
    uu_b = opts$limits$upper
  }

  if(is.null(opts$discretization)){
    dd_eta <- 50
  }else{
    dd_eta <- opts$discretization
  }

  if(is.null(opts$multistart)){
    opts$multistart=5
  }

  # Useful for choosing limits for eta
  cubeVertex<-matrix(c(ll_b[1],uu_b[1]),nrow=1)
  ii=1
  while(ii<d){
    ii=ii+1
    cubeVertex <- cbind(cubeVertex,cubeVertex)
    cubeVertex <-rbind(cubeVertex, c(rep(ll_b[ii],ncol(cubeVertex)/2),c(rep(uu_b[ii],ncol(cubeVertex)/2))))
  }

  p=nrow(matrix(allPsi[[1]],ncol=d))

  allMaxPoints<-list()
  allMinPoints<-list()
  results<-list(min=data.frame(matrix(NA,ncol=num_Psi,nrow = dd_eta^p)),max=data.frame(matrix(NA,ncol=num_Psi,nrow = dd_eta^p)))

  # Save the design, useful for plotting functions
  Design<- matrix(NA,ncol=num_Psi,nrow=dd_eta*p)

  # Loop over the different Psi
  for(i in seq(num_Psi)){

    if(!is.null(opts$verb))
      if(opts$verb)
        cat("Psi number ",i," of ",num_Psi,"\n")

    # take current Psi
    cPsi = allPsi[[i]]

    p = nrow(cPsi)
    if(is.null(p)){
      cPsi <-matrix(cPsi,ncol=d)
      p=nrow(cPsi)
    }
    # Choose limits for etas for current Psi
    mmEtas<-apply(crossprod(t(cPsi),cubeVertex),1,min)
    MMetas<-apply(crossprod(t(cPsi),cubeVertex),1,max)

    if(p==1){
      etas1<-matrix(seq(from = mmEtas,to = MMetas,length.out = dd_eta),ncol=1)
      Design[,i]<-etas1
    }else if(p==2){
      etas1<-expand.grid(seq(from = mmEtas[1],to = MMetas[1],length.out = dd_eta),seq(from = mmEtas[2],to = MMetas[2],length.out = dd_eta))
      Design[,i]<-c(seq(from = mmEtas[1],to = MMetas[1],length.out = dd_eta),seq(from = mmEtas[2],to = MMetas[2],length.out = dd_eta))
    }else{
      stop("ERROR: no implementation for p>2!")
    }

    pSup<-apply(etas1,1,function(x){a_res=getProfileSup_optim(eta = x,Psi = matrix(cPsi,ncol=d),f = f,fprime = fprime,d = d,options = opts);gc();return(a_res)})
    pInf<-apply(etas1,1,function(x){a_res=getProfileInf_optim(eta = x,Psi = matrix(cPsi,ncol=d),f = f,fprime = fprime,d = d,options = opts);gc();return(a_res)})

    results$max[[i]] <- sapply(pSup,function(x){x$val})
    allMaxPoints[[i]]<- sapply(pSup,function(x){x$aux$solution})

    results$min[[i]] <- sapply(pInf,function(x){x$val})
    allMinPoints[[i]]<- sapply(pSup,function(x){x$aux$solution})



    if(!is.null(opts$plts)){
      if(opts$plts){
        if(p==1){
          plot(etas1,results$min[,i],ylim=c(min(results$min[,i]),max(results$max[,i])),type='l',main=paste("Psi number",i))
          lines(etas1,results$max[,i])
        }else if(p==2){
          par(mfrow=c(1,2))
          image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(results$max[[i]],ncol=dd_eta),col=gray.colors(20),main=bquote(P[Psi[.(i)]]^sup ~ "f"))
          contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(results$max[[i]],ncol=dd_eta),nlevels = 10,add=T)

          image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(results$min[[i]],ncol=dd_eta),col=gray.colors(20),main=bquote(P[Psi[.(i)]]^inf ~ "f"))
          contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(results$min[[i]],ncol=dd_eta),nlevels = 10,add=T)
          par(mfrow=c(1,1))
        }else{
          stop("We should never be here!")
        }
      }
    }


  }

  toReturn<-list(res=results)
  toReturn$Design<-Design
  if(is.null(opts$heavyReturn))
    opts$heavyReturn=FALSE
  if(opts$heavyReturn){
    toReturn$minima<-allMinPoints
    toReturn$maxima<-allMaxPoints
  }
  return(toReturn)
}


# approxProfileExtrema
#' @author Dario Azzimonti
#' @name approxProfileExtrema
#' @title Approximate profile extrema functions
#' @description Evaluate profile extrema for a set of Psi with approximations at few values
#' @param f the function to be evaluated
#' @param fprime derivative of the function
#' @param d dimension of the input domain
#' @param allPsi a list containing the matrices Psi (dim \eqn{pxd}) for which to compute the profile extrema
#' @param opts a list containing the options for this function and the subfunctions \link{getProfileSup_optim}, \link{getProfileInf_optim} or \link{getProfileExtrema}. The options only for approxProfileExtrema are
#' \itemize{
#' \item \code{limits:} an optional list with the upper and lower limits of input space dimension, if NULL then \code{limits=list(upper=rep(1,d),lower=rep(0,d))}
#' \item \code{smoother:} Select which smoother to use:a string that selects which smoother to use: \itemize{
#'       \item \code{"1order"}: first order interpolation with gradient
#'       \item \code{"splineSmooth"}: smoothing spline with default degrees of freedom (DEFAULT OPTION)
#'       \item \code{"quantSpline"}: profile inf and profile sup approximated with quantile spline regression at levels 0.1 and 0.9 respectively
#' }
#' \item \code{heavyReturn:} If TRUE returns also all minimizers, default is FALSE.
#' \item \code{initDesign:} A list of the same length as allPsi containing the designs of few points where the expensive sup is evaluated. If Null it is automatically initialized
#' \item \code{fullDesignSize:} The full design where the function is approximated.
#' \item \code{multistart:} number of multistarts for optim procedure.
#' \item \code{numMCsamples:} number of MC samples for the sup.
#' \item \code{plts:} If TRUE, plots the max/min functions at each coordinate, default is FALSE.
#' \item \code{verb:} If TRUE, outputs intermediate results, default is FALSE.
#' }
#' @return a list of two data frames (min, max) of the evaluations of \eqn{f_sup(x_i) = sup_{x_j \neq i} f(x_1,\dots,x_d) } and \eqn{f_inf(x_i) = inf_{x_j \neq i} f(x_1,\dots,x_d) }
#' for each i at the design Design. By default Design is a 100 equally spaced points for each dimension. It can be changed by defining it in options$Design
#' @examples
#' # Compute the oblique profile extrema with approximate optimization on 2d example
#'
#' # Define the function
#' testF <- function(x,params,v1=c(1,0),v2=c(0,1)){
#' return(sin(crossprod(v1,x)*params[1]+params[2])+cos(crossprod(v2,x)*params[3]+params[4])-1.5)
#' }
#'
#' testFprime <- function(x,params,v1=c(1,0),v2=c(0,1)){
#'   return(matrix(c(params[1]*v1[1]*cos(crossprod(v1,x)*params[1]+params[2])-
#'                   params[3]*v2[1]*sin(crossprod(v2,x)*params[3]+params[4]),
#'                  params[1]*v1[2]*cos(crossprod(v1,x)*params[1]+params[2])-
#'                   params[3]*v2[2]*sin(crossprod(v2,x)*params[3]+params[4])),ncol=1))
#' }
#'
#'
#' # Define the main directions of the function
#' theta=pi/6
#' pparams<-c(1,0,10,0)
#' vv1<-c(cos(theta),sin(theta))
#' vv2<-c(cos(theta+pi/2),sin(theta+pi/2))
#'
#' # Define optimizer friendly function
#' f <-function(x){
#' return(testF(x,pparams,vv1,vv2))
#' }
#' fprime <- function(x){
#'  return(testFprime(x,pparams,vv1,vv2))
#' }
#'
#' # Define list of directions where to evaluate the profile extrema
#' all_Psi <- list(Psi1=vv1,Psi2=vv2)
#'
#' # Evaluate profile extrema along directions of all_Psi
#' allOblique<-approxProfileExtrema(f=f,fprime = fprime,d = 2,allPsi = all_Psi,
#'                                  opts = list(plts=FALSE,heavyReturn=TRUE))
#'
#' \donttest{
#' # Consider threshold=0
#' threshold <- 0
#'
#' # Plot oblique profile extrema functions
#' plotMaxMin(allOblique,allOblique$Design,threshold = threshold)
#'
#' ## Since the example is two dimensional we can visualize the regions excluded by the profile extrema
#' # evaluate the function at a grid for plots
#' inDes<-seq(0,1,,100)
#' inputs<-expand.grid(inDes,inDes)
#' outs<-apply(X = inputs,MARGIN = 1,function(x){return(testF(x,pparams,v1=vv1,v2=vv2))})
#'
#' # obtain the points where the profiles take the threshold value
#' cccObl<-getChangePoints(threshold = threshold,allRes = allOblique,Design = allOblique$Design)
#'
#' # visualize the functions and the regions excluded
#'
#' image(inDes,inDes,matrix(outs,ncol=100),col=grey.colors(20),main="Example and oblique profiles")
#' contour(inDes,inDes,matrix(outs,ncol=100),add=TRUE,nlevels = 20)
#' contour(inDes,inDes,matrix(outs,ncol=100),add=TRUE,levels = c(threshold),col=4,lwd=1.5)
#' plotOblique(cccObl$alwaysEx$`0`[[1]],all_Psi[[1]],col=3)
#' plotOblique(cccObl$alwaysEx$`0`[[2]],all_Psi[[2]],col=3)
#' plotOblique(cccObl$neverEx$`0`[[1]],all_Psi[[1]],col=2)
#' plotOblique(cccObl$neverEx$`0`[[2]],all_Psi[[2]],col=2)
#'
#' }
#' @export
approxProfileExtrema=function(f,fprime=NULL,d,allPsi,opts=NULL){

  # Set up options
  if(is.null(opts$limits)){
    limits<-list(lower=rep(0,d),upper=rep(1,d))
  }

  num_Psi<-length(allPsi)

  # initialize opts
  if(is.null(opts$limits)){
    ll_b = rep(0,d)
    uu_b = rep(1,d)
  }else{
    ll_b = limits$lower
    uu_b = limits$upper
  }

  if(is.null(opts$fullDesignSize)){
    opts$fullDesignSize<-80*d
  }

  if(is.null(opts$multistart)){
    opts$multistart=5
  }
  if(is.null(opts$initDesign))
    opts$initDesign<-list()

  # Useful for choosing limits for eta
  cubeVertex<-matrix(c(ll_b[1],uu_b[1]),nrow=1)
  ii=1
  while(ii<d){
    ii=ii+1
    cubeVertex <- cbind(cubeVertex,cubeVertex)
    cubeVertex <-rbind(cubeVertex, c(rep(ll_b[ii],ncol(cubeVertex)/2),c(rep(uu_b[ii],ncol(cubeVertex)/2))))
  }

  p=nrow(matrix(allPsi[[1]],ncol=d))

  exact_allMaxPoints<-list()
  exact_allMinPoints<-list()
  exact_results<-list(min=data.frame(matrix(NA,ncol=num_Psi,nrow = ceiling(sqrt(d*p)*10))),max=data.frame(matrix(NA,ncol=num_Psi,nrow = ceiling(sqrt(d*p)*10))))

  # initialize the max/min values
  sK_max<-data.frame(matrix(NA,nrow=opts$fullDesignSize^p,ncol=num_Psi))
  sK_min<-data.frame(matrix(NA,nrow=opts$fullDesignSize^p,ncol=num_Psi))

  # Save Design to return
  Design<- matrix(NA,ncol=num_Psi,nrow=opts$fullDesignSize*p)

  # Loop over the different Psi
  for(i in seq(num_Psi)){

    if(!is.null(opts$verb))
      if(opts$verb)
        cat("Psi number ",i," of ",num_Psi,"\n")

    # take current Psi
    cPsi = allPsi[[i]]

    p = nrow(cPsi)
    if(is.null(p)){
      cPsi <-matrix(cPsi,ncol=d)
      p=nrow(cPsi)
    }
    # Choose limits for etas for current Psi
    mmEtas<-apply(crossprod(t(cPsi),cubeVertex),1,min)
    MMetas<-apply(crossprod(t(cPsi),cubeVertex),1,max)


    # Get initial design
    if(length(opts$initDesign)<num_Psi){
      if(p==1){
        opts$initDesign[[i]]<-matrix(seq(from=mmEtas[1],to=MMetas[1],,ceiling(sqrt(d)*10)),ncol=1)
      }else{
        ### NEEDS TEST!
        opts$initDesign[[i]]<-matrix(mmEtas,ncol=p,byrow = T,nrow=ceiling(sqrt(d*p)*10))+maximinLHS(ceiling(sqrt(d*p)*10),p)%*%diag(MMetas-mmEtas,ncol=p)
      }

    }


    pSup<-apply(opts$initDesign[[i]],1,function(x){return(getProfileSup_optim(eta = x,Psi = matrix(cPsi,ncol=d),f = f,fprime = fprime,d = d,options = opts))})
    gc()
    pInf<-apply(opts$initDesign[[i]],1,function(x){return(getProfileInf_optim(eta = x,Psi = matrix(cPsi,ncol=d),f = f,fprime = fprime,d = d,options = opts))})
    gc()

    exact_results$max[[i]] <- sapply(pSup,function(x){x$val})
    exact_allMaxPoints[[i]]<- sapply(pSup,function(x){x$aux$solution})
    exact_results$min[[i]] <- sapply(pInf,function(x){x$val})
    exact_allMinPoints[[i]]<- sapply(pInf,function(x){x$aux$solution})

    # the gradient of the sup/inf function is saved
    if(!is.null(fprime) & p==1){
      nn<-length(exact_results$max[[i]])
      exact_grad_min<-matrix(NA,ncol=1,nrow=nn)
      exact_grad_max<-matrix(NA,ncol=1,nrow=nn)
      for(ll in seq(nn)){

        yy<-exact_allMinPoints[[i]][,ll]
        exact_grad_min[ll,]<-fprime(yy)[i,] ### THIS IS WRONG!
        yy<-exact_allMaxPoints[[i]][,ll]
        exact_grad_max[ll,]<-fprime(yy)[i,] ### THIS IS WRONG!
      }
    }
    if(p==1){
      newPoints<-matrix(seq(mmEtas,MMetas,,opts$fullDesignSize),ncol=1)
      Design[,i]<-newPoints
    }else{
      newPoints<-expand.grid(seq(mmEtas[1],MMetas[1],,opts$fullDesignSize),seq(mmEtas[2],MMetas[2],,opts$fullDesignSize))
      Design[,i]<-c(seq(mmEtas[1],MMetas[1],,opts$fullDesignSize),seq(mmEtas[2],MMetas[2],,opts$fullDesignSize))
    }


    if(p==1){

      if(!is.null(opts$smoother) && opts$smoother=="1order"){
        #    sK_max[,i]<- kGradSmooth(newPoints=newPoints,profPoints = exact_allMaxPoints[[i]],
        #                                 profEvals=exact_results$max[[i]], profGradient=exact_grad_max)

        #      sK_min[,i]<-kGradSmooth(newPoints=newPoints,profPoints = exact_allMaxPoints[[i]],
        #                                 profEvals=exact_results$min[[i]], profGradient=exact_grad_min)
        warning(paste("The method ",opts$smoother," is not implemented yet!"))
      }else if(!is.null(opts$smoother) && opts$smoother=="quantSpline"){
        dd_fit<-data.frame(x=opts$initDesign[[i]],y=exact_results$max[[i]])
        fit.max <- rq(y ~ bs(x, df=max(nn%/%1.5,3)),data=dd_fit,tau=0.9)
        dd_fit<-data.frame(x=opts$initDesign[[i]],y=exact_results$min[[i]])
        fit.min <- rq(y ~ bs(x, df=max(nn%/%1.5,3)), tau=0.1,data=dd_fit)
        newPoints<-data.frame(x=newPoints)

        sK_max[,i]<- predict.rq(object = fit.max,newdata = newPoints)
        sK_min[,i]<- predict.rq(object = fit.min,newdata = newPoints)
      }else {
        sk_maxSmooth<-smooth.spline(x=opts$initDesign[[i]],y=exact_results$max[[i]])
        sk_minSmooth<-smooth.spline(x=opts$initDesign[[i]],y=exact_results$min[[i]])

        sK_max[,i]<-predict(sk_maxSmooth,newPoints)$y
        sK_min[,i]<-predict(sk_minSmooth,newPoints)$y
      }

    }else{
      approxKm<-km(design = opts$initDesign[[i]],response = exact_results$max[[i]],covtype = "matern3_2",control = list(trace=F))
      sK_max[,i]<-predict.km(object = approxKm,newdata = newPoints,type = "UK",checkNames = F)$mean

      approxKm<-km(design = opts$initDesign[[i]],response = exact_results$min[[i]],covtype = "matern3_2",control = list(trace=F))
      sK_min[,i]<-predict.km(object = approxKm,newdata = newPoints,type = "UK",checkNames = F)$mean
    }


    ## Debug plots
    if(!is.null(opts$debug) && opts$debug){

      warning("The debug plots do not work!")

      # par(mfrow=c(1,2))
      # image(x=Design[1:opts$fullDesignSize,i],y=Design[(opts$fullDesignSize+1):(opts$fullDesignSize*2),i],
      #       z=matrix(sK_max[[i]],ncol=opts$fullDesignSize),col=gray.colors(20),main=sprintf("Psup Psi %i",i))
      # contour(x=Design[1:opts$fullDesignSize,i],y=Design[(opts$fullDesignSize+1):(opts$fullDesignSize*2),i],
      #       z=matrix(sK_max[[i]],ncol=opts$fullDesignSize),nlevels = 10,add=T)
      # contour(x=Design[1:opts$fullDesignSize,i],y=Design[(opts$fullDesignSize+1):(opts$fullDesignSize*2),i],
      #         z=matrix(sK_max[[i]],ncol=opts$fullDesignSize),levels = threshold,add=T,col=2)
      #
      # image(x=Design[1:opts$fullDesignSize,i],y=Design[(opts$fullDesignSize+1):(opts$fullDesignSize*2),i],
      #       z=matrix(sK_min[[i]],ncol=opts$fullDesignSize),col=gray.colors(20),main=sprintf("Pinf Psi %i",i))
      # contour(x=Design[1:opts$fullDesignSize,i],y=Design[(opts$fullDesignSize+1):(opts$fullDesignSize*2),i],
      #         z=matrix(sK_min[[i]],ncol=opts$fullDesignSize),nlevels = 10,add=T)
      # contour(x=Design[1:opts$fullDesignSize,i],y=Design[(opts$fullDesignSize+1):(opts$fullDesignSize*2),i],
      #         z=matrix(sK_min[[i]],ncol=opts$fullDesignSize),levels = threshold,add=T,col=2)
      # par(mfrow=c(1,1))
      #    par(mfrow=c(1,2))
      # Plot  max/min function
      # ylimTemp<-range(c(sK_min[,coord],aa$res$min[,coord],sK_max[,coord],aa$res$max[,coord]))
      # plot(aa$minima[((coord-1)*nn+1):(coord*nn),coord],aa$res$min[,coord],main=paste("Coordinate",coord,", Min"),
      #      xlab="x",ylab="f",ylim=ylimTemp)
      # lines(newPoints,sK_min[,coord],type='l',col=3)
      # points(aa$maxima[((coord-1)*nn+1):(coord*nn),coord],aa$res$max[,coord])
      # lines(newPoints,sK_max[,coord],type='l',col=3)
      #
      # if(!is.null(opts$smooth) && opts$smooth){
      #   points(sk_minSmooth$x,sk_minSmooth$y,col=2)
      #   points(sk_maxSmooth$x,sk_maxSmooth$y,col=2)
      # }
    }
  }

  res=list(min=sK_min,max=sK_max)
  toReturn<-list(res=res)
  toReturn$Design <- Design
  if(!is.null(opts$heavyReturn) && opts$heavyReturn){
    aa<-list(e_res=exact_results,e_maxPts= exact_allMaxPoints,e_minPts=exact_allMinPoints)
    aa$design<-opts$initDesign
    toReturn$profPoints<-aa
  }
  return(toReturn)



}

#' @author Dario Azzimonti
#' @name gradKm_dnewdata
#' @title Gradient of posterior mean and variance
#' @description Computes the gradient of the posterior mean and variance of the kriging model in \code{object} at the points \code{newdata}.
#' @param object a \link[DiceKriging]{km} object
#' @param newdata a vector, matrix or data frame containing the points where to perform predictions.
#' @param type a character corresponding to the type of kriging family (\code{"SK"} or \code{"UK"}).
#' @param se.compute an optional boolean indicating whether to compute the posterior variance or not. Default is TRUE.
#' @param bias.correct an optional boolean to correct bias in the UK variance. Default is FALSE.
#' @param light.return an optional boolean indicating whether to return additional variables. Default is FALSE.
#' @return Returns a list containing \itemize{
#' \item \code{mean:} the gradient of the posterior mean at \code{newdata}.
#' \item \code{trend:} the gradient of the trend at \code{newdata}.
#' \item \code{s2:} the gradient of the posterior variance at \code{newdata}.
#' }
#' @export
gradKm_dnewdata<-function (object, newdata, type, se.compute = TRUE, light.return = FALSE, bias.correct = FALSE)
{
  nugget.flag <- object@covariance@nugget.flag
  X <- object@X
  y <- object@y
  newdata <- as.matrix(newdata)
  d.newdata <- ncol(newdata)
  if (!identical(d.newdata, object@d)) {
    stop("newdata must have the same numbers of columns than the experimental design")
  }
  if (!identical(colnames(newdata), colnames(X))) {
    colnames(newdata) <- colnames(X)
  }
  T <- object@T
  z <- object@z
  M <- object@M
  beta <- object@trend.coef

  F.newdata<-model.matrix(object@trend.formula, data = data.frame(newdata))

  # Pre process formula to remove things that are not in derivative tables
  newFormula<-as.formula(gsub(")","",gsub("I(","",object@trend.formula,fixed=TRUE),fixed=TRUE))

  der.formula<-deriv(newFormula,namevec = colnames(object@X),function.arg = TRUE)
  trendDeriv<-function(x){
    result<-matrix(NA,ncol=ncol(x),nrow=nrow(x))
    for(i in seq(nrow(x)))
      result[i,]<-attr(do.call(der.formula,as.list(x[i,])),"gradient")
    return(cbind(rep(0,nrow(x)),result))
  }
  F.newdata.dx <- trendDeriv(newdata) #model.matrix(trendDeriv, data = data.frame(newdata))
  y.predict.trend<-matrix(NA,nrow=nrow(newdata),ncol=ncol(newdata))
  y.predict.complement<-matrix(NA,nrow=nrow(newdata),ncol=ncol(newdata))

  if(length(beta)>1){
    y.predict.trend <-sweep(F.newdata.dx,MARGIN = 2,beta,"*")[,-1]
  }else{
    y.predict.trend <- F.newdata.dx[,-1]
  }

  Tinv.c.newdata<-array(NA,dim = c(nrow(X),ncol(X),nrow(newdata)))
  c.deriv.newdata<-array(NA,dim = c(nrow(X),ncol(X),nrow(newdata)))
  c.newdata <- covMat1Mat2(object@covariance, X1 = X, X2 = newdata,
                           nugget.flag = object@covariance@nugget.flag)

  for(i in seq(nrow(newdata))){
    c.deriv.newdata[,,i]<-covVector.dx(object = object@covariance,x = newdata[i,],X = X,c = c.newdata[,i])
    Tinv.c.newdata[,,i] <- backsolve(t(T), c.deriv.newdata[,,i], upper.tri = FALSE)
    y.predict.complement[i,] <- crossprod(Tinv.c.newdata[,,i], z)
  }

  y.predict <- y.predict.trend + y.predict.complement
  output.list <- list()
  output.list$trend <- y.predict.trend
  output.list$mean <- y.predict

  if(se.compute){
    if (!is(object@covariance, "covUser")) {
      total.sd2 <- 0
    }
    else {
      stop("User covariance kernel derivatives not implemented")
    }

    s2.predict.1<-matrix(NA,nrow=nrow(newdata),ncol=ncol(newdata))
    for(i in seq(nrow(newdata))){
      s2.predict.1[i,]<-crossprod(Tinv.c.newdata[,,i],backsolve(t(T),c.newdata[,i],upper.tri = FALSE))
    }
    if (type == "SK") {
      s2.predict <- total.sd2 - 2*s2.predict.1
    }else if (type == "UK") {
      T.M <- chol(t(M) %*% M)
      s2.predict.mat <- matrix(NA,nrow=nrow(newdata),ncol=ncol(newdata))
      for(i in seq(nrow(newdata))){
        s2.predict.mat[i,]<-crossprod(backsolve(t(T.M), t(F.newdata[i,] -  backsolve(t(T),c.newdata[,i])%*% M), upper.tri = FALSE),
                                      backsolve(t(T.M), t(F.newdata.dx[i,-1] - crossprod(Tinv.c.newdata[,,i], M)), upper.tri = FALSE))
      }
      s2.predict <- total.sd2 - 2*s2.predict.1 + 2*s2.predict.mat
      if (bias.correct)
        s2.predict <- s2.predict * object@n/(object@n - object@p)
    }
#    lower95 <- y.predict - q95 * sqrt(s2.predict)
#    upper95 <- y.predict + q95 * sqrt(s2.predict)
    output.list$s2 <- s2.predict
#    output.list$lower95 <- lower95
#    output.list$upper95 <- upper95
  }
#  if (!light.return) {
#    output.list$c <- c.newdata
#    output.list$Tinv.c <- Tinv.c.newdata
# }
#  if ((se.compute) || (cov.compute)) {
#    if (!is(object@covariance, "covUser")) {
#      total.sd2 <- object@covariance@sd2
#    }
#    else {
#      m <- nrow(newdata)
#      total.sd2 <- rep(NA, m)
#      for (i in 1:m) {
#        total.sd2[i] <- object@covariance@kernel(newdata[i,
#                                                        ], newdata[i, ])
#     }
#   }
#    if (object@covariance@nugget.flag) {
#      total.sd2 <- total.sd2 + object@covariance@nugget
#    }
#  }
#  if (se.compute) {
#    s2.predict.1 <- apply(Tinv.c.newdata, 2, crossprod)
#    if (type == "SK") {
#      s2.predict <- pmax(total.sd2 - s2.predict.1, 0)
#      s2.predict <- as.numeric(s2.predict)
#      q95 <- qnorm(0.975)
#    }
#    else if (type == "UK") {
#      T.M <- chol(t(M) %*% M)
#      s2.predict.mat <- backsolve(t(T.M), t(F.newdata.dx -
#                                              t(Tinv.c.newdata) %*% M), upper.tri = FALSE)
#      s2.predict.2 <- apply(s2.predict.mat, 2, crossprod)
#      s2.predict <- pmax(total.sd2 - s2.predict.1 + s2.predict.2,
#                         0)
#      s2.predict <- as.numeric(s2.predict)
#      if (bias.correct)
#        s2.predict <- s2.predict * object@n/(object@n -
#                                                object@p)
#       q95 <- qt(0.975, object@n - object@p)
#     }
#     lower95 <- y.predict - q95 * sqrt(s2.predict)
#     upper95 <- y.predict + q95 * sqrt(s2.predict)
#     output.list$sd <- sqrt(s2.predict)
#     output.list$lower95 <- lower95
#     output.list$upper95 <- upper95
#   }
#   if (cov.compute) {
#     C.newdata <- covMatrix(object@covariance, newdata)[[1]]
#     cond.cov <- C.newdata - crossprod(Tinv.c.newdata)
#     if (type == "UK") {
#       T.M <- chol(t(M) %*% M)
#       s2.predict.mat <- backsolve(t(T.M), t(F.newdata.dx -
#                                               t(Tinv.c.newdata) %*% M), upper.tri = FALSE)
#       cond.cov <- cond.cov + crossprod(s2.predict.mat)
#       if (bias.correct)
#         cond.cov <- cond.cov * object@n/(object@n - object@p)
#     }
#     output.list$cov <- cond.cov
#   }
  return(output.list)
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
## usethis namespace: end

#' @title profExtrema package
#' @description Computation and plots of profile extrema functions. The package main functions are: \describe{
#'    \item{\strong{Computation:}}{ \itemize{
#'     \item \code{\link{coordinateProfiles}}: Given a \link[DiceKriging]{km} objects computes the coordinate profile extrema function for the posterior mean and its quantiles.
#'     \item \code{\link{coordProf_UQ}}: UQ part of \code{coordinateProfiles}.
#'     \item \code{\link{obliqueProfiles}}: Given a \link[DiceKriging]{km} objects computes the profile extrema functions for a generic list of matrices Psi for the posterior mean and its quantiles.
#'     \item \code{\link{obliqueProf_UQ}}: The UQ part of \code{obliqueProfiles}.
#'     \item \code{\link{getAllMaxMin}}: computes coordinate profile extrema with full optimization for a deterministic function.
#'     \item \code{\link{approxMaxMin}}: approximates coordinate profile extrema for a deterministic function.
#'     \item \code{\link{getProfileExtrema}}: computes profile extrema given a list of matrices Psi for a deterministic function.
#'     \item \code{\link{approxProfileExtrema}}: approximates profile extrema given a list of matrices Psi for a deterministic function.
#'    } }
#'    \item{\strong{Plotting:}}{ \itemize{
#'     \item \code{\link{plot_univariate_profiles_UQ}}: plots for the results of \code{coordProf_UQ} or \code{obliqueProf_UQ}. Note that this function only works for univariate profiles.
#'     \item \code{\link{plotBivariateProfiles}}: plots the bivariate maps results of a call to \code{obliqueProfiles} with a two dimensional projection matrix Psi.
#'     \item \code{\link{plotMaxMin}}: simple plotting function for univariate profile extrema.
#'     \item \code{\link{plotOneBivProfile}}: simple plotting function for bivariate profile extrema.
#'    }  }
#'
#' }
#' @details Package: profExtrema \cr
#' Type: Package \cr
#' Version: 0.2.1 \cr
#' Date: 2020-03-20
#'
#' @author Dario Azzimonti (dario.azzimonti@@gmail.com) .
#' @name profExtrema
#' @import microbenchmark
#' @importFrom DiceKriging predict.km covMat1Mat2 covVector.dx covMatrix simulate km
#' @importFrom KrigInv integration_design
#' @importFrom pGPx krig_weight_GPsimu optim_dist_measure grad_kweights
#' @importFrom quantreg rq predict.rq
#' @importFrom grDevices cairo_pdf dev.off gray.colors adjustcolor pdf
#' @importFrom graphics abline contour image legend lines par plot points hist axis
#' @importFrom stats deriv model.matrix optim predict rnorm quantile runif smooth.spline
#' @importFrom utils combn
#' @importFrom utils modifyList
#' @importFrom MASS Null
#' @importFrom rcdd makeH scdd lpcdd d2q q2d
#' @importFrom lhs maximinLHS
#' @importFrom splines bs
#' @importFrom graphics polygon layout
#' @importFrom methods is
#' @importFrom grDevices heat.colors terrain.colors
#' @importFrom stats as.formula
#' @importFrom RColorBrewer brewer.pal
#' @note This work was supported in part the Hasler Foundation, grant number 16065 and by the Swiss National Science Foundation, grant number 167199. The author warmly thanks David Ginsbourger, Jérémy Rohmer and Déborah Idier for fruitful discussions and accurate, thought provoking suggestions.
#' @references
# @importFrom lpSolve lp
#'
#' Azzimonti, D., Bect, J., Chevalier, C., and Ginsbourger, D. (2016). Quantifying uncertainties on excursion sets under a Gaussian random field prior. SIAM/ASA Journal on Uncertainty Quantification, 4(1):850-874.
#'
#' Azzimonti, D., Ginsbourger, D., Rohmer, J. and Idier, D. (2017+). Profile extrema for visualizing and quantifying uncertainties on excursion regions. Application to coastal flooding. arXiv:1710.00688.
#'
#' Chevalier, C. (2013). Fast uncertainty reduction strategies relying on Gaussian process models. PhD thesis, University of Bern.
#'
#' Chevalier, C., Picheny, V., Ginsbourger, D. (2014). An efficient and user-friendly implementation of batch-sequential inversion strategies based on kriging. Computational Statistics & Data Analysis, 71: 1021-1034.
#'
#' Johnson, S. G.  The NLopt nonlinear-optimization package, http://ab-initio.mit.edu/nlopt
#'
#' Koenker, R. (2017). quantreg: Quantile Regression. R package version 5.33.
#'
#' Nocedal, J. and Wright, S. J. (2006). Numerical Optimization, second edition. Springer- Verlag, New York.
#'
#' Neuwirth, E. (2014). RColorBrewer: ColorBrewer Palettes. R package version 1.1-2.
#'
#' Roustant, O., Ginsbourger, D., Deville, Y. (2012). DiceKriging, DiceOptim: Two R Packages for the Analysis of Computer Experiments by Kriging-Based Metamodeling and Optimization. Journal of Statistical Software, 51(1): 1-55.
#'
NULL

# getMax function
#' @author Dario Azzimonti
#' @name getMax
#' @title Coordinate profile sup function
#' @description Compute coordinate profile sup functions
#' @param x one dimensional point where the function is to be evaluated
#' @param f function to be optimized (takes a vector y of dimension d and returns a real number)
#' @param fprime derivative of f (same format)
#' @param coord integer selecting the dimension that is fixed, the other ones are optimized over
#' @param d dimension of the input for f
#' @param options a list containing the options to be passed to optim:
#' \itemize{
#' \item \code{par:} contains the starting point (a point in dimension d-1)
#' \item \code{method:} is the string denoting the chosen method for the optimization (see optim for details)
#' \item \code{lower:} the lower bounds for the optimization domain (see optim for details)
#' \item \code{upper:} the upper bounds for the optimization domain (see optim for details)
#' }
#' @return a real value corresponding to \eqn{max_{x_1,\dots, x_{coord-1},x_{coord+1}, \dots, x_d} f(x_1,\dots,x_d)}
getMax = function(x,f,fprime,coord,d,options=NULL){
  fxFixed=function(y){
    res<-NA
    if(coord==1)
      res<-f(c(x,y))
    if(coord==d)
      res<-f(c(y,x))
    if(coord>1 && coord<d)
      res<-f(c(y[1:(coord-1)],x,y[coord:(d-1)]))
    return(res)
  }
  if(!is.null(fprime)){
    fprimexFixed=function(y){
      res<-NA
      if(coord==1)
        res<-fprime(c(x,y))
      if(coord==d)
        res<-fprime(c(y,x))
      if(coord>1 && coord<d)
        res<-fprime(c(y[1:(coord-1)],x,y[coord:(d-1)]))
      return(res[-coord])
    }
  }else{
    fprimexFixed<-NULL
  }

  startingPoint<-runif(d-1)
  if(!is.null(options$par)){
    startingPoint<-options$par
  }

#  if(d<=2){
#    methd="Brent"
#    lower=0
#    upper=1
#  }else{
    methd="L-BFGS-B"
    lower=rep(0,d-1)
    upper=rep(1,d-1)
#  }
  trace=0

  if(!is.null(options$method))
    methd=options$method
  if(!is.null(options$lower))
    lower=options$lower
  if(!is.null(options$upper))
    upper=options$upper
  if(!is.null(options$trace))
    trace=options$trace


  optRes<-list(value=-Inf,aux=NA)
  if(!is.null(options$multistart)){
    optRess<-NULL
    for(i in seq(options$multistart)){
      startingPoint<-runif(d-1,min = lower,max=upper)
      if(!is.null(options$par) && i==1)
        startingPoint<-options$par
      optRess<-optim(par = startingPoint,fn = fxFixed,gr = fprimexFixed,control = list(fnscale=-1,trace=trace,pgtol=1e-4),method = methd,lower = lower,upper = upper)
      if(optRess$value>=optRes$value){
        optRes=optRess
      }else{
        break
      }
    }
  }else{
    optRes<-optim(par = startingPoint,fn = fxFixed,gr = fprimexFixed,control = list(fnscale=-1,trace=trace,pgtol=1e-4),method = methd,lower = lower,upper = upper)
  }



  return(list(val=optRes$value,aux=optRes))

}

# getMin function
#' @author Dario Azzimonti
#' @name getMin
#' @title Coordinate profile inf function
#' @description Compute coordinate profile inf functions
#' @param x one dimensional point where the function is to be evaluated
#' @param f function to be optimized (takes a vector y of dimension d and returns a real number)
#' @param fprime derivative of f (same format)
#' @param coord integer selecting the dimension that is fixed, the other ones are optimized over
#' @param d dimension of the input for f
#' @param options a list containing the options to be passed to optim:
#' \itemize{
#' \item \code{par:} contains the starting point (a point in dimension d-1)
#' \item \code{method:} is the string denoting the chosen method for the optimization (see optim for details)
#' \item \code{lower:} the lower bounds for the optimization domain (see optim for details)
#' \item \code{upper:} the upper bounds for the optimization domain (see optim for details)
#' }
#' @return a real value corresponding to \eqn{min_{x_1,\dots, x_{coord-1},x_{coord+1}, \dots, x_d} f(x_1,\dots,x_d)}
getMin = function(x,f,fprime,coord,d,options=NULL){
  fxFixed=function(y){
    res<-NA
    if(coord==1)
      res<-f(c(x,y))
    if(coord==d)
      res<-f(c(y,x))
    if(coord>1 && coord<d)
      res<-f(c(y[1:(coord-1)],x,y[coord:(d-1)]))
    return(res)
  }
  if(!is.null(fprime)){
    fprimexFixed=function(y){
      res<-NA
      if(coord==1)
        res<-fprime(c(x,y))
      if(coord==d)
        res<-fprime(c(y,x))
      if(coord>1 && coord<d)
        res<-fprime(c(y[1:(coord-1)],x,y[coord:(d-1)]))
      return(res[-coord])
    }
  }else{
    fprimexFixed<-NULL
  }

  startingPoint<-runif(d-1)
  if(!is.null(options$par)){
    startingPoint<-options$par
  }

  # if(d<=2){
  #   methd="Brent"
  #   lower=0
  #   upper=1
  # }else{
  methd="L-BFGS-B"
  lower<-rep(0,d-1)
  upper<-rep(1,d-1)
  # }
  trace=0

  if(!is.null(options$method))
    methd=options$method
  if(!is.null(options$lower))
    lower=options$lower
  if(!is.null(options$upper))
    upper=options$upper
  if(!is.null(options$trace))
    trace=options$trace

  optRes<-list(value=Inf,aux=NA)
  if(!is.null(options$multistart)){
    optRess<-NULL
    for(i in seq(options$multistart)){
      startingPoint<-runif(d-1,min = lower,max=upper)
      if(!is.null(options$par) && i==1)
        startingPoint<-options$par
      optRess<-optim(par = startingPoint,fn = fxFixed,gr = fprimexFixed,control = list(trace=trace),method = methd,lower = lower,upper = upper)
      if(optRess$value<optRes$value){
        optRes=optRess
      }else{
        break
      }
    }
  }else{
    optRes<-optim(par = startingPoint,fn = fxFixed,gr = fprimexFixed,control = list(trace=trace),method = methd,lower = lower,upper = upper)
  }

  return(list(val=optRes$value,aux=optRes))

}


# getMaxMinMC function
#' @author Dario Azzimonti
#' @name getMaxMinMC
#' @title Coordinate profile extrema with MC
#' @description Compute coordinate profile extrema with MC
#' @param x one dimensional point where the function is to be evaluated
#' @param f function to be optimized (takes a vector y of dimension d and returns a real number)
#' @param fprime derivative of f (same format)
#' @param coord integer selecting the dimension that is fixed, the other ones are optimized over
#' @param d dimension of the input for f
#' @param options a list containing the options to be passed to the MC optimizer:
#' \itemize{
#' \item \code{par:} contains the starting point (a point in dimension d-1)
#' \item \code{numMCsamples:} number of MC samples
#' \item \code{rand} string that chooses the type of randomness in MC: "unif" (uniform in [lower,upper]), "norm" (independent normal with mean 0 and variance 1)
#' \item \code{lower:} the lower bounds for the optimization domain (see optim for details)
#' \item \code{upper:} the upper bounds for the optimization domain (see optim for details)
#' }
#' @return a real value corresponding to \eqn{max_{x_1,\dots, x_{coord-1},x_{coord+1}, \dots, x_d} f(x_1,\dots,x_d)}
getMaxMinMC = function(x,f,fprime,coord,d,options=NULL){
  fxFixed=function(y){
    res<-NA
    if(coord==1)
      res<-f(c(x,y))
    if(coord==d)
      res<-f(c(y,x))
    if(coord>1 && coord<d)
      res<-f(c(y[1:(coord-1)],x,y[coord:(d-1)]))
    return(res)
  }
  if(!is.null(fprime)){
    fprimexFixed=function(y){
      res<-NA
      if(coord==1)
        res<-fprime(c(x,y))
      if(coord==d)
        res<-fprime(c(y,x))
      if(coord>1 && coord<d)
        res<-fprime(c(y[1:(coord-1)],x,y[coord:(d-1)]))
      return(res[-coord])
    }
  }else{
    fprimexFixed<-NULL
  }

  startingPoint<-runif(d-1)
  if(!is.null(options$par)){
    startingPoint<-options$par
  }


  lower=rep(0,d-1)
  upper=rep(1,d-1)
  numMCsamples=1000
  trace=0

  if(!is.null(options$numMCsamples))
    numMCsamples=options$numMCsamples
  if(!is.null(options$lower))
    lower=options$lower
  if(!is.null(options$upper))
    upper=options$upper
  if(!is.null(options$trace))
    trace=options$trace


  optRes<-list(value=-Inf,aux=NA)

  # sample numSamplesMC/2 uniformly(/sobol?) in domain
  # evaluate f and fprime there
  # choose
#  for(i in seq(options$)){
  if(is.null(options$rand) | options$rand=="unif"){
    firstPassage<-matrix(runif((d-1)*numMCsamples,min=lower,max=upper),ncol=(d-1))
  }else if(!is.null(options$rand) & options$rand=="norm"){
    firstPassage<-matrix(rnorm((d-1)*numMCsamples),ncol=(d-1))
  }
  firstRes<-apply(firstPassage,1,fxFixed)

  idxMax<-which.max(firstRes)
  idxMin<-which.min(firstRes)


  return(list(max=list(val=firstRes[idxMax],pts=firstPassage[idxMax,]),
              min=list(val=firstRes[idxMin],pts=firstPassage[idxMin,]),
              aux=list(points=firstPassage,vals=firstRes)))

}


#' @author Dario Azzimonti
#' @name getAllMaxMin
#' @title Coordinate profile extrema with BFGS optimization
#' @description Evaluate coordinate profile extrema with full optimization.
#' @param f the function to be evaluated
#' @param fprime derivative of the function
#' @param d dimension of the input domain
#' @param options a list containing the options for this function and the subfunctions getMax, getMin
#' see documentation of getMax, getMin for details. The options only for getAllMaxMin are
#' \itemize{
#' \item \code{Design:} an optional design matrix with the discretization of each dimension, if NULL then for each dimension Design[,coord] = seq(0,1,length.out=100)
#' \item \code{heavyReturn:} If TRUE returns also all minimizers, default is FALSE.
#' \item \code{plts:} If TRUE, plots the max/min functions at each coordinate, default is FALSE.
#' \item \code{verb:} If TRUE, outputs intermediate results, default is FALSE.
#' \item \code{MonteCarlo:} If TRUE, use the MC optimizer otherwise use standard optim.
#' }
#' @return a list of two data frames (min, max) of the evaluations of \eqn{f_sup(x_i) = sup_{x_j \neq i} f(x_1,\dots,x_d) } and \eqn{f_inf(x_i) = inf_{x_j \neq i} f(x_1,\dots,x_d) }
#' for each i at the design Design. By default Design is a 100 equally spaced points for each dimension. It can be changed by defining it in options$Design
#' @examples
#' if (!requireNamespace("DiceKriging", quietly = TRUE)) {
#' stop("DiceKriging needed for this example to work. Please install it.",
#'      call. = FALSE)
#' }
#' # Compute the coordinate profile extrema with full optimization on 2d example
#'
#' # Define the function
#' g=function(x){
#'   return(-branin(x))
#' }
#' # Define the gradient
#' gprime = function(x){
#'   x1 = x[1]*15-5
#'   x2 = x[2]*15
#'   f1prime = (15*25)/(4*pi^4)*x1^3 - (15*75)/(2*pi^3)*x1^2 +
#'   (80*15)/(pi^2)*x1 - (5*15)/(pi^2)*x2*x1 +
#'   10*15/pi*x2 - 60*15/pi-10*15* (1 - 1/(8*pi))*sin(x1)
#'   f2prime = 2*15*(x2-5/(4*pi^2)*x1^2 +5/pi*x1-6)
#'   return(c(-f1prime,-f2prime))
#' }
#' # set up dimension
#' coordProf<-getAllMaxMin(f = g,fprime = gprime,d=2,options = list(multistart=4,heavyReturn=TRUE))
#'
#' \donttest{
#' # Consider threshold=-10
#' threshold<- -10
#' # obtain the points where the profiles take the threshold value
#' pp_change<-getChangePoints(threshold = threshold,allRes = coordProf)
#' # evaluate g at a grid and plot the image
#' x<-seq(0,1,,100)
#' grid<-expand.grid(x,x)
#' g_evals<- apply(X = grid,MARGIN = 1,FUN = g)
#' image(x = x,y = x,z = matrix(g_evals,nrow = 100),col = grey.colors(20))
#' contour(x=x,y=x,z=matrix(g_evals,nrow = 100), add=TRUE, nlevels = 20)
#' contour(x=x,y=x,z=matrix(g_evals,nrow = 100), add=TRUE, levels = threshold,col=2)
#' abline(h = pp_change$neverEx$`-10`[[2]],col="darkgreen",lwd=2)
#' abline(v = pp_change$neverEx$`-10`[[1]],col="darkgreen",lwd=2)
#' # Plot the coordinate profiles and a threshold
#' plotMaxMin(allRes = coordProf,threshold = threshold,changes = TRUE)
#' }
#' @export
getAllMaxMin<-function(f,fprime=NULL,d,options=NULL){

  if(is.null(options$Design)){
    warning("Design not chosen, by default [0,1]^d")
    Design<-replicate(d,seq(0,1,,100))
  }else{
    Design<-options$Design
  }
  nPtsPerDim<-nrow(Design)

  allMaxPoints<-matrix(NA,nrow=(nPtsPerDim*d),ncol=d)
  allMinPoints<-matrix(NA,nrow=(nPtsPerDim*d),ncol=d)
  results<-list(min=data.frame(matrix(NA,ncol=d,nrow = nPtsPerDim)),max=data.frame(matrix(NA,ncol=d,nrow = nPtsPerDim)))
  normGradRes<-array(NA,dim = c(nPtsPerDim,d,2))
  gradOpt<-array(NA,dim = c(nPtsPerDim,d,2))
#  tempMCvals<-list()
#  tempMCpoints<-list()
  for(coord in seq(d)){
    if(!is.null(options$verb))
      if(options$verb)
        cat("Coordinate ",coord," of ",d,"\n")
#    tempMCvals[[coord]]<-matrix(NA,ncol=nPtsPerDim,nrow=options$numMCsamples)
#    tempMCpoints[[coord]]<-array(NA,dim = c(nPtsPerDim,options$numMCsamples,d))
    for(i in seq(nPtsPerDim)){
      if(!is.null(options$verb))
        if(options$verb)
          if(i%%10==0)
            cat("Iteration ",i," of ",nPtsPerDim,"\n")

      if(!is.null(options$MonteCarlo) && options$MonteCarlo){
        temp<-getMaxMinMC(x=Design[i,coord],f = f,fprime = fprime,coord,d,options=options)
#        tempMCvals[[coord]][,i]<-temp$aux$vals
#        pp<-matrix(NA,ncol=d,nrow=options$numMCsamples)
#        pp[,-coord]<-temp$aux$points
#        pp[,coord]<-rep(Design[i,coord],options$numMCsamples)
#        tempMCpoints[[coord]][i,,]<-pp
        results$max[i,coord]<-temp$max$val
        allMaxPoints[(coord-1)*nPtsPerDim+i,-coord]<-temp$max$pts
        allMaxPoints[(coord-1)*nPtsPerDim+i,coord]<-Design[i,coord]

        results$min[i,coord]<-temp$min$val
        allMinPoints[(coord-1)*nPtsPerDim+i,-coord]<-temp$min$pts
        allMinPoints[(coord-1)*nPtsPerDim+i,coord]<-Design[i,coord]

      }else{
        options$par<-allMaxPoints[getClosePoints(x = Design[i,coord],allPts = allMaxPoints,whichDim = coord),-coord]
        options$par<-options$par*(1+rnorm(n=1,sd=diff(range(Design[,coord]))/500))
        if(!length(options$par))
          options$par<-NULL
        tempMax<-getMax(x=Design[i,coord],f = f,fprime = fprime,coord,d,options=options)
        results$max[i,coord]<-tempMax$val
        allMaxPoints[(coord-1)*nPtsPerDim+i,-coord]<-tempMax$aux$par
        allMaxPoints[(coord-1)*nPtsPerDim+i,coord]<-Design[i,coord]

        options$par<-allMinPoints[getClosePoints(x = Design[i,coord],allPts = allMinPoints,whichDim = coord),-coord]
        # 'regularization'
        options$par<-options$par*(1+rnorm(n=1,sd=diff(range(Design[,coord]))/500))
        if(!length(options$par))
          options$par<-NULL
        tempMin<-getMin(Design[i,coord],f = f,fprime = fprime,coord= coord,d = d,options=options)
        results$min[i,coord]<-tempMin$val
        allMinPoints[(coord-1)*nPtsPerDim+i,-coord]<-tempMin$aux$par
        allMinPoints[(coord-1)*nPtsPerDim+i,coord]<-Design[i,coord]

      }

      if(!is.null(fprime)){
        gr<-fprime(allMaxPoints[(coord-1)*nPtsPerDim+i,])
        gradOpt[i,coord,2]<-gr[coord]
        normGradRes[i,coord,2]<-norm(gr[-coord],"2")

        gr<-fprime(allMinPoints[(coord-1)*nPtsPerDim+i,])
        gradOpt[i,coord,1]<-gr[coord]
        normGradRes[i,coord,1]<-norm(gr[-coord],"2")
      }
    }

    if(!is.null(options$plts))
      if(options$plts){
        plot(Design[,coord],results$min[,coord],ylim=c(min(results$min[,coord]),max(results$max[,coord])),type='l',main=paste("Coordinate",coord))
        lines(Design[,coord],results$max[,coord])

 #       boxplot(tempMCvals[[coord]])
 #       lines(apply(tempMCvals[[coord]],2,max),col=2)
 #       lines(apply(tempMCvals[[coord]],2,min),col=2)
      }
    if(!is.null(options$verb))
      if(options$verb)
        cat("###############################\n")
  }

  toReturn<-list(res=results)
  if(is.null(options$heavyReturn))
    options$heavyReturn=FALSE
  if(options$heavyReturn){
    toReturn$minima<-allMinPoints
    toReturn$maxima<-allMaxPoints
    toReturn$gradRes<-normGradRes
    toReturn$gradOpt<-gradOpt
  }
  return(toReturn)
}



# approxMaxMin
#' @author Dario Azzimonti
#' @name approxMaxMin
#' @title Approximate coordinate profile functions
#' @description Evaluate profile extrema over other variables with approximations at few values
#' @param f the function to be evaluated
#' @param fprime derivative of the function
#' @param d dimension of the input domain
#' @param opts a list containing the options for this function and the subfunctions getMax, getMin or getMaxMinMC, see documentation of getMax, getMin, getMaxMinMC for details. The options only for approxMaxMin are
#' \itemize{
#' \item \code{limits:} an optional list with the upper and lower limits of each dimension, if NULL then for each dimension limits are 0,1
#' \item \code{smoother:} Select which smoother to use:a string that selects which smoother to use: \itemize{
#'       \item \code{"1order"}: first order interpolation with gradient
#'       \item \code{"splineSmooth"}: smoothing spline with default degrees of freedom (DEFAULT OPTION)
#'       \item \code{"quantSpline"}: profile inf and profile sup approximated with quantile spline regression at levels 0.1 and 0.9 respectively
#' }
#' \item \code{heavyReturn:} If TRUE returns also all minimizers, default is FALSE.
#' \item \code{initDesign:} The design of few points where the expensive sup is evaluated.
#' \item \code{fullDesignSize:} The full design where the function is approximated.
#' \item \code{multistart:} number of multistarts for optim procedure.
#' \item \code{MonteCarlo:} if TRUE, computes sup with Monte Carlo procedure.
#' \item \code{numMCsamples:} number of MC samples for the sup.
#' \item \code{plts:} If TRUE, plots the max/min functions at each coordinate, default is FALSE.
#' \item \code{verb:} If TRUE, outputs intermediate results, default is FALSE.
#' }
#' @return a list of two data frames (min, max) of the evaluations of \eqn{f_sup(x_i) = sup_{x_j \neq i} f(x_1,\dots,x_d) } and \eqn{f_inf(x_i) = inf_{x_j \neq i} f(x_1,\dots,x_d) }
#' for each i at the design Design. By default Design is a 100 equally spaced points for each dimension. It can be changed by defining it in options$Design
#' @examples
#' if (!requireNamespace("DiceKriging", quietly = TRUE)) {
#' stop("DiceKriging needed for this example to work. Please install it.",
#'      call. = FALSE)
#' }
#' # Compute the coordinate profile extrema with full optimization on 2d example
#'
#' # Define the function
#' g=function(x){
#'   return(-branin(x))
#' }
#' # Define the gradient
#' gprime = function(x){
#'   x1 = x[1]*15-5
#'   x2 = x[2]*15
#'   f1prime = (15*25)/(4*pi^4)*x1^3 - (15*75)/(2*pi^3)*x1^2 +
#'   (80*15)/(pi^2)*x1 - (5*15)/(pi^2)*x2*x1 +
#'   10*15/pi*x2 - 60*15/pi-10*15* (1 - 1/(8*pi))*sin(x1)
#'   f2prime = 2*15*(x2-5/(4*pi^2)*x1^2 +5/pi*x1-6)
#'   return(matrix(c(-f1prime,-f2prime),nrow=1))
#' }
#'
#' # generic approximation options
#' init_des<-lhs::maximinLHS(15,2)
#' options_approx<- list(multistart=4,heavyReturn=TRUE,initDesign=init_des,fullDesignSize=100)
#'
#' # 1order approximation
#' options_approx$smoother<-"1order"
#' coordProf_approx_1order<-approxMaxMin(f = g,fprime = gprime,d=2,opts = options_approx)
#'
#' # quantile regression
#' options_approx$smoother<-"quantSpline"
#' coordProf_approx_quantReg<-approxMaxMin(f = g,fprime = gprime,d=2,opts = options_approx)
#'
#'
#' \donttest{
#' # Consider threshold=-10
#' threshold<- -10
#' # obtain the points where the profiles take the threshold value
#' pp_change<-getChangePoints(threshold = threshold,allRes = coordProf_approx_quantReg)
#' # evaluate g at a grid and plot the image
#' x<-seq(0,1,,100)
#' grid<-expand.grid(x,x)
#' g_evals<- apply(X = grid,MARGIN = 1,FUN = g)
#' image(x = x,y = x,z = matrix(g_evals,nrow = 100),col = grey.colors(20))
#' contour(x=x,y=x,z=matrix(g_evals,nrow = 100), add=TRUE, nlevels = 20)
#' contour(x=x,y=x,z=matrix(g_evals,nrow = 100), add=TRUE, levels = threshold,col=2)
#' abline(h = pp_change$neverEx$`-10`[[2]],col="darkgreen",lwd=2)
#' abline(v = pp_change$neverEx$`-10`[[1]],col="darkgreen",lwd=2)
#' # Plot the coordinate profiles and a threshold
#' plotMaxMin(allRes = coordProf_approx_1order,threshold = threshold,changes = TRUE)
#' plotMaxMin(allRes = coordProf_approx_quantReg,threshold = threshold,changes = TRUE)
#' }
#' @export
approxMaxMin=function(f,fprime=NULL,d,opts=NULL){
  if(is.null(opts$limits)){
    limits<-list(lower=rep(0,d),upper=rep(1,d))
  }
#  nPtsPerDim<-rep(100,d)
  if(is.null(opts$initDesign)){
    opts$initDesign<-maximinLHS(ceiling(sqrt(d)*10),d)
  }
  # It is better to include points on the border
#  if(!is.null(opts$smoother) && opts$smoother=="quantSpline"){
    for(coord in seq(d)){
      opts$initDesign[,coord]<-c(limits$lower[coord],sort(opts$initDesign[,coord])[c(-1,-nrow(opts$initDesign))],limits$upper[coord])
    }
#  }
  if(is.null(opts$fullDesignSize)){
    opts$fullDesignSize<-80*d
  }

  aa<-getAllMaxMin(f,fprime,d,options=list(Design=opts$initDesign,multistart=opts$multistart,
                                           heavyReturn=T,MonteCarlo=opts$MonteCarlo,numMCsamples=opts$numMCsamples,verb=opts$verb))

  # initialize the max/min values
  sK_max<-data.frame(matrix(NA,nrow=opts$fullDesignSize,ncol=d))
  sK_min<-data.frame(matrix(NA,nrow=opts$fullDesignSize,ncol=d))

  if(!is.null(fprime)){
    aa$res$grad_min<-matrix(NA,ncol=ncol(aa$res$min),nrow=nrow(aa$res$min))
    colnames(aa$res$grad_min)<-colnames(aa$res$min)

    aa$res$grad_max<-matrix(NA,ncol=ncol(aa$res$max),nrow=nrow(aa$res$max))
    colnames(aa$res$grad_max)<-colnames(aa$res$max)
  }

  for(coord in seq(d)){
    # the gradient of the sup/inf function is saved
    nn<-length(aa$res$max[,coord])
    for(i in seq(nn)){
      if(!is.null(fprime)){
        yy<-aa$minima[(coord-1)*nn+i,]
        aa$res$grad_min[i,coord]<-fprime(yy)[,coord]
        yy<-aa$maxima[(coord-1)*nn+i,]
        aa$res$grad_max[i,coord]<-fprime(yy)[,coord]
      }

    }
    newPoints<-matrix(seq(limits$lower[coord],limits$upper[coord],,opts$fullDesignSize),ncol=1)

    if(!is.null(opts$smoother) && opts$smoother=="1order"){
      sK_max[,coord]<- kGradSmooth(newPoints=newPoints,profPoints = aa$maxima[((coord-1)*nn+1):(coord*nn),coord],
                                   profEvals=aa$res$max[,coord], profGradient=aa$res$grad_max[,coord])

      sK_min[,coord]<-kGradSmooth(newPoints=newPoints,profPoints = aa$minima[((coord-1)*nn+1):(coord*nn),coord],
                                  profEvals=aa$res$min[,coord], profGradient=aa$res$grad_min[,coord])
    }else if(!is.null(opts$smoother) && opts$smoother=="quantSpline"){
      dd_fit<-data.frame(x=aa$maxima[((coord-1)*nn+1):(coord*nn),coord],y=aa$res$max[,coord])
      fit.max <- rq(y ~ bs(x, df=max(nn%/%1.5,3)),data=dd_fit,tau=0.9)
      dd_fit<-data.frame(x=aa$minima[((coord-1)*nn+1):(coord*nn),coord],y=aa$res$min[,coord])
      fit.min <- rq(y ~ bs(x, df=max(nn%/%1.5,3)), tau=0.1,data=dd_fit)
      newPoints<-data.frame(x=newPoints)

      sK_max[,coord]<- predict.rq(object = fit.max,newdata = newPoints)
      sK_min[,coord]<- predict.rq(object = fit.min,newdata = newPoints)
    }else {
      sk_maxSmooth<-smooth.spline(x=aa$maxima[((coord-1)*nn+1):(coord*nn),coord],y=aa$res$max[,coord])
      sk_minSmooth<-smooth.spline(x=aa$minima[((coord-1)*nn+1):(coord*nn),coord],y=aa$res$min[,coord])

      sK_max[,coord]<-predict(sk_maxSmooth,newPoints)$y
      sK_min[,coord]<-predict(sk_minSmooth,newPoints)$y
    }




    ## Debug plots
    if(!is.null(opts$debug) && opts$debug){
  #    par(mfrow=c(1,2))
      # Plot  max/min function
      ylimTemp<-range(c(sK_min[,coord],aa$res$min[,coord],sK_max[,coord],aa$res$max[,coord]))
      plot(aa$minima[((coord-1)*nn+1):(coord*nn),coord],aa$res$min[,coord],main=paste("Coordinate",coord,", Min"),
           xlab="x",ylab="f",ylim=ylimTemp)
      lines(newPoints,sK_min[,coord],type='l',col=3)
      points(aa$maxima[((coord-1)*nn+1):(coord*nn),coord],aa$res$max[,coord])
      lines(newPoints,sK_max[,coord],type='l',col=3)

      if(!is.null(opts$smooth) && opts$smooth){
        points(sk_minSmooth$x,sk_minSmooth$y,col=2)
        points(sk_maxSmooth$x,sk_maxSmooth$y,col=2)
      }


      # Plot min function
#      ylimTemp<-range(c(sK_min[,coord],aa$res$min[,coord]))
#      plot(aa$minima[((coord-1)*nn+1):(coord*nn),coord],aa$res$min[,coord],main=paste("Coordinate",coord,", Min"),
#           xlab="x",ylab="f")
#      lines(newPoints,sK_min[,coord],type='l',col=3)
#      points(sk_minSmooth$x,sk_minSmooth$y,col=2)

      # This code uses gg_min, which needs to be initialized before the function
#      ylimTemp<-c(min(c(gg_min[,coord],aa$res$min[,coord])),max(c(gg_min[,coord],aa$res$min[,coord])))
#      plot(dd,gg_min[,coord],ylim=ylimTemp,type='l',main=paste("Coordinate",coord,", Min"),
#           xlab="x",ylab="f")
#      points(aa$minima[((coord-1)*nn+1):(coord*nn),coord],aa$res$min[,coord])
#      points(aa$minima[((coord-1)*nn+1):(coord*nn),coord],aa$res$grad_min[,coord],pch=17)
#      lines(dd,allRes_km$res$min[,coord])
#      abline(h=0,col=3)
#      abline(h=-10,col=2)
#      lines(newPoints,sK_min[,coord],col=4)
      #  points(allMinPoints[(coord-1)*nPtsPerDim[coord]+1:i,coord],results$min[1:i,coord])

      # Plot the max function
#      ylimTemp<-range(c(sK_max[,coord],aa$res$max[,coord]))
#      plot(aa$maxima[((coord-1)*nn+1):(coord*nn),coord],aa$res$max[,coord],main=paste("Coordinate",coord,", Max"),
#           xlab="x",ylab="f")
#      lines(newPoints,sK_max[,coord],type='l',col=3)
    #  lines(pp_sk_max_smooth$x,pp_sk_max_smooth$y,type='l',col=2)
#      points(sk_maxSmooth$x,sk_maxSmooth$y,col=2)

      # This code uses gg_min, which needs to be initialized before the function
#      ylimTemp<-c(min(c(gg_max[,coord],aa$res$max[,coord])),max(c(gg_max[,coord],aa$res$max[,coord])))
#      plot(dd,gg_max[,coord],ylim=ylimTemp,type='l',main=paste("Coordinate",coord,", Max"),
#           xlab="x",ylab="f")
#      points(aa$maxima[((coord-1)*nn+1):(coord*nn),coord],aa$res$max[,coord])
#      points(aa$maxima[((coord-1)*nn+1):(coord*nn),coord],aa$res$grad_max[,coord],pch=17)
#      lines(dd,allRes_km$res$max[,coord])
#      abline(h=0,col=3)
#      abline(h=-10,col=2)
#      lines(newPoints,sK_max[,coord],col=4)
      #  points(allMaxPoints[(coord-1)*nPtsPerDim[coord]+1:i,coord],results$max[1:i,coord])
    }
  }

  res=list(min=sK_min,max=sK_max)
  toReturn<-list(res=res)

  if(!is.null(opts$heavyReturn) && opts$heavyReturn){
    aa$design<-opts$initDesign
    toReturn$profPoints<-aa
  }
  return(toReturn)

}


#' @author Dario Azzimonti
#' @title First order approximation
#' @name kGradSmooth
#' @description Compute first order approximation of function from evaluations and gradient
#' @param newPoints vector of points where to approximate the function
#' @param profPoints locations where the function was evaluated
#' @param profEvals value of the evaluation at profPoints
#' @param profGradient value of the gradient at profPoints
#' @return approximated values of the function at newPoints
kGradSmooth<-function(newPoints,profPoints,profEvals,profGradient){
  newPoints<-matrix(newPoints,ncol=1)
  clPts<-apply(newPoints,MARGIN = 1,FUN = function(x){which.min(abs(x-profPoints))})
  ww<-rep(NA,nrow(newPoints))
  for(i in seq(nrow(newPoints))){
    ww[i]<-profEvals[clPts[i]]+(newPoints[i,]-profPoints[clPts[i]])*profGradient[clPts[i]]
  }
  return(ww)
}


#' @author Dario Azzimonti
#' @title Find close points
#' @name getClosePoints
#' @description Obtain points close in one specific dimension
#' @param x one dimensional point
#' @param allPts dataframe containing a list of d dimensional points
#' @param whichDim integer defining the dimension of x
#' @return the index in allPts (row number) of the closest point in allPts to x along the whichDim dimension
getClosePoints<-function(x,allPts,whichDim){
  rr<-which.min(abs(x-allPts[,whichDim]))
  if(!length(rr))
    rr<-NULL
  return(rr)
}

# coordinateProfiles function
#' @author Dario Azzimonti
#' @title Coordinate profiles starting from a kriging model
#' @description The function coordinateProfiles computes the profile extrema functions for the posterior mean of a Gaussian process and its confidence bounds
#' @param object either a \link[DiceKriging]{km} model or a list containing partial results. If \code{object} is a km model then all computations are carried out. If \code{object} is a list, then the function carries out all computations to complete the list results.
#' @param threshold the threshold of interest
#' @param options_full an optional list of options for getAllMaxMin, see \link{getAllMaxMin} for details.
#' @param options_approx an optional list of options for approxMaxMin, see \link{approxMaxMin} for details.
#' @param uq_computations boolean, if TRUE the uq computations for the profile mean are computed.
#' @param plot_level an integer to select the plots to return (0=no plots, 1=basic plots, 2= all plots)
#' @param plot_options an optional list of parameters for plots. See \link{setPlotOptions} for currently available options.
#' @param CI_const an optional vector containing the constants for the CI. If not NULL, then profiles extrema for \eqn{m_n(x) \pm CI_const[i]*s_n(x,x)} are computed.
#' @param return_level an integer to select the amount of details returned
#' @param ... additional parameters to be passed to \link{coordProf_UQ}.
#' @return If return_level=1 a list containing
#' \itemize{
#' \item \code{profMean_full:} the results of \code{getAllMaxMin} for the posterior mean
#' \item \code{profMean_approx:} the results of \code{approxMaxMin} for the posterior mean
#' \item \code{res_UQ:} the results of \code{coordProf_UQ} for the posterior mean
#' }
#' if return_level=2 the same list as above but also including
#' \itemize{
#' \item \code{abs_err:} the vector of maximum absolute approximation errors for the profile inf /sup on posterior mean for the chosen approximation
#' \item \code{times:}  a list containing
#' 	\itemize{
#' 	\item \code{full:} computational time for the full computation of profile extrema
#' 	\item \code{approx:} computational time for the approximate computation of profile extrema
#' 	}
#' }
#' @examples
#' if (!requireNamespace("DiceKriging", quietly = TRUE)) {
#' stop("DiceKriging needed for this example to work. Please install it.",
#'      call. = FALSE)
#' }
#' # Compute a kriging model from 50 evaluations of the Branin function
#' # Define the function
#' g=function(x){
#'   return(-branin(x))
#' }
#' gp_des<-lhs::maximinLHS(20,2)
#' reals<-apply(gp_des,1,g)
#' kmModel<-km(design = gp_des,response = reals,covtype = "matern3_2")
#'
#' threshold=-10
#'
#' # Compute coordinate profiles on the posterior mean
#' # Increase multistart and size of designs for more precise results
#' options_full<-list(multistart=2,heavyReturn=TRUE, Design = replicate(2,seq(0,1,,50)))
#' init_des<-lhs::maximinLHS(12,2)
#' options_approx<- list(multistart=2,heavyReturn=TRUE,initDesign=init_des,fullDesignSize=50)
#' cProfilesMean<-coordinateProfiles(object=kmModel,threshold=threshold,options_full=options_full,
#'                                   options_approx=options_approx,uq_computations=FALSE,
#'                                   plot_level=3,plot_options=NULL,CI_const=NULL,return_level=2)
#' \dontrun{
#' # Coordinate profiles with UQ with approximate profiles
#' plot_options<-list(save=FALSE, titleProf = "Coordinate profiles",
#'                    title2d = "Posterior mean",qq_fill=TRUE)
#' cProfilesUQ<-coordinateProfiles(object=cProfilesMean,threshold=threshold,options_full=options_full,
#'                                   options_approx=options_approx,uq_computations=TRUE,
#'                                   plot_level=3,plot_options=NULL,CI_const=NULL,return_level=2)
#'
#' # Coordinate profiles with UQ with fully optim profiles
#' options_full_sims<-list(multistart=4,heavyReturn=TRUE, Design = replicate(2,seq(0,1,,60)))
#' cProfilesUQ<-coordinateProfiles(object=cProfilesMean,threshold=threshold,options_full=options_full,
#'                                   options_approx=options_approx,uq_computations=TRUE,
#'                                   plot_level=3,plot_options=NULL,CI_const=NULL,return_level=2,
#'                                   options_full_sims=options_full_sims)
#' }
#' @export
coordinateProfiles = function(object,threshold,options_full=NULL,options_approx=NULL,uq_computations=FALSE,plot_level=0,plot_options=NULL,CI_const=NULL,return_level=1,...){

  # Check object
  if(is(object,"km")){
    object<-list(kmModel=object)
  }else if(!is.list(object)){
    stop("object must be either a list or a km object")
  }

  # Options setup
  if(is.null(options_full)){
    options_full<-list(multistart=8,heavyReturn=TRUE,Design = replicate(object$kmModel@d,seq(0,1,,100)))
  }

  if(is.null(options_approx)){
    init_des<-maximinLHS(ceiling(sqrt(object$kmModel@d)*10),object$kmModel@d)
    options_approx<- list(multistart=8,heavyReturn=TRUE,initDesign=init_des,fullDesignSize=100,smoother="1order")
  }

  # save the dimension of the input
  d<-object$kmModel@d

  # save number of thresholds
  num_T<-length(threshold)

  # Set up plot options
  if(!is.null(options_full$Design))
    plot_options$design = options_full$Design
  plot_options<-setPlotOptions(plot_options = plot_options,d=d,num_T=num_T,kmModel=object$kmModel)

  ## posterior mean part ##
  # Let us define the posterior mean function and gradient in a 'optimizer-friendly' way
  g <-function(y){
    y<-matrix(y,ncol=d)
    colnames(y)<-colnames(object$kmModel@X)
    return(predict.km(object = object$kmModel,newdata = y,type="UK",light.return = TRUE,se.compute = FALSE)$mean)
  }
  gprime <- function(y){
    y<-matrix(y,ncol=d)
    colnames(y)<-colnames(object$kmModel@X)
    return(gradKm_dnewdata(object = object$kmModel,newdata = y,type = "UK",light.return = TRUE,se.compute = FALSE)$mean)
  }

  # Compute profile mean full, if not already computed
  if(is.null(object$profMean_full)){
    timeIn<-get_nanotime()
    object$profMean_full<-getAllMaxMin(f = g,fprime = gprime,d = d,options = options_full)
    timeMean_full<-(get_nanotime()-timeIn)*1e-9
  }else{
    if(return_level>1 && is.null(object$more$times$full))
      object$more<-list(times=list(full=NA))
  }
  #    profMean_full,profMean_approx,res_UQ
  #    times<-list(full=timeMean_full,approx=timeMean_approx)
  #    results$more<-list(abs_err=abs_err,times=times)
  # Compute profile extrema with full optim



  # Obtain the threshold points selected by profile extrema
  changePP<-getChangePoints(threshold = threshold,allRes = object$profMean_full)


  # Plot posterior mean, ONLY for dimension==2
  if(plot_level>=2 && d==2){
    # since dimension==2 we can plot the posterior mean
    newdata<-expand.grid(seq(0,1,,100),seq(0,1,,100))
    colnames(newdata)<-colnames(object$kmModel@X)
    pred2d<-predict.km(object$kmModel,newdata = newdata,type = "UK",light.return = TRUE,se.compute = FALSE)

    # Plot the posterior mean and visualize the actual excursion set and the regions of no-excursion according to the profile extrema functions.
    if(plot_options$save)
      cairo_pdf(filename = paste(plot_options$folderPlots,"2dpostMean_1",plot_options$id_save,".pdf",sep=""),width = 12,height = 12)
    par(mar = c(5, 5, 4, 2) + 0.1)
    image(matrix(pred2d$mean,nrow = 100),col=gray.colors(20), main=plot_options$title2d,xlab = "", ylab = "", #colnames(object$kmModel@X)[1],ylab= colnames(object$kmModel@X)[2],
          cex.main=3,cex.axis=1.8,cex.lab=2.8)
    contour(matrix(pred2d$mean,nrow = 100),add=T,nlevels = 10,lwd=1.5,labcex=1.2)
    contour(matrix(pred2d$mean,nrow = 100),add=T,levels = threshold,col=plot_options$col_thresh,lwd=3,labcex=1.5)
    for(tt in seq(num_T)){
      abline(v = changePP$neverEx[[tt]][[1]],col=plot_options$col_CCPthresh_nev[tt],lwd=2.5)
      abline(h = changePP$neverEx[[tt]][[2]],col=plot_options$col_CCPthresh_nev[tt],lwd=2.5)
      abline(h = changePP$alwaysEx[[tt]][[2]],col=plot_options$col_CCPthresh_alw[tt],lwd=2.5)
      abline(h = changePP$alwaysEx[[tt]][[2]],col=plot_options$col_CCPthresh_alw[tt],lwd=2.5)
    }
    if(plot_options$fun_evals>0){
      points(object$kmModel@X,pch=17,cex=1.6)
    }
    if(plot_options$save)
      dev.off()
  }

  # Plot the profile extrema functions
  if(plot_level>0){
    colnames(object$profMean_full$res$min)<-plot_options$coord_names
    if(plot_options$save)
      cairo_pdf(filename = paste(plot_options$folderPlots,"profMean_full",plot_options$id_save,".pdf",sep=""),width = 12,height = 12)
    plotMaxMin(allRes = object$profMean_full,threshold = threshold,Design = options_full$Design)
    if(plot_options$save)
      dev.off()
  }

  # Compute profile extrema with approximation, if not already computed
  if(is.null(object$profMean_approx)){
    timeIn<-get_nanotime()
    object$profMean_approx<-approxMaxMin(f = g,fprime = gprime,d = d,opts = options_approx)
    timeMean_approx<-(get_nanotime()-timeIn)*1e-9
  }else{
    if(return_level>1 && is.null(object$more$times$approx))
      object$more$times$approx=NA
  }

  if(plot_level>0){
    oldpar<-par()
    if(plot_options$save)
      cairo_pdf(filename = paste(plot_options$folderPlots,"profMean_comparison",plot_options$id_save,".pdf",sep=""),width = 12,height = 12)
    mfrows<-switch(d,"1"=c(1,1),"2"=c(2,1),"3"=c(2,2),"4"=c(2,2),"5"=c(3,2),"6"=c(3,2))
    par(mfrow=mfrows, mar = c(4, 5, 3, 1) + 0.1)
    for(coord in seq(d)){
      if(is.null(plot_options$ylim)){
        ylimTemp<-c(min(object$profMean_full$res$min[,coord]),max(object$profMean_full$res$max[,coord]))
      }else{
        ylimTemp<-plot_options$ylim[coord,]
      }
      # create the title of each plot
      title_string<-paste("Coordinate",plot_options$coord_names[coord])

      plot(plot_options$design[,coord],object$profMean_full$res$min[,coord],ylim=ylimTemp,type='l',main=title_string,
           xlab=expression(eta),ylab="f",lty=1,cex.main=3,cex.lab=2.5,cex.axis=1.8)
      lines(plot_options$design[,coord],object$profMean_full$res$max[,coord],lty=1)
      lines(plot_options$design[,coord],object$profMean_approx$res$max[,coord],lty=3,col=4,lwd=2)
      points(object$profMean_approx$profPoints$design[,coord],object$profMean_approx$profPoints$res$max[,coord],col=4)
      lines(plot_options$design[,coord],object$profMean_approx$res$min[,coord],lty=4,col=4,lwd=2)
      points(object$profMean_approx$profPoints$design[,coord],object$profMean_approx$profPoints$res$min[,coord],col=4)
      for(tt in seq(num_T)){
        abline(v=changePP$neverEx[[tt]][[coord]],col=plot_options$col_CCPthresh_nev[tt],lwd=2.5)
        abline(v=changePP$alwaysEx[[tt]][[coord]],col=plot_options$col_CCPthresh_alw[tt],lwd=2.5)
      }
      abline(h = threshold,col=plot_options$col_thresh,lwd=2)
      if(!is.null(plot_options$legend)){
        legend("bottomleft",c(as.expression(substitute(paste(P[coord]^sup,f," (full)"),list(coord=coord))),
                              as.expression(substitute(paste(P[coord]^inf,f," (full)"),list(coord=coord))),
                              as.expression(substitute(paste(P[coord]^sup,f," (1order)"),list(coord=coord))),
                              as.expression(substitute(paste(P[coord]^inf,f," (1order)"),list(coord=coord))),
                              "threshold"),
               lty=c(1,1,3,4,1),col=c(1,1,4,4,2),cex=1.8,lwd=c(2,2,2,2,2))
      }
    }
    par(mfrow=c(1,1))
    if(plot_options$save)
      dev.off()
    #  par(oldpar)
  }

  # save absolute errors with approximation
  if(nrow(object$profMean_approx$res$min)==nrow(object$profMean_full$res$min)){
    abs_err_min<-max(abs(object$profMean_approx$res$min-object$profMean_full$res$min)/object$profMean_full$res$min)
    abs_err_max<-max(abs(object$profMean_approx$res$max-object$profMean_full$res$max)/object$profMean_full$res$max)
    abs_err<-c(abs_err_min,abs_err_max)
  }else{
    warning("Different discretizations for approx and full profiles, error not computed!")
    abs_err = c(NA,NA)
  }
  ## End posterior mean part ##

  ## quantiles part ##
  if(!is.null(CI_const)){
    g_s2<-function(y){
      y<-matrix(y,ncol=d)
      colnames(y)<-colnames(object$kmModel@X)
      return(predict.km(object = object$kmModel,newdata = y,type="UK",light.return = TRUE,se.compute = TRUE)$sd)
    }
    g_s2prime <- function(y){
      y<-matrix(y,ncol=d)
      colnames(y)<-colnames(object$kmModel@X)
      s2<-predict.km(object = object$kmModel,newdata = y,type="UK",light.return = TRUE,se.compute = TRUE)$sd
      s2grad<-gradKm_dnewdata(object = object$kmModel,newdata = y,type = "UK",se.compute = TRUE)$s2
      return(0.5*s2grad/s2)
    }

    # Compute profile mean full, if not already computed
    if(is.null(object$profSd_full)){
      timeIn<-get_nanotime()
      object$profSd_full<-getAllMaxMin(f = g_s2,fprime = g_s2prime,d = d,options = options_full)
      timeVar_full<-(get_nanotime()-timeIn)*1e-9
    }else{
      if(return_level>1 && is.null(object$more$times$full))
        object$more$times$full_var=NA
    }

    # Obtain the threshold points selected by profile extrema
    prof_CI_const_full<-list()
    temp<-list()
    changePP_CI_full<-list()
    for(i in seq(length(CI_const))){
      temp_m<-list(res=list())
      temp_m$res<-list(min= object$profMean_full$res$min - CI_const[i]*object$profSd_full$res$min,
                       max= object$profMean_full$res$max - CI_const[i]*object$profSd_full$res$max)
      temp_M<-list(res=list())
      temp_M$res<-list(min= object$profMean_full$res$min + CI_const[i]*object$profSd_full$res$min,
                       max= object$profMean_full$res$max + CI_const[i]*object$profSd_full$res$max)
      prof_CI_const_full[[i]]<-list(lower=temp_m, upper=temp_M)
      changePP_CI_full[[i]]<-list(lower=getChangePoints(threshold = threshold,allRes = temp_m),
                                  upper=getChangePoints(threshold = threshold,allRes = temp_M))
    }

    # Compute profile extrema var with approximation, if not already computed
    #    if(is.null(object$profVar_approx)){
    #      timeIn<-get_nanotime()
    #      object$profVar_approx<-approxMaxMin(f = g_s2,fprime = g_s2prime,d = d,opts = options_approx)
    #      timeVar_approx<-(get_nanotime()-timeIn)*1e-9
    #    }else{
    #      if(return_level>1 && is.null(object$more$times$approx))
    #        object$more$times$approx=NA
    #    }

    if(plot_level>0){
      oldpar<-par()
      if(plot_options$save)
        cairo_pdf(filename = paste(plot_options$folderPlots,"profMeanCI_comparison",plot_options$id_save,".pdf",sep=""),width = 12,height = 12)
      mfrows<-switch(d,"1"=c(1,1),"2"=c(2,1),"3"=c(2,2),"4"=c(2,2),"5"=c(3,2),"6"=c(3,2))
      par(mfrow=mfrows, mar = c(4, 5, 3, 1) + 0.1)
      for(coord in seq(d)){
        if(is.null(plot_options$ylim)){
          ylimTemp_CI<-c(object$profMean_full$res$min[,coord],object$profMean_full$res$max[,coord])
          for(i in seq(length(CI_const))){
            ylimTemp_CI<-c(ylimTemp_CI,prof_CI_const_full[[i]]$lower$res$min[,coord],
                           prof_CI_const_full[[i]]$lower$res$max[,coord],
                           prof_CI_const_full[[i]]$upper$res$min[,coord],
                           prof_CI_const_full[[i]]$upper$res$max[,coord])
          }
          ylimTemp_CI<-range(ylimTemp_CI)
        }else{
          ylimTemp<-plot_options$ylim[coord,]
        }

        # create the title of each plot
        title_string<-paste("Coordinate",plot_options$coord_names[coord])
        # full mean
        plot(plot_options$design[,coord],object$profMean_full$res$min[,coord],ylim=ylimTemp_CI,type='l',main=title_string,
             xlab=expression(eta),ylab="f",lty=1,cex.main=3,cex.lab=2.5,cex.axis=1.8)
        lines(plot_options$design[,coord],object$profMean_full$res$max[,coord],lty=1)
        # approx mean
        lines(plot_options$design[,coord],object$profMean_approx$res$max[,coord],lty=3,col=4,lwd=2)
        points(object$profMean_approx$profPoints$design[,coord],object$profMean_approx$profPoints$res$max[,coord],col=4)
        lines(plot_options$design[,coord],object$profMean_approx$res$min[,coord],lty=4,col=4,lwd=2)
        points(object$profMean_approx$profPoints$design[,coord],object$profMean_approx$profPoints$res$min[,coord],col=4)
        for(tt in seq(num_T)){
          abline(v=changePP$neverEx[[tt]][[coord]],col=plot_options$col_CCPthresh_nev[tt],lwd=2.5)
          abline(v=changePP$alwaysEx[[tt]][[coord]],col=plot_options$col_CCPthresh_alw[tt],lwd=2.5)
        }
        # full CI_const
        CI_cols1<-heat.colors(n = length(CI_const)+1,alpha=0.7)
        CI_cols2<-terrain.colors(n = length(CI_const)+1,alpha=0.7)
        for(i in seq(length(CI_const))){
          lines(plot_options$design[,coord],prof_CI_const_full[[i]]$upper$res$max[,coord],lty=4,col=CI_cols1[i])
          lines(plot_options$design[,coord],prof_CI_const_full[[i]]$lower$res$max[,coord],lty=4,col=CI_cols1[i])
          lines(plot_options$design[,coord],prof_CI_const_full[[i]]$upper$res$min[,coord],lty=4,col=CI_cols2[i])
          lines(plot_options$design[,coord],prof_CI_const_full[[i]]$lower$res$min[,coord],lty=4,col=CI_cols2[i])
        }
        abline(h = threshold,col=plot_options$col_thresh,lwd=2)
        if(!is.null(plot_options$legend)){
          legend("bottomleft",c(as.expression(substitute(paste(P[coord]^sup,f," (full)"),list(coord=coord))),
                                as.expression(substitute(paste(P[coord]^inf,f," (full)"),list(coord=coord))),
                                as.expression(substitute(paste(P[coord]^sup,f," (1order)"),list(coord=coord))),
                                as.expression(substitute(paste(P[coord]^inf,f," (1order)"),list(coord=coord))),
                                "threshold"),
                 lty=c(1,1,3,4,1),col=c(1,1,4,4,2),cex=1.8,lwd=c(2,2,2,2,2))
        }
      }
      par(mfrow=c(1,1))
      if(plot_options$save)
        dev.off()
      #  par(oldpar)
    }
    object$profCI_full<-prof_CI_const_full
  }


  ## UQ part ##
  #  res_UQ<-NULL
  if(uq_computations){# && is.null(object$res_UQ)){
    if(is.null(object$res_UQ$kmModel))
      object$res_UQ$kmModel=object$kmModel
    object$res_UQ<-coordProf_UQ(object=object$res_UQ,threshold=threshold,allResMean=object$profMean_full,
                                options_approx=options_approx,plot_level=max(0,plot_level-1),
                                plot_options=plot_options,return_level=max(1,return_level-1),...)

  }else{
    object$res_UQ<-NULL
  }
  #results<-object =list(kmModel=object$kmModel, profMean_full=object$profMean_full,profMean_approx=object$profMean_approx,res_UQ=object$res_UQ)
  if(return_level==1){
    return(object)
  }else{
    if(is.null(object$more)){
      times<-list(full=timeMean_full,approx=timeMean_approx)
      object$more<-list(abs_err=abs_err,times=times)
    }
    return(object)
  }

}

# obliqueProfiles function
#' @author Dario Azzimonti
#' @name obliqueProfiles
#' @title Oblique coordinate profiles starting from a kriging model
#' @description The function obliqueProfiles computes the (oblique) profile extrema functions for the posterior mean of a Gaussian process and its confidence bounds
#' @param object either a \link[DiceKriging]{km} model or a list containing partial results. If \code{object} is a km model then all computations are carried out. If \code{object} is a list, then the function carries out all computations to complete the list results.
#' @param allPsi a list containing the matrices Psi (dim \eqn{pxd}) for which to compute the profile extrema
#' @param threshold the threshold of interest
#' @param options_full an optional list of options for getProfileExtrema, see \link{getProfileExtrema} for details.
#' @param options_approx an optional list of options for approxProfileExtrema, see \link{approxProfileExtrema} for details.
#' @param uq_computations boolean, if TRUE the uq computations for the profile mean are computed.
#' @param plot_level an integer to select the plots to return (0=no plots, 1=basic plots, 2= all plots)
#' @param plot_options an optional list of parameters for plots. See \link{setPlotOptions} for currently available options.
#' @param CI_const an optional vector containing the constants for the CI. If not NULL, then profiles extrema for \eqn{m_n(x) \pm CI_const[i]*s_n(x,x)} are computed.
#' @param return_level an integer to select the amount of details returned
#' @param ... additional parameters to be passed to \link{obliqueProf_UQ}.
#' @return If return_level=1 a list containing
#' \itemize{
#' \item \code{profMean_full:} the results of \code{getProfileExtrema} for the posterior mean
#' \item \code{profMean_approx:} the results of \code{approxProfileExtrema} for the posterior mean
#' \item \code{res_UQ:} the results of \code{obliqueProf_UQ} for the posterior mean
#' }
#' if return_level=2 the same list as above but also including
#' \itemize{
#' \item \code{abs_err:} the vector of maximum absolute approximation errors for the profile inf /sup on posterior mean for the chosen approximation
#' \item \code{times:}  a list containing
#' 	\itemize{
#' 	\item \code{full:} computational time for the full computation of profile extrema
#' 	\item \code{approx:} computational time for the approximate computation of profile extrema
#' 	}
#' }
#' @examples
#' if (!requireNamespace("DiceKriging", quietly = TRUE)) {
#' stop("DiceKriging needed for this example to work. Please install it.",
#'      call. = FALSE)
#' }
#' # Compute a kriging model from 50 evaluations of the Branin function
#' # Define the function
#' g=function(x){
#'   return(-branin(x))
#' }
#' gp_des<-lhs::maximinLHS(20,2)
#' reals<-apply(gp_des,1,g)
#' kmModel<-km(design = gp_des,response = reals,covtype = "matern3_2")
#'
#' threshold=-10
#'
#' # Compute oblique profiles on the posterior mean
#' # (for theta=0 it is equal to coordinateProfiles)
#' options_full<-list(multistart=4,heavyReturn=TRUE,discretization=100)
#' options_approx<- list(multistart=4,heavyReturn=TRUE,initDesign=NULL,fullDesignSize=100)
#' theta=pi/4
#' allPsi = list(Psi1=matrix(c(cos(theta),sin(theta)),ncol=2),
#' Psi2=matrix(c(cos(theta+pi/2),sin(theta+pi/2)),ncol=2))
#'
#' \dontrun{
#' profMeans<-obliqueProfiles(object = kmModel,allPsi = allPsi,threshold = threshold,
#'                            options_full = options_full,options_approx = options_approx,
#'                            uq_computations = FALSE,plot_level = 3,plot_options = NULL,
#'                            CI_const = NULL,return_level = 2)
#'
#'
#' # Approximate oblique profiles with UQ
#' plot_options<-list(save=FALSE, titleProf = "Coordinate profiles",
#'                    title2d = "Posterior mean",qq_fill=TRUE)
#' options_sims<-list(nsim=150)
#' obProfUQ<-obliqueProfiles(object=profMeans,threshold=threshold,allPsi = allPsi,
#'                            options_full=options_full, options_approx=options_approx,
#'                            uq_computations=TRUE, plot_level=3,plot_options=NULL,
#'                            CI_const=NULL,return_level=2,options_sims=options_sims)
#' }
#' @export
obliqueProfiles = function(object,allPsi,threshold,options_full=NULL,options_approx=NULL,uq_computations=FALSE,plot_level=0,plot_options=NULL,CI_const=NULL,return_level=1,...){

  # Check object
  if(is(object,"km")){
    object<-list(kmModel=object)
  }else if(!is.list(object)){
    stop("object must be either a list or a km object")
  }

  # save the dimension of the input
  d<-object$kmModel@d

  # Options setup
  if(is.null(options_full)){
    options_full<-list(multistart=8,heavyReturn=TRUE,discretization=100)
  }

  # Initialize options_approx
  if(is.null(options_approx)){
    options_approx<- list(multistart=8,heavyReturn=TRUE,fullDesignSize=100,smoother=NULL)
  }

  # initialize limits
  if(is.null(options_approx$limits)){
    ll_b = rep(0,d)
    uu_b = rep(1,d)
  }else{
    ll_b = options_approx$limits$lower
    uu_b = options_approx$limits$upper
  }
  # Useful for choosing limits for eta
  cubeVertex<-matrix(c(ll_b[1],uu_b[1]),nrow=1)
  ii=1
  while(ii<d){
    ii=ii+1
    cubeVertex <- cbind(cubeVertex,cubeVertex)
    cubeVertex <-rbind(cubeVertex, c(rep(ll_b[ii],ncol(cubeVertex)/2),c(rep(uu_b[ii],ncol(cubeVertex)/2))))
  }

  # Initialize initial design for approximation
  num_Psi = length(allPsi)
  if(is.null(options_approx$initDesign)){
    init_des<-list()
    for(i in seq(num_Psi)){
      p<-nrow(allPsi[[i]])
      # Choose limits for etas for current Psi
      mmEtas<-apply(crossprod(t(allPsi[[i]]),cubeVertex),1,min)
      MMetas<-apply(crossprod(t(allPsi[[i]]),cubeVertex),1,max)

      # Get initial design
      if(length(options_approx$initDesign)<num_Psi){
        if(p==1){
          init_des[[i]]<-matrix(seq(from=mmEtas[1],to=MMetas[1],,ceiling(sqrt(d)*10)),ncol=1)
        }else{
          ### NEEDS TEST!
          init_des[[i]]<-matrix(mmEtas,ncol=p,byrow = T,nrow=ceiling(sqrt(d*p)*10))+maximinLHS(ceiling(sqrt(d*p)*10),p)%*%diag(MMetas-mmEtas,ncol=p)
        }
      }
  #    init_des<-maximinLHS(ceiling(sqrt(object$kmModel@d)*10),object$kmModel@d)
    }
    options_approx$initDesign<-init_des
  }
  p<-nrow(allPsi[[1]])

  # save number of thresholds
  num_T<-length(threshold)

  # Set up plot options
  plot_options<-setPlotOptions(plot_options = plot_options,d=d,num_T=num_T,kmModel=object$kmModel)

  ## posterior mean part ##
  # Let us define the posterior mean function and gradient in a 'optimizer-friendly' way
  g <-function(y){
    y<-matrix(y,ncol=d)
    colnames(y)<-colnames(object$kmModel@X)
    return(predict.km(object = object$kmModel,newdata = y,type="UK",light.return = TRUE,se.compute = FALSE)$mean)
  }
  gprime <- function(y){
    y<-matrix(y,ncol=d)
    colnames(y)<-colnames(object$kmModel@X)
    return(t(gradKm_dnewdata(object = object$kmModel,newdata = y,type = "UK",light.return = TRUE,se.compute = FALSE)$mean))
  }

  # Compute profile mean full, if not already computed
  if(is.null(object$profMean_full)){
    timeIn<-get_nanotime()
    object$profMean_full<-getProfileExtrema(f = g,fprime = gprime,d = d,allPsi = allPsi,opts = options_full)
    timeMean_full<-(get_nanotime()-timeIn)*1e-9
  }else{
    if(return_level>1 && is.null(object$more$times$full))
      object$more<-list(times=list(full=NA))
  }
  #    profMean_full,profMean_approx,res_UQ
  #    times<-list(full=timeMean_full,approx=timeMean_approx)
  #    results$more<-list(abs_err=abs_err,times=times)
  # Compute profile extrema with full optim

  gc()

  # Obtain the threshold points selected by profile extrema
  if(p==1){
    changePP<-getChangePoints(threshold = threshold,allRes = object$profMean_full,Design = object$profMean_full$Design)
  }else{
    changePP<-NULL
  }

  # Plot posterior mean, ONLY for dimension==2
  if(plot_level>=2 & num_Psi==2 & d==2){

    # since dimension==2 we can plot the posterior mean
    newdata<-expand.grid(seq(ll_b[1],uu_b[1],,100),seq(ll_b[2],uu_b[2],,100))
    colnames(newdata)<-colnames(object$kmModel@X)
    pred2d<-predict.km(object$kmModel,newdata = newdata,type = "UK",light.return = TRUE,se.compute = FALSE)

    # Plot the posterior mean and visualize the actual excursion set and the regions of no-excursion according to the profile extrema functions.
    if(plot_options$save)
      cairo_pdf(filename = paste(plot_options$folderPlots,"2dpostMean_1",plot_options$id_save,".pdf",sep=""),width = 12,height = 12)
    par(mar = c(5, 5, 4, 2) + 0.1,mfrow=c(1,1))
    image(matrix(pred2d$mean,nrow = 100),col=gray.colors(20), main=plot_options$title2d,xlab = "", ylab = "", #colnames(object$kmModel@X)[1],ylab= colnames(object$kmModel@X)[2],
          cex.main=3,cex.axis=1.8,cex.lab=2.8)
    contour(matrix(pred2d$mean,nrow = 100),add=T,nlevels = 10,lwd=1.5,labcex=1.2)
    contour(matrix(pred2d$mean,nrow = 100),add=T,levels = threshold,col=plot_options$col_thresh,lwd=3,labcex=1.5)
    for(tt in seq(num_T)){
      plotOblique(changePoints = changePP$alwaysEx[[tt]][[1]],direction = allPsi[[1]],col=plot_options$col_CCPthresh_alw[tt],lwd=2.5)
      plotOblique(changePoints = changePP$alwaysEx[[tt]][[2]],direction = allPsi[[2]],col=plot_options$col_CCPthresh_alw[tt],lwd=2.5)
      plotOblique(changePoints = changePP$neverEx[[tt]][[1]],direction = allPsi[[1]],col=plot_options$col_CCPthresh_nev[tt],lwd=2.5)
      plotOblique(changePoints = changePP$neverEx[[tt]][[2]],direction = allPsi[[2]],col=plot_options$col_CCPthresh_nev[tt],lwd=2.5)
    }
    if(plot_options$fun_evals>0){
      points(object$kmModel@X,pch=17,cex=1.6)
    }
    if(plot_options$save)
      dev.off()
  }

  # Plot the profile extrema functions
  if(plot_level>0){
    if(p==1){
      if(plot_options$save)
        cairo_pdf(filename = paste(plot_options$folderPlots,"profMean_full",plot_options$id_save,".pdf",sep=""),width = 12,height = 12)
      colnames(object$profMean_full$res$min)<-plot_options$coord_names
      plotMaxMin(allRes = object$profMean_full,threshold = threshold,Design = object$profMean_full$Design)
    }else{
      if(plot_options$save)
        pdf(file = paste(plot_options$folderPlots,"profMean_full",plot_options$id_save,".pdf",sep=""),width = 18,height = 9)
      plotOneBivProfile(allRes = object$profMean_full,allPsi = allPsi,Design=object$profMean_full$Design,threshold=threshold,main_addendum="(mean, full)",xlab=expression(eta[1]),ylab=expression(eta[2]))
    }
    if(plot_options$save)
      dev.off()
  }

  # Compute profile extrema with approximation, if not already computed
  if(is.null(object$profMean_approx)){
    timeIn<-get_nanotime()
    object$profMean_approx<-approxProfileExtrema(f = g,fprime = gprime,d = d,allPsi = allPsi,opts = options_approx)
    timeMean_approx<-(get_nanotime()-timeIn)*1e-9
  }else{
    if(return_level>1 && is.null(object$more$times$approx))
      object$more$times$approx=NA
  }

  if(plot_level>0 & p==1){
    oldpar<-par()
    if(plot_options$save)
      cairo_pdf(filename = paste(plot_options$folderPlots,"profMean_comparison",plot_options$id_save,".pdf",sep=""),width = 12,height = 12)
    mfrows<-switch(d,"1"=c(1,1),"2"=c(2,1),"3"=c(2,2),"4"=c(2,2),"5"=c(3,2),"6"=c(3,2))
    par(mfrow=mfrows, mar = c(4, 5, 3, 1) + 0.1)
    for(coord in seq(num_Psi)){
      if(is.null(plot_options$ylim)){
        ylimTemp<-c(min(object$profMean_full$res$min[,coord]),max(object$profMean_full$res$max[,coord]))
      }else{
        ylimTemp<-plot_options$ylim[coord,]
      }
      # create the title of each plot
      title_string<-paste("Coordinate",plot_options$coord_names[coord])

      plot(object$profMean_full$Design[,coord],object$profMean_full$res$min[,coord],ylim=ylimTemp,type='l',main=title_string,
           xlab=expression(eta),ylab="f",lty=1,cex.main=3,cex.lab=2.5,cex.axis=1.8)
      lines(object$profMean_full$Design[,coord],object$profMean_full$res$max[,coord],lty=1)
      lines(object$profMean_approx$Design[,coord],object$profMean_approx$res$max[,coord],lty=3,col=4,lwd=2)
      points(t(object$profMean_approx$profPoints$design[[coord]]),object$profMean_approx$profPoints$e_res$max[,coord],col=4)
      lines(object$profMean_approx$Design[,coord],object$profMean_approx$res$min[,coord],lty=4,col=4,lwd=2)
      points(t(object$profMean_approx$profPoints$design[[coord]]),object$profMean_approx$profPoints$e_res$min[,coord],col=4)
      for(tt in seq(num_T)){
        abline(v=changePP$neverEx[[tt]][[coord]],col=plot_options$col_CCPthresh_nev[tt],lwd=2.5)
        abline(v=changePP$alwaysEx[[tt]][[coord]],col=plot_options$col_CCPthresh_alw[tt],lwd=2.5)
      }
      abline(h = threshold,col=plot_options$col_thresh,lwd=2)
      if(!is.null(plot_options$legend)){
        legend("bottomleft",c(as.expression(substitute(paste(P[coord]^sup,f," (full)"),list(coord=coord))),
                              as.expression(substitute(paste(P[coord]^inf,f," (full)"),list(coord=coord))),
                              as.expression(substitute(paste(P[coord]^sup,f," (1order)"),list(coord=coord))),
                              as.expression(substitute(paste(P[coord]^inf,f," (1order)"),list(coord=coord))),
                              "threshold"),
               lty=c(1,1,3,4,1),col=c(1,1,4,4,2),cex=1.8,lwd=c(2,2,2,2,2))
      }
    }
    par(mfrow=c(1,1))
    if(plot_options$save)
      dev.off()
    #  par(oldpar)
  }else{
    if(plot_options$save)
      pdf(file = paste(plot_options$folderPlots,"profMean_approx",plot_options$id_save,".pdf",sep=""),width = 18,height = 9)
    plotOneBivProfile(allRes = object$profMean_approx,allPsi = allPsi,Design=object$profMean_approx$Design,threshold=threshold,main_addendum="(mean, approx)",xlab=expression(eta[1]),ylab=expression(eta[2]))
    if(plot_options$save)
      dev.off()
  }

  # save absolute errors with approximation
  if(nrow(object$profMean_approx$res$min)==nrow(object$profMean_full$res$min)){
  abs_err_min<-max(abs(object$profMean_approx$res$min-object$profMean_full$res$min)/object$profMean_full$res$min)
  abs_err_max<-max(abs(object$profMean_approx$res$max-object$profMean_full$res$max)/object$profMean_full$res$max)
  abs_err<-c(abs_err_min,abs_err_max)
  }else{
    warning("Different discretizations for approx and full profiles, error not computed!")
    abs_err = c(NA,NA)
  }
#   abs_err = c(NA,NA)
  ## End posterior mean part ##

  ## quantiles part ##
  if(!is.null(CI_const)){
    g_s2<-function(y){
      y<-matrix(y,ncol=d)
      colnames(y)<-colnames(object$kmModel@X)
      return(predict.km(object = object$kmModel,newdata = y,type="UK",light.return = TRUE,se.compute = TRUE)$sd)
    }
    g_s2prime <- function(y){
      y<-matrix(y,ncol=d)
      colnames(y)<-colnames(object$kmModel@X)
      s2<-predict.km(object = object$kmModel,newdata = y,type="UK",light.return = TRUE,se.compute = TRUE)$sd
      s2grad<-gradKm_dnewdata(object = object$kmModel,newdata = y,type = "UK",se.compute = TRUE)$s2
      return(0.5*t(s2grad)/s2)
    }

    # Compute profile sd full, if not already computed
    if(is.null(object$profSd_full)){
      timeIn<-get_nanotime()
      object$profSd_full<-getProfileExtrema(f = g_s2,fprime = g_s2prime,d = d,allPsi = allPsi,opts = options_full)
      timeVar_full<-(get_nanotime()-timeIn)*1e-9
    }else{
      if(return_level>1 && is.null(object$more$times$full))
        object$more$times$full_var=NA
    }

    # Obtain the threshold points selected by profile extrema
    prof_CI_const_full<-list()
    temp<-list()
    changePP_CI_full<-list()
    for(i in seq(length(CI_const))){
      temp_m<-list(res=list())
      temp_m$res<-list(min= object$profMean_full$res$min - CI_const[i]*object$profSd_full$res$min,
                       max= object$profMean_full$res$max - CI_const[i]*object$profSd_full$res$max)
      temp_M<-list(res=list())
      temp_M$res<-list(min= object$profMean_full$res$min + CI_const[i]*object$profSd_full$res$min,
                       max= object$profMean_full$res$max + CI_const[i]*object$profSd_full$res$max)
      prof_CI_const_full[[i]]<-list(lower=temp_m, upper=temp_M)

      if(p==1)
        changePP_CI_full[[i]]<-list(lower=getChangePoints(threshold = threshold,allRes = temp_m,Design = object$profMean_full$Design),
                                  upper=getChangePoints(threshold = threshold,allRes = temp_M,Design = object$profMean_full$Design))
    }

    # Compute profile extrema var with approximation, if not already computed
    #    if(is.null(object$profVar_approx)){
    #      timeIn<-get_nanotime()
    #      object$profVar_approx<-approxMaxMin(f = g_s2,fprime = g_s2prime,d = d,opts = options_approx)
    #      timeVar_approx<-(get_nanotime()-timeIn)*1e-9
    #    }else{
    #      if(return_level>1 && is.null(object$more$times$approx))
    #        object$more$times$approx=NA
    #    }

    if(plot_level>0 & p==1){
      oldpar<-par()
      if(plot_options$save)
        cairo_pdf(filename = paste(plot_options$folderPlots,"profMeanCI_comparison",plot_options$id_save,".pdf",sep=""),width = 12,height = 12)
      mfrows<-switch(d,"1"=c(1,1),"2"=c(2,1),"3"=c(2,2),"4"=c(2,2),"5"=c(3,2),"6"=c(3,2))
      par(mfrow=mfrows, mar = c(4, 5, 3, 1) + 0.1)
      for(coord in seq(num_Psi)){
        if(is.null(plot_options$ylim)){
          ylimTemp_CI<-c(object$profMean_full$res$min[,coord],object$profMean_full$res$max[,coord])
          for(i in seq(length(CI_const))){
            ylimTemp_CI<-c(ylimTemp_CI,prof_CI_const_full[[i]]$lower$res$min[,coord],
                           prof_CI_const_full[[i]]$lower$res$max[,coord],
                           prof_CI_const_full[[i]]$upper$res$min[,coord],
                           prof_CI_const_full[[i]]$upper$res$max[,coord])
          }
          ylimTemp_CI<-range(ylimTemp_CI)
        }else{
          ylimTemp<-plot_options$ylim[coord,]
        }

        # create the title of each plot
        title_string<-paste("Coordinate",plot_options$coord_names[coord])
        # full mean
        plot(object$profMean_full$Design[,coord],object$profMean_full$res$min[,coord],ylim=ylimTemp_CI,type='l',main=title_string,
             xlab=expression(eta),ylab="f",lty=1,cex.main=3,cex.lab=2.5,cex.axis=1.8)
        lines(object$profMean_full$Design[,coord],object$profMean_full$res$max[,coord],lty=1)
        # approx mean
        lines(object$profMean_approx$Design[,coord],object$profMean_approx$res$max[,coord],lty=3,col=4,lwd=2)
        points(t(object$profMean_approx$profPoints$design[[coord]]),object$profMean_approx$profPoints$e_res$max[,coord],col=4)
        lines(object$profMean_approx$Design[,coord],object$profMean_approx$res$min[,coord],lty=4,col=4,lwd=2)
        points(t(object$profMean_approx$profPoints$design[[coord]]),object$profMean_approx$profPoints$e_res$min[,coord],col=4)
        for(tt in seq(num_T)){
          abline(v=changePP$neverEx[[tt]][[coord]],col=plot_options$col_CCPthresh_nev[tt],lwd=2.5)
          abline(v=changePP$alwaysEx[[tt]][[coord]],col=plot_options$col_CCPthresh_alw[tt],lwd=2.5)
        }
        # full CI_const
        CI_cols1<-heat.colors(n = length(CI_const)+1,alpha=0.7)
        CI_cols2<-terrain.colors(n = length(CI_const)+1,alpha=0.7)
        for(i in seq(length(CI_const))){
          lines(object$profMean_full$Design[,coord],prof_CI_const_full[[i]]$upper$res$max[,coord],lty=4,col=CI_cols1[i])
          lines(object$profMean_full$Design[,coord],prof_CI_const_full[[i]]$lower$res$max[,coord],lty=4,col=CI_cols1[i])
          lines(object$profMean_full$Design[,coord],prof_CI_const_full[[i]]$upper$res$min[,coord],lty=4,col=CI_cols2[i])
          lines(object$profMean_full$Design[,coord],prof_CI_const_full[[i]]$lower$res$min[,coord],lty=4,col=CI_cols2[i])
        }
        abline(h = threshold,col=plot_options$col_thresh,lwd=2)
        if(!is.null(plot_options$legend)){
          legend("bottomleft",c(as.expression(substitute(paste(P[coord]^sup,f," (full)"),list(coord=coord))),
                                as.expression(substitute(paste(P[coord]^inf,f," (full)"),list(coord=coord))),
                                as.expression(substitute(paste(P[coord]^sup,f," (1order)"),list(coord=coord))),
                                as.expression(substitute(paste(P[coord]^inf,f," (1order)"),list(coord=coord))),
                                "threshold"),
                 lty=c(1,1,3,4,1),col=c(1,1,4,4,2),cex=1.8,lwd=c(2,2,2,2,2))
        }
      }
      par(mfrow=c(1,1))
      if(plot_options$save)
        dev.off()
      #  par(oldpar)
    }else{

      for(i in seq(length(CI_const))){
        if(plot_options$save)
          pdf(file = paste(plot_options$folderPlots,"profMeanCI",CI_const[i]*100,"_lower",plot_options$id_save,".pdf",sep=""),width = 18,height = 9)
        plotOneBivProfile(allRes = prof_CI_const_full[[i]]$lower,allPsi = allPsi,Design=object$profMean_full$Design,threshold=threshold,main_addendum=paste("(CI,",CI_const[i],", lower)"),xlab=expression(eta[1]),ylab=expression(eta[2]))
        if(plot_options$save)
          dev.off()

        if(plot_options$save)
          pdf(file = paste(plot_options$folderPlots,"profMeanCI",CI_const[i]*100,"_upper",plot_options$id_save,".pdf",sep=""),width = 18,height = 9)
        plotOneBivProfile(allRes = prof_CI_const_full[[i]]$upper,allPsi = allPsi,Design=object$profMean_full$Design,threshold=threshold,main_addendum=paste("(CI,",CI_const[i],", upper)"),xlab=expression(eta[1]),ylab=expression(eta[2]))
        if(plot_options$save)
          dev.off()
      }
    }
    object$profCI_full<-prof_CI_const_full
  }


  ## UQ part ##
  #  res_UQ<-NULL
  if(uq_computations){# && is.null(object$res_UQ)){
  #  warning("UQ computations need to be checked!")
    if(is.null(object$res_UQ$kmModel))
      object$res_UQ$kmModel=object$kmModel
    object$res_UQ<-obliqueProf_UQ(object=object$res_UQ,threshold=threshold,allPsi=allPsi,allResMean=object$profMean_full,
                                options_approx=options_approx,plot_level=max(0,plot_level-1),
                                plot_options=plot_options,return_level=max(1,return_level-1),...)

  }else{
    object$res_UQ<-NULL
  }
  #results<-object =list(kmModel=object$kmModel, profMean_full=object$profMean_full,profMean_approx=object$profMean_approx,res_UQ=object$res_UQ)
  if(return_level==1){
    return(object)
  }else{
    if(is.null(object$more)){
      times<-list(full=timeMean_full,approx=timeMean_approx)
      object$more<-list(abs_err=abs_err,times=times)
    }
    return(object)
  }

}

# getPointProportion function
#' @author Dario Azzimonti
#' @title Obtain proportion of true observations in excursion set
#' @name getPointProportion
#' @description Computes the proportion of observations in the excursion set from true function evaluations,
#'  binned by the grid determined with \code{xBins}, \code{yBins}.
#' @param pp a matrix of dimension nPts x 2 with the true points locations in 2 dimensions. 
#' @param xBins numerical vector with the ordered breaks of the grid along the x axis
#' @param yBins numerical vector with the ordered breaks of the grid along the y axis
#' @param whichAbove boolean vector of dimension nPts, selects the points above
#' @param plt if not \code{TRUE} plots the grid, the points and the counts for each cell.
#' @return a list containing \code{above}, the counts of points in excursion, \code{full} the counts per cell of all points,
#'  \code{freq}, the relative frequence. 
#' @export
getPointProportion <- function(pp,xBins, yBins, whichAbove, plt=FALSE){
  
  nnR <- length(xBins)
  nnC <- length(yBins)
  
  # get bins for full pp 
  binxy <- data.frame(x=findInterval(pp[,1], xBins),
                      y=findInterval(pp[,2], yBins))
  
  init_table<- table(binxy)
  full_table<-matrix(0,ncol=(nnC-1),nrow=(nnR-1))
  full_table[sort(unique(binxy$x)),sort(unique(binxy$y))]<-init_table
  colnames(full_table) <- 1:(nnC-1)
  rownames(full_table) <- 1:(nnR-1)
  
  if(plt){
    plot(expand.grid(xBins,yBins), t="n", xaxs="i", yaxs="i",main="full")
    points(pp, col="blue", pch="+") 
    abline(v=xBins, h=yBins)
    
    ddd <- as.data.frame.table(full_table)
    names(ddd)[1:2]<-c("x","y")
    xxx <- xBins[-length(xBins)] + 0.5*diff(xBins)
    ddd$x <- xxx[ddd$x]
    yyy <- yBins[-length(yBins)] + 0.5*diff(yBins)
    ddd$y <- yyy[ddd$y]
    with(ddd, text(x, y, label=Freq))
  }
  
  # get bins for above pp 
  binxy <- data.frame(x=findInterval(pp[whichAbove,1], xBins),
                      y=findInterval(pp[whichAbove,2], yBins))
  
  init_table<- table(binxy)
  above_table<-matrix(0,ncol=(nnC-1),nrow=(nnR-1))
  above_table[sort(unique(binxy$x)),sort(unique(binxy$y))]<-init_table
  colnames(above_table) <- 1:(nnC-1)
  rownames(above_table) <- 1:(nnR-1)
  
  freqEx<-above_table/full_table
  freqEx[is.nan(freqEx)]<-rep(0,sum(is.nan(freqEx)))
  
  if(plt){
    plot(expand.grid(xBins,yBins), t="n", xaxs="i", yaxs="i",main="whichAbove")
    points(pp[whichAbove,], col="blue", pch="+") 
    abline(v=xBins, h=yBins)
    
    ddd <- as.data.frame.table(above_table)
    names(ddd)[1:2]<-c("x","y")
    xxx <- xBins[-length(xBins)] + 0.5*diff(xBins)
    ddd$x <- xxx[ddd$x]
    yyy <- yBins[-length(yBins)] + 0.5*diff(yBins)
    ddd$y <- yyy[ddd$y]
    with(ddd, text(x, y, label=Freq))
  }
  
  return(list(above=above_table,full=full_table,freq=freqEx))
  
}

## Graphical functions
# plotMaxMin function
#' @author Dario Azzimonti
#' @title Plot coordinate profiles
#' @name plotMaxMin
#' @description Plot coordinate profiles, for dimension up to 6.
#' @param allRes list containing the list \code{res} which contains the computed minima and maxima. The object returned by the function \code{getAllMaxMin}.
#' @param Design a d dimensional design corresponding to the points
#' @param threshold if not \code{NULL} plots the level
#' @param changes boolean, if not \code{FALSE} plots the points where profile extrema take values near the threshold.
#' @param trueEvals if not \code{NULL} adds to each plot the data points and the observed value
#' @param ... additional parameters to be passed to the plot function
#' @return plots the sup and inf of the function for each dimension. If threshold is not NULL
#' @export
# you can add km_model=NULL in signature and uncomment the lines() below to plot the trend.
# Still needs work on how to evaluate directly the formulas.
plotMaxMin<-function(allRes,Design=NULL,threshold=NULL,changes=FALSE,trueEvals=NULL,...){
  d<-ncol(allRes$res$min)
  if(is.null(Design)){
    if(!is.null(allRes$minima)){
      minR<-apply(allRes$minima,2,range)
      maxR<-apply(allRes$maxima,2,range)
      Design<-matrix(NA,ncol=d,nrow=100)
      for(i in seq(d)){
        Design[,i]<-seq(min(c(minR[,i],maxR[,i])),max(c(minR[,i],maxR[,i])),,100)
      }

    }else{
      warning("Design not selected, automatically chosen between 0 and 1")
      Design<-replicate(d,seq(0,1,,nrow(allRes$res$min)))
    }
  }

  if(changes){
    pp_change<-getChangePoints(threshold = threshold,Design = Design,allRes=allRes)
  }

  mfrows<-switch(d,"1"=c(1,1),"2"=c(2,1),"3"=c(2,2),"4"=c(2,2),"5"=c(3,2),"6"=c(3,2))
  oldpar<-par()
  # mar= c(5, 4, 4, 2) + 0.1
  par(mfrow=mfrows, mar = c(4, 3, 3, 1) + 0.1)
  for(coord in seq(d)){
    ylimTemp<-c(min(allRes$res$min[,coord]),max(allRes$res$max[,coord]))
    title_string<-paste("Coordinate",colnames(allRes$res$min)[coord])
    plot(Design[,coord],allRes$res$min[,coord],ylim=ylimTemp,type='l',main=title_string,
         xlab="x",ylab="f",...)
    lines(Design[,coord],allRes$res$max[,coord])

    if(!is.null(threshold))
      abline(h = threshold,col=2)
    if(changes){
      for(cc in seq(names(pp_change$alwaysEx))){
        abline(v = pp_change$alwaysEx[[cc]][[coord]],col=4)
        abline(v = pp_change$neverEx[[cc]][[coord]],col="darkgreen")
      }
    }
    if(!is.null(trueEvals))
      points(trueEvals[,coord],trueEvals[,(d+1)])
  }
  par(mfrow=c(1,1))
}


# getChangePoints
#' @author Dario Azzimonti
#' @name getChangePoints
#' @title Coordinate profiles crossing points
#' @description Obtain the points where the coordinate profile extrema functions cross the threshold
#' @param allRes list containing the list \code{res} which contains the computed minima and maxima. The object returned by the function \code{getAllMaxMin}.
#' @param Design a d dimensional design corresponding to the points
#' @param threshold if not null plots the level
#' @return returns a list containing two lists with d components where
#' \itemize{
#' \item alwaysEx: each component is a numerical vector indicating the points \eqn{x_i} where \eqn{inf_{x^{-i}}f(x) >} \code{threshold};
#' \item neverEx: each component is a numerical vector indicating the points \eqn{x_i} where \eqn{sup_{x^{-i}}f(x) <} \code{threshold}.
#' }
#' @export
getChangePoints <- function(threshold,Design=NULL,allRes){
  d<-ncol(allRes$res$min)
  if(is.null(Design))
    Design<-replicate(d,seq(0,1,,nrow(allRes$res$min)))

  # adapt the function to many thresholds
  num_t<-length(threshold)
  alwaysExcursion<-array(NA,dim = c(nrow(Design),d,num_t))
  neverExcursion<-array(NA,dim = c(nrow(Design),d,num_t))

  chPointsAlwaysEx<-replicate(n = num_t,list())
  names(chPointsAlwaysEx)<-round(threshold)
  chPointsNeverEx<-replicate(n = num_t,list())
  names(chPointsNeverEx)<-round(threshold)

  for(j in seq(num_t)){
    alwaysExcursion[,,j] <- allRes$res$min > threshold[j]
    neverExcursion[,,j] <- allRes$res$max < threshold[j]

    for(i in seq(d)){
      if(!length(Design[alwaysExcursion[,i,j],i])){
        chPointsAlwaysEx[[j]][[i]]<-min(Design[,i])
      }else{
        chPointsAlwaysEx[[j]][[i]]<-getSegments(Design[alwaysExcursion[,i,j],i])
      }

      if(!length(Design[neverExcursion[,i,j],i])){
        chPointsNeverEx[[j]][[i]]<-min(Design[,i])
      }else{
        chPointsNeverEx[[j]][[i]]<-getSegments(Design[neverExcursion[,i,j],i])
      }

    }
  }
  return(list(alwaysEx=chPointsAlwaysEx,neverEx=chPointsNeverEx))
}




# getSegments function
#' @author Dario Azzimonti
#' @name getSegments
#' @title getSegments
#' @description Auxiliary function for \code{getChangePoints}
#' @param y a vector
# @param Design a d dimensional design corresponding to the points
# @param threshold if not null plots the level
#' @return plots the sup and inf of the function for each dimension. If threshold is not NULL
#' @export
getSegments = function(y){
  edges<-c(y[1])
  minDiffy<-min(unique(diff(y)))
  allJumps<-abs(diff(y)-minDiffy)>1e-15
  edges<-c(edges,y[allJumps])
  edges<-c(edges,y[c(FALSE,allJumps)],y[length(y)])
  return(sort(edges))
}



# plotOblique function
#' @author Dario Azzimonti
#' @name plotOblique
#' @title plotOblique
#' @description Auxiliary function for 2d plotting of excluded regions
#' @param changePoints Numerical vector with the change points (usually if \code{cp=getChangePoints(...)}, then this is cc$alwaysEx[[1]][[1]] for example)
#' @param direction The Psi vector used for the direction
#' @param ... parameters to be passed to abline
#' @return adds to the current plot the lines \eqn{x} s.t. \code{direction}^T \eqn{x} = \code{changePoints[i]} for all i
#' @export
plotOblique<-function(changePoints,direction,...){

  av<- changePoints/direction[2]
  bv<- -direction[1]/direction[2]*(changePoints!=0)

  for(i in seq(length(changePoints))){
    if(direction[2]==0){
      abline(v=changePoints[i]/direction[1],...)
    }else{
      if(av[i]==0){
        next
      }else{
        abline(a=av[i],b=bv[i],...)
      }
    }

  }
}


# plotOneBivProfile function
#' @author Dario Azzimonti
#' @title Plot bivariate profiles
#' @name plotOneBivProfile
#' @description Plots the bivariate profiles stored in \code{allRes} for each Psi in \code{allPsi}.
#' @param allRes list containing the list \code{res} which contains the computed minima and maxima. The object returned by the function \code{getProfileExtrema}.
#' @param allPsi a list containing the matrices Psi (dim \eqn{2xd}) for which to compute the profile extrema
#' @param Design a matrix of dimension \eqn{(2d)x numPsi} encoding the first (\code{Design[1:d,]}) and the second ((\code{Design[(d+1):(2*d),]})) axis values.
#' @param threshold if not \code{NULL} plots the level as a contour.
#' @param trueEvals if not \code{NULL} adds to each plot the data points and the observed value
#' @param main_addendum additional string to add to image title. Default is empty string.
#' @param ... additional parameters to be passed to the plot function
#' @return plots the 2d maps of the profile sup and inf in \code{allRes} for each Psi in \code{allPsi}. If threshold is not NULL also contours the threshold level.
#' @seealso plotBivariateProfiles
#' @export
plotOneBivProfile<-function(allRes,allPsi,Design=NULL,threshold=NULL,trueEvals=NULL,main_addendum="",...){


  num_Psi<-length(allPsi)



  if(is.null(Design)){
      warning("Design not selected, auto choice not implemented")
      return(NULL)
  }

  dd_eta<-length(Design[,1])/2

  bkg_col <- gray.colors(20)

  for(i in seq(num_Psi)){

    zzBackground_max <- matrix(allRes$res$max[[i]],ncol=dd_eta)
    zzBackground_min <- matrix(allRes$res$min[[i]],ncol=dd_eta)

#    par(mfrow=c(1,2))
    layout(matrix(1:4,nrow=1),widths=c(0.45,0.05,0.45,0.05))
    par(mar=c(4,4,4,0.2) + 0.2)
    image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackground_max,col=bkg_col,
          main=bquote(P[Psi[.(i)]]^sup ~ "f " ~ .(main_addendum)),...)
    contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackground_max,nlevels = 10,add=T)
    contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackground_max,levels = threshold,add=T,col=2,lwd=1.4)

    legend_info<-hist(zzBackground_max,plot=FALSE,breaks=20)
    which_col_leg <- which(legend_info$counts>0)
    par(mar=c(5,0.2,5,2.8))
    image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
    axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg),...)#,mgp=c(0,.5,0))

    par(mar=c(4,4,4,0.2) + 0.2)
    image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackground_min,col=bkg_col,main=bquote(P[Psi[.(i)]]^inf ~ "f" ~ .(main_addendum)),...)
    contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackground_min,nlevels = 10,add=T)
    contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackground_min,levels = threshold,add=T,col=2,lwd=1.4)

    legend_info<-hist(zzBackground_min,plot=FALSE,breaks=20)
    which_col_leg <- which(legend_info$counts>0)
    par(mar=c(5,0.2,5,2.8))
    image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
    axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg),...)#,mgp=c(0,.5,0))
  }
#  par(mfrow=c(1,1))


}

# Function to plot the univariate profile extrema functions with UQ
#' @name plot_univariate_profiles_UQ
#' @description Function to plot the univariate profile extrema functions with UQ
#' @title Univariate profile extrema with UQ
#' @param objectUQ an object returned by \link{coordProf_UQ} or the object saved in \code{obj$res_UQ}, if \code{obj} is the object returned by \link{coordinateProfiles}.
#' @param plot_options a list containing the same elements as the one passed to \link{coordinateProfiles}
#' @param nsims number of simulations
#' @param threshold threshold of interest
#' @param nameFile the central name of the plot file
#' @param quantiles_uq a vector containing the quantiles to be computed
#' @param profMean the profile coordinate extrema functions for the mean. It is saved in \code{obj$profMean_full} or \code{obj$profMean_approx} if \code{obj} is an object returned by \link{coordinateProfiles}.
#' @param typeProf a string to choose with type of profile extrema for simulations to plot \itemize{
#' \item "approx": plots only the approximate profile extrema for simulations
#' \item "full": plots only the full profile extrema for simulations
#' \item "both": plots both the approximate and full profile extrema for simulations
#' }
#' @return Plots either to the default graphical device or to pdf (according to the options passed in \code{plot_options})
#' @export
plot_univariate_profiles_UQ<-function(objectUQ,plot_options,nsims,threshold,nameFile="prof_UQ",quantiles_uq=c(0.05,0.95),profMean=NULL,typeProf="approx"){

  d<-objectUQ$kmModel@d

  num_T<-length(threshold)

  plot_options<-setPlotOptions(plot_options = plot_options,d=d,num_T=num_T,kmModel = objectUQ$kmModel)

  d <- ncol(plot_options$design)

  if(plot_options$save)
    cairo_pdf(filename = paste(plot_options$folderPlots,nameFile,plot_options$id_save,".pdf",sep=""),width = 12,height = 12)
  mfrows<-switch(d,"1"=c(1,1),"2"=c(2,1),"3"=c(2,2),"4"=c(2,2),"5"=c(3,2),"6"=c(3,2))
  par(mfrow=mfrows, mar = c(4, 5, 3, 1) + 0.1)
  for(coord in seq(d)){
    # create the title of each plot
    title_string<-paste("Coordinate",plot_options$coord_names[coord])
    # determine the plot lims
    if(is.null(plot_options$ylim)){
      if(typeProf=="approx"){
        ylimTemp<-range(c(range(objectUQ$profSups[coord,,]),range(objectUQ$profInfs[coord,,])))
      }else if(typeProf=="full"){
        ylimTemp<-range(c(range(objectUQ$profSups_full[coord,,]),range(objectUQ$profInfs_full[coord,,])))
      }else{
        ylimTemp<-range(c(range(objectUQ$profSups[coord,,]),range(objectUQ$profInfs[coord,,]),
                          range(objectUQ$profSups_full[coord,,]),range(objectUQ$profInfs_full[coord,,])))
      }
      if(!is.null(objectUQ$bound$bound)){
        ylimTemp<-range(c(ylimTemp,
                          range(objectUQ$bound$bound$lower$res$min[,coord],na.rm = TRUE),
                          range(objectUQ$bound$bound$upper$res$min[,coord],na.rm = TRUE),
                          range(objectUQ$bound$bound$lower$res$max[,coord],na.rm = TRUE),
                          range(objectUQ$bound$bound$upper$res$max[,coord],na.rm = TRUE)))
      }
      if(!is.null(profMean)){
        ylimTemp<-range(c(ylimTemp,
                          range(profMean$res$min[,coord]),range(profMean$res$max[,coord])))
        #ylimTemp<-c(min(profMean$res$min[,coord]),max(profMean$res$max[,coord]))
        plot(plot_options$design[,coord],profMean$res$min[,coord],ylim=ylimTemp,type='l',main=title_string,
             xlab=expression(eta),ylab="f",lty=1,cex.main=3,cex.lab=2.5,cex.axis=1.8)
      }else{
        ylimTemp<-range(c(ylimTemp,
                          range(objectUQ$prof_quantiles_approx$`0.5`$res$min[,coord]),
                          range(objectUQ$prof_quantiles_approx$`0.5`$res$max[,coord])))
        plot(plot_options$design[,coord],objectUQ$prof_quantiles_approx$`0.5`$res$min[,coord],ylim=ylimTemp,type='l',main=title_string,
             xlab=expression(eta),ylab="f",lty=1,cex.main=3,cex.lab=2.5,cex.axis=1.8)
      }
    }else{
      ylimTemp<-plot_options$ylim[coord,]
      if(!is.null(profMean)){
        plot(plot_options$design[,coord],profMean$res$min[,coord],ylim=ylimTemp,type='l',main=title_string,
             xlab=expression(eta),ylab="f",lty=1,cex.main=3,cex.lab=2.5,cex.axis=1.8)
      }else{
        plot(plot_options$design[,coord],objectUQ$prof_quantiles_approx$`0.5`$res$min[,coord],ylim=ylimTemp,type='l',main=title_string,
             xlab=expression(eta),ylab="f",lty=1,cex.main=3,cex.lab=2.5,cex.axis=1.8)
      }
    }
    for(i in seq(nsims)){
      if(typeProf=="approx"){
 #       dd<-seq(0,1,,length.out = length(objectUQ$profInfs[coord,,i]))
        lines(plot_options$design[,coord], objectUQ$profSups[coord,,i],col=adjustcolor('darkolivegreen3',alpha.f=0.5),lwd=0.2)
        lines(plot_options$design[,coord], objectUQ$profInfs[coord,,i],col=adjustcolor('darkolivegreen3',alpha.f=0.5),lwd=0.2)
      }else if (typeProf=="full"){
 #       dd<-seq(0,1,,length.out = length(objectUQ$profInfs_full[coord,,i]))
        lines(plot_options$design[,coord], objectUQ$profSups_full[coord,,i],col=adjustcolor('darkolivegreen3',alpha.f=0.5),lwd=0.2)
        lines(plot_options$design[,coord], objectUQ$profInfs_full[coord,,i],col=adjustcolor('darkolivegreen3',alpha.f=0.5),lwd=0.2)
      }else{
  #      dd<-seq(0,1,,length.out = length(objectUQ$profInfs[coord,,i]))
        lines(plot_options$design[,coord], objectUQ$profSups[coord,,i],col=adjustcolor('darkolivegreen3',alpha.f=0.3),lwd=0.2)
        lines(plot_options$design[,coord], objectUQ$profInfs[coord,,i],col=adjustcolor('darkolivegreen3',alpha.f=0.3),lwd=0.2)
        lines(plot_options$design[,coord], objectUQ$profSups_full[coord,,i],col=adjustcolor('darkorange3',alpha.f=0.3),lwd=0.2)
        lines(plot_options$design[,coord], objectUQ$profInfs_full[coord,,i],col=adjustcolor('darkorange3',alpha.f=0.3),lwd=0.2)
      }
    }
    if(plot_options$qq_fill){
      if(typeProf=="approx"){
        polygon(c(plot_options$design[,coord], rev(plot_options$design[,coord])),
                c(objectUQ$prof_quantiles_approx[[2]]$res$min[,coord],
                  rev(objectUQ$prof_quantiles_approx[[1]]$res$min[,coord])),
                col = plot_options$qq_fill_colors$approx, border = NA)
        polygon(c(plot_options$design[,coord], rev(plot_options$design[,coord])),
                c(objectUQ$prof_quantiles_approx[[2]]$res$max[,coord],
                  rev(objectUQ$prof_quantiles_approx[[1]]$res$max[,coord])),
                col = plot_options$qq_fill_colors$approx, border = NA)
      }else{
        polygon(c(plot_options$design[,coord], rev(plot_options$design[,coord])),
                c(objectUQ$prof_quantiles_full[[2]]$res$min[,coord],
                  rev(objectUQ$prof_quantiles_full[[1]]$res$min[,coord])),
                col = plot_options$qq_fill_colors$approx, border = NA)
        polygon(c(plot_options$design[,coord], rev(plot_options$design[,coord])),
                c(objectUQ$prof_quantiles_full[[2]]$res$max[,coord],
                  rev(objectUQ$prof_quantiles_full[[1]]$res$max[,coord])),
                col = plot_options$qq_fill_colors$approx, border = NA)
      }
    }
    #  lines(plot_options$design[,coord],allRes_approx$res$max[,coord],lty=3,col=4,lwd=2)
    #  points(allRes_approx$profPoints$design[,coord],allRes_approx$profPoints$res$max[,coord],col=4)
    #  lines(plot_options$design[,coord],allRes_approx$res$min[,coord],lty=4,col=4,lwd=2)
    #  points(allRes_approx$profPoints$design[,coord],allRes_approx$profPoints$res$min[,coord],col=4)
    ## @@@@
    ### NB changePP to be implemented!!!!!!!
    if(!is.null(profMean)){
      changePP<-getChangePoints(threshold = threshold,allRes = profMean,Design = profMean$Design)
    }

    if(typeProf=="approx" || typeProf=="both"){
    ccPP_approx<-list()
    for(j in seq(length(quantiles_uq))){
      ccPP_approx[[j]]<-getChangePoints(threshold = threshold,allRes = objectUQ$prof_quantiles_approx[[j]],objectUQ$Design_approx)
    }
    names(ccPP_approx)<-quantiles_uq
    }

    if(typeProf=="full" || typeProf=="both"){
      ccPP_full<-list()
      for(j in seq(length(quantiles_uq))){
        ccPP_full[[j]]<-getChangePoints(threshold = threshold,allRes = objectUQ$prof_quantiles_full[[j]],objectUQ$Design_full)
      }
      names(ccPP_full)<-quantiles_uq
    }

    ## @@@
    abline(h = threshold,col=plot_options$col_thresh,lwd=2)
    if(typeProf=="approx"){
      for(j in seq(length(quantiles_uq))){
        lines(plot_options$design[,coord],objectUQ$prof_quantiles_approx[[j]]$res$max[,coord],lty=j+1,col=1,lwd=1.5)
        lines(plot_options$design[,coord],objectUQ$prof_quantiles_approx[[j]]$res$min[,coord],lty=j+1,col=1,lwd=1.5)
        for(tt in seq(num_T)){
          abline(v=ccPP_approx[[j]]$alwaysEx[[tt]][[coord]],col=plot_options$col_CCPthresh_alw[tt],lwd=2,lty=2)
          abline(v=ccPP_approx[[j]]$neverEx[[tt]][[coord]],col=plot_options$col_CCPthresh_nev[tt],lwd=2,lty=3)
        }
      }
      for(i in seq(num_T)){
        if(!is.null(profMean)){
          abline(v=changePP$neverEx[[i]][[coord]],col=plot_options$col_CCPthresh_nev[i],lwd=2.5)
          abline(v=changePP$alwaysEx[[i]][[coord]],col=plot_options$col_CCPthresh_alw[i],lwd=2.5)
        }else{
          abline(v=ccPP_approx[[j]]$alwaysEx[[i]][[coord]],col=plot_options$col_CCPthresh_alw[i],lwd=2.5,lty=1)
          abline(v=ccPP_approx[[j]]$neverEx[[i]][[coord]],col=plot_options$col_CCPthresh_nev[i],lwd=2.5,lty=1)
        }
      }
    }else if(typeProf=="full"){
      if(!is.null(objectUQ$prof_quantiles_full)){
        for(j in seq(length(quantiles_uq))){
          lines(plot_options$design[,coord],objectUQ$prof_quantiles_full[[j]]$res$max[,coord],lty=j+1,col=1,lwd=1.5)
          lines(plot_options$design[,coord],objectUQ$prof_quantiles_full[[j]]$res$min[,coord],lty=j+1,col=1,lwd=1.5)
          for(tt in seq(num_T)){
            abline(v=ccPP_full[[j]]$alwaysEx[[tt]][[coord]],col=plot_options$col_CCPthresh_alw[tt],lwd=2,lty=2)
            abline(v=ccPP_full[[j]]$neverEx[[tt]][[coord]],col=plot_options$col_CCPthresh_nev[tt],lwd=2,lty=3)
          }
        }
        for(i in seq(num_T)){
          if(!is.null(profMean)){
            abline(v=changePP$neverEx[[i]][[coord]],col=plot_options$col_CCPthresh_nev[i],lwd=2.5)
            abline(v=changePP$alwaysEx[[i]][[coord]],col=plot_options$col_CCPthresh_alw[i],lwd=2.5)
          }else{
            abline(v=ccPP_full[[j]]$alwaysEx[[i]][[coord]],col=plot_options$col_CCPthresh_alw[i],lwd=2.5)
            abline(v=ccPP_full[[j]]$neverEx[[i]][[coord]],col=plot_options$col_CCPthresh_nev[i],lwd=2.5)
          }
        }
      }
    }else{
      for(j in seq(length(quantiles_uq))){
        lines(plot_options$design[,coord],objectUQ$prof_quantiles_approx[[j]]$res$max[,coord],lty=j+1,col=1,lwd=1.5)
        lines(plot_options$design[,coord],objectUQ$prof_quantiles_approx[[j]]$res$min[,coord],lty=j+1,col=1,lwd=1.5)
        lines(plot_options$design[,coord],objectUQ$prof_quantiles_full[[j]]$res$max[,coord],lty=j+1,col=adjustcolor('brown3',alpha.f=0.5),lwd=1.5)
        lines(plot_options$design[,coord],objectUQ$prof_quantiles_full[[j]]$res$min[,coord],lty=j+1,col=adjustcolor('brown3',alpha.f=0.5),lwd=1.5)
        for(tt in seq(num_T)){
          abline(v=ccPP_approx[[j]]$alwaysEx[[tt]][[coord]],col=4,lwd=2,lty=2)
          abline(v=ccPP_approx[[j]]$neverEx[[tt]][[coord]],col=4,lwd=2,lty=3)
          abline(v=ccPP_full[[j]]$alwaysEx[[tt]][[coord]],col=adjustcolor('brown4',alpha.f=0.5),lwd=2,lty=3)
          abline(v=ccPP_full[[j]]$neverEx[[tt]][[coord]],col=adjustcolor('brown3',alpha.f=0.5),lwd=2,lty=3)
        }
      }
      for(i in seq(num_T)){
        if(!is.null(profMean)){
          abline(v=changePP$neverEx[[i]][[coord]],col=plot_options$col_CCPthresh_nev[i],lwd=2.5)
          abline(v=changePP$alwaysEx[[i]][[coord]],col=plot_options$col_CCPthresh_alw[i],lwd=2.5)
        }else{
          abline(v=ccPP_full[[j]]$alwaysEx[[i]][[coord]],col=plot_options$col_CCPthresh_alw[i],lwd=2.5)
          abline(v=ccPP_full[[j]]$neverEx[[i]][[coord]],col=plot_options$col_CCPthresh_nev[i],lwd=2.5)
        }
      }
    }
    if(!is.null(profMean)){
      lines(plot_options$design[,coord],profMean$res$max[,coord],lty=1)
      lines(plot_options$design[,coord],profMean$res$min[,coord],lty=1)
    }else{
      lines(plot_options$design[,coord],objectUQ$prof_quantiles_approx$`0.5`$res$max[,coord],lty=1)
      lines(plot_options$design[,coord],objectUQ$prof_quantiles_approx$`0.5`$res$min[,coord],lty=1)
    }
    if(plot_options$fun_evals>1){
      points(objectUQ$kmModel@X[,coord],objectUQ$kmModel@y,pch=17)
    }
    # bound
    if(!is.null(objectUQ$bound$bound)){
      lines(plot_options$design[,coord],objectUQ$bound$bound$lower$res$min[,coord],col=plot_options$bound_cols[1],lty=2,lwd=1.5)
      lines(plot_options$design[,coord],objectUQ$bound$bound$lower$res$max[,coord],col=plot_options$bound_cols[2],lty=2,lwd=1.5)
      lines(plot_options$design[,coord],objectUQ$bound$bound$upper$res$min[,coord],col=plot_options$bound_cols[1],lty=3,lwd=1.5)
      lines(plot_options$design[,coord],objectUQ$bound$bound$upper$res$max[,coord],col=plot_options$bound_cols[2],lty=3,lwd=1.5)

      if(plot_options$qq_fill){
        polygon(c(plot_options$design[,coord], rev(plot_options$design[,coord])),
                c(objectUQ$bound$bound$upper$res$min[,coord],
                  rev(objectUQ$bound$bound$lower$res$min[,coord])),
                col = plot_options$qq_fill_colors$bound_min, border = NA)
        polygon(c(plot_options$design[,coord], rev(plot_options$design[,coord])),
                c(objectUQ$bound$bound$upper$res$max[,coord],
                  rev(objectUQ$bound$bound$lower$res$max[,coord])),
                col = plot_options$qq_fill_colors$bound_max, border = NA)
      }else{
        lines(plot_options$design[,coord],objectUQ$bound$approx$lower$res$min[,coord],col=5,lty=2,lwd=1.2)
        lines(plot_options$design[,coord],objectUQ$bound$approx$lower$res$max[,coord],col=5,lty=2,lwd=1.2)
        lines(plot_options$design[,coord],objectUQ$bound$approx$upper$res$min[,coord],col=5,lty=3,lwd=1.2)
        lines(plot_options$design[,coord],objectUQ$bound$approx$upper$res$max[,coord],col=5,lty=3,lwd=1.2)
      }
    }
    if(!is.null(plot_options$legend)){
      legend("bottomleft",c(as.expression(substitute(paste(P[coord]^sup,f,"/",P[coord]^inf,f," (full)"),list(coord=coord))),
                            as.expression(substitute(paste(P[coord]^sup,f[qq]),list(coord=coord,qq=quantiles_uq[1]))),
                            as.expression(substitute(paste(P[coord]^inf,f[qq]),list(coord=coord,qq=quantiles_uq[1]))),
                            as.expression(substitute(paste(P[coord]^sup,f[qq]),list(coord=coord,qq=quantiles_uq[2]))),
                            as.expression(substitute(paste(P[coord]^inf,f[qq]),list(coord=coord,qq=quantiles_uq[2]))),
                            "threshold"),
             lty=c(1,2,2,3,3,1),col=c(1,1,1,1,1,2),cex=1.2,lwd=c(2,2,2,2,2,2))
    }
  }
  par(mfrow=c(1,1))
  if(plot_options$save)
    dev.off()

}


#' @name setPlotOptions
#' @title Set-up the plot options when NULL
#' @description  Function to set-up plot options for \link{plot_univariate_profiles_UQ}, \link{plotBivariateProfiles}, \link{coordinateProfiles}, \link{coordProf_UQ}, \link{obliqueProfiles} and \link{obliqueProf_UQ}.
#' @author Dario Azzimonti
#' @param plot_options the list of plot options to set-up
#' @param d number of coordinates
#' @param num_T number of thresholds of interest
#' @param kmModel a \link[DiceKriging]{km} model, used to obtain the coordinates names.
#' @return the properly set-up list containing the following fields \itemize{
#' \item \code{save:} boolean, if TRUE saves the plots in \code{folderPlots}
#' \item \code{folderPlots:} a string containing the destination folder for plots, if \code{save==TRUE} default is \code{./}
#' \item \code{ylim:} a matrix \code{coord}x2 containing the ylim for each coordinate, if NULL in \code{plot_options} this is left NULL and automatically set at the plot time.
#' \item \code{titleProf:} a string containing the title for the coordinate profile plots, default is \code{"Coordinate profiles"}
#' \item \code{title2d:} a string containing the title for the 2d plots (if the input is 2d), default is \code{"Posterior mean"}
#' \item \code{design:} a \eqn{dxr} matrix where \eqn{d} is the input dimension and \eqn{r} is the size of the discretization for plots at each dimension
#' \item \code{coord_names:} a \eqn{d}-vector of characters naming the dimensions. If NULL and \code{kmModel} not NULL then it is the names of \code{kmModel@X} otherwise \code{x_1,...,x_d}
#' \item \code{id_save:} a string to be added to the plot file names, useful for serial computations on HPC, left as in \code{plot_options}.
#' \item \code{qq_fill:} if TRUE it fills the region between the first 2 quantiles in \code{quantiles_uq} and between the upper and lower bound in \code{objectUQ$bound$bound}, if \code{NULL}, it is set as \code{FALSE}.
#' \item \code{bound_cols:} a vector of two strings containing the names of the colors for upper and lower bound plots.
#' \item \code{qq_fill_colors:} a list containing the colors for qq_fill: \code{approx} for 2 quantiles, \code{bound_min} for bounds on the profile inf, \code{bound_max} for profile sup. Initialized only if \code{qq_fill==TRUE}.
#' \item \code{col_CCPthresh_nev:} Color palette of dimension \code{num_T} for the colors of the vertical lines delimiting the intersections between the profiles sup and the thresholds
#' \item \code{col_CCPthresh_alw:} Color palette of dimension \code{num_T} for the colors of the vertical lines delimiting the intersections between the profiles inf and the thresholds
#' \item \code{col_thresh:} Color palette of dimension \code{num_T} for the colors of the thresholds
#' \item \code{fun_evals:} integer denoting the level of plot for the true evaluations. \itemize{
#' \item 0: default, no plots for true evaluations;
#' \item 1: plot the true evaluations as points in 2d plots, no true evaluation plots in 1d;
#' \item 2: plot true evaluations, in 2d with different color for values above threshold;
#' \item 3: plot true evaluations, in 2d plots in color, with background of the image colored as proportion of points inside excursion; }
#' }
#' if all the fields are already filled then returns \code{plot_options}
#' @export
setPlotOptions<-function(plot_options=NULL,d,num_T,kmModel=NULL){

  # Start with saving, folder options, titles
  full_plot_options <-list(save=F, folderPlots = './', id_save="",
                           titleProf="Profile extrema", title2d="Posterior mean")



  # Design plot
  full_plot_options$design<-matrix(NA,ncol=d,nrow=100)
  for(i in seq(d)){
    full_plot_options$design[,i]<-seq(0,1,,100)
  }


  # coordinate names
  if(is.null(kmModel)){
    full_plot_options$coord_names<-apply(matrix(1:d,ncol=1),1,function(x) paste("x_",x,sep=''))
  }else{
    full_plot_options$coord_names<-colnames(kmModel@X)
  }

  ## Colors set-up

  # AlwaysEx colors
  full_plot_options$col_CCPthresh_alw<-adjustcolor(RColorBrewer::brewer.pal(n=max(num_T+1,3),name='Purples'),offset = c(-0.25, -0.25, -0.25, 0))
  full_plot_options$col_CCPthresh_alw<-full_plot_options$col_CCPthresh_alw[(2:length(full_plot_options$col_CCPthresh_alw))]
  if(num_T==1)
    full_plot_options$col_CCPthresh_alw[2]

  # NeverEx colors
  full_plot_options$col_CCPthresh_nev<-adjustcolor(RColorBrewer::brewer.pal(n=max(num_T+1,3),name='Greens'),offset = c(-0.25, -0.25, -0.25, 0))
  full_plot_options$col_CCPthresh_nev<-full_plot_options$col_CCPthresh_nev[(2:length(full_plot_options$col_CCPthresh_nev))]
  if(num_T==1)
    full_plot_options$col_CCPthresh_nev[2]

  # threshold colors
  full_plot_options$col_thresh<-adjustcolor(RColorBrewer::brewer.pal(n=max(num_T+1,3),name='Reds'),offset = c(-0.3, -0.3, -0.3, 0))
  full_plot_options$col_thresh<-full_plot_options$col_thresh[(2:length(full_plot_options$col_thresh))]
  if(num_T==1)
    full_plot_options$col_thresh[2]

  # bound colors
  # first one is min, second one is max
  full_plot_options$bound_cols<-c("darkseagreen4","deepskyblue4")



  # plot function evaluations
  full_plot_options$fun_evals <- 0

  # qq_fill
  full_plot_options$qq_fill <- FALSE
  # qq_fill colors
  full_plot_options$qq_fill_colors<-list(
    approx=adjustcolor("red",alpha.f=0.5),
    bound_max=adjustcolor(full_plot_options$bound_cols[2],alpha.f=0.2),
    bound_min=adjustcolor(full_plot_options$bound_cols[1],alpha.f=0.2) )

  if(is.null(plot_options)){
    plot_options<-full_plot_options
  }else{
    plot_options <- modifyList(full_plot_options,plot_options)
  }
  return(plot_options)
}

# mean_Delta_T function
#' @author Dario Azzimonti
#' @name mean_Delta_T
#' @title mean function of difference process
#' @description The function mean_Delta_T computes the mean function of the difference process \eqn{Z_x - \widetilde{Z}_x} at \code{x}.
#' @param x a matrix \eqn{r x d} containing the \eqn{r} points where the function is to be computed.
#' @param kmModel the \link[DiceKriging]{km} model of the Gaussian process \eqn{Z}.
#' @param simupoints the matrix \eqn{l x d} containing the pilot points \eqn{G}.
#' @param T.mat the upper triangular factor of the Choleski decomposition of the covariance matrix of \code{rbind(kmModel@X,simupoints)}
#' @param F.mat the evaluation of the trend function at \code{rbind(kmModel@X,simupoints)}, see \link[stats]{model.matrix}.
#' @return the value of the mean function at \code{x} for the difference process \eqn{Z^\Delta = Z_x - \widetilde{Z}_x}.
mean_Delta_T<-function(x,kmModel,simupoints,T.mat,F.mat){
  x<-matrix(x,ncol=kmModel@d)
  colnames(x)<-colnames(kmModel@X)
  obj <- krig_weight_GPsimu(object=kmModel,simu_points=simupoints,krig_points=x,T.mat = T.mat,F.mat = F.mat)
  mu_n<-predict.km(object = kmModel,newdata = x,type = "UK",se.compute = FALSE,light.return = TRUE)$mean
  mu_E<-predict.km(object = kmModel,newdata = simupoints,type = "UK",se.compute = FALSE,light.return = TRUE)$mean
  return(mu_n-obj$krig.mean.init-crossprod(obj$Lambda.end,mu_E))
  #  return(-tcrossprod(weights,predict.km(object = kmModel,newdata = simupoints,type = "UK",se.compute = FALSE,cov.compute = FALSE,light.return = FALSE)$mean))
}

# grad_mean_Delta_T function
#' @author Dario Azzimonti
#' @name grad_mean_Delta_T
#' @title Gradient of the mean function of difference process
#' @description The function grad_mean_Delta_T computes the gradient for the mean function of the difference process \eqn{Z_x - \widetilde{Z}_x} at \code{x}.
#' @param x a matrix \eqn{r x d} containing the \eqn{r} points where the function is to be computed.
#' @param kmModel the \link[DiceKriging]{km} model of the Gaussian process \eqn{Z}.
#' @param simupoints the matrix \eqn{l x d} containing the pilot points \eqn{G}.
#' @param T.mat the upper triangular factor of the Choleski decomposition of the covariance matrix of \code{rbind(kmModel@X,simupoints)}
#' @param F.mat the evaluation of the trend function at \code{rbind(kmModel@X,simupoints)}, see \link[stats]{model.matrix}.
#' @return the value of the gradient for the mean function at \code{x} for the difference process \eqn{Z^\Delta = Z_x - \widetilde{Z}_x}.
grad_mean_Delta_T<-function(x,kmModel,simupoints,T.mat,F.mat){
  x<-matrix(x,ncol=kmModel@d)
  colnames(x)<-colnames(kmModel@X)
  obj_deriv<-grad_kweights(object = kmModel,simu_points = simupoints,krig_points = matrix(x,ncol=kmModel@d),T.mat = T.mat,F.mat = F.mat)
  krig_mean_init <- matrix(obj_deriv$krig.mean.init,ncol=kmModel@d)
  weights <- t(obj_deriv$Lambda.end)

  krig_mean_init <- matrix(obj_deriv$krig.mean.init,ncol=kmModel@d)
  weights <- t(obj_deriv$Lambda.end)

  rr<-gradKm_dnewdata(object = kmModel,newdata=x,type = "UK",se.compute = F)$mean-krig_mean_init - tcrossprod(predict.km(object = kmModel,newdata = simupoints,type = "UK",se.compute = FALSE,cov.compute = FALSE,light.return = FALSE)$mean,weights)
  return(rr) # return(- tcrossprod(predict.km(object = kmModel,newdata = simupoints,type = "UK",se.compute = FALSE,cov.compute = FALSE,light.return = FALSE)$mean,weights))
}


# var_Delta_T function
#' @author Dario Azzimonti
#' @name var_Delta_T
#' @title Variance function of difference process
#' @description The function var_Delta_T computes the gradient for the variance function of the difference process \eqn{Z_x - \widetilde{Z}_x} at \code{x}.
#' @param x a matrix \eqn{r x d} containing the \eqn{r} points where the function is to be computed.
#' @param kmModel the \link[DiceKriging]{km} model of the Gaussian process \eqn{Z}.
#' @param simupoints the matrix \eqn{l x d} containing the pilot points \eqn{G}.
#' @param T.mat the upper triangular factor of the Choleski decomposition of the covariance matrix of \code{rbind(kmModel@X,simupoints)}
#' @param F.mat the evaluation of the trend function at \code{rbind(kmModel@X,simupoints)}, see \link[stats]{model.matrix}.
#' @return the value of the variance function at \code{x} for the difference process \eqn{Z^\Delta = Z_x - \widetilde{Z}_x}.
var_Delta_T<-function(x,kmModel,simupoints,T.mat,F.mat){
  x<-matrix(x,ncol=kmModel@d)
  colnames(x)<-colnames(kmModel@X)
  obj <- krig_weight_GPsimu(object=kmModel,simu_points=simupoints,krig_points=x,T.mat = T.mat,F.mat = F.mat)
  #  krig.mean.init <- matrix(obj$krig.mean.init,nrow=1)
  weights <- t(obj$Lambda.end)
  var_post<-predict.km(object = kmModel,newdata = x,type = "UK",se.compute =TRUE)$sd^2
  kEE<-predict.km(object = kmModel,newdata = simupoints,type = "UK",cov.compute = TRUE,light.return = TRUE)$cov #covMatrix(object = kmModel@covariance,X = rbind(kmModel@X,simupoints))$C
  return(list(vv=var_post, vv_E= -weights%*%tcrossprod(kEE,weights)))
}


# grad_var_Delta_T function
#' @author Dario Azzimonti
#' @name grad_var_Delta_T
#' @title Gradient of the variance function of difference process
#' @description The function grad_var_Delta_T computes the gradient for the variance function of the difference process \eqn{Z_x - \widetilde{Z}_x} at \code{x}.
#' @param x a matrix \eqn{r x d} containing the \eqn{r} points where the function is to be computed.
#' @param kmModel the \link[DiceKriging]{km} model of the Gaussian process \eqn{Z}.
#' @param simupoints the matrix \eqn{l x d} containing the pilot points \eqn{G}.
#' @param T.mat the upper triangular factor of the Choleski decomposition of the covariance matrix of \code{rbind(kmModel@X,simupoints)}
#' @param F.mat the evaluation of the trend function at \code{rbind(kmModel@X,simupoints)}, see \link[stats]{model.matrix}.
#' @return the value of the gradient for the variance function at \code{x} for the difference process \eqn{Z^\Delta = Z_x - \widetilde{Z}_x}.
grad_var_Delta_T<-function(x,kmModel,simupoints,T.mat,F.mat){
  x<-matrix(x,ncol=kmModel@d)
  colnames(x)<-colnames(kmModel@X)
  obj<-krig_weight_GPsimu(object=kmModel,simu_points=simupoints,krig_points=x,T.mat = T.mat,F.mat = F.mat)
  obj_deriv<-grad_kweights(object = kmModel,simu_points = simupoints,krig_points = matrix(x,ncol=kmModel@d),T.mat = T.mat,F.mat = F.mat)
  #  krig_mean_init <- matrix(obj_deriv$krig.mean.init,ncol=kmModel@d)
  d_weights <- t(obj_deriv$Lambda.end)
  weights<-t(obj$Lambda.end)
  ggVar<-gradKm_dnewdata(object = kmModel,newdata = x,type = "UK",se.compute = TRUE)$s2

  kEE<-predict.km(object = kmModel,newdata = simupoints,type = "UK",cov.compute = TRUE,light.return = TRUE)$cov

  return(list(ggV=t(ggVar), ggRest= - 2*d_weights%*%tcrossprod(kEE,weights)))
}

# prof_mean_var_Delta function
#' @author Dario Azzimonti
#' @name prof_mean_var_Delta
#' @title Profile extrema for the mean and variance functions of difference process
#' @description The function prof_mean_var_Delta computes the profile extrema functions for the mean and variance functions of the difference process \eqn{Z_x - \widetilde{Z}_x} at \code{x}.
#' @param kmModel the \link[DiceKriging]{km} model of the Gaussian process \eqn{Z}.
#' @param simupoints the matrix \eqn{l x d} containing the pilot points \eqn{G}.
#' @param allPsi optional list of matrices (dim \eqn{pxd}) for which to compute the profile extrema. If NULL coordinate profiles are computed.
#' @param options_approx an optional list of options for \link{approxMaxMin} (or \link{approxProfileExtrema} if \code{allPsi} not NULL).
#' @param options_full_sims an optional list of options for \link{getAllMaxMin}(or \link{approxProfileExtrema} if \code{allPsi} not NULL). If NULL the full computations are not excuted. NOTE: this computations might be very expensive!
#' @param T.mat the upper triangular factor of the Choleski decomposition of the covariance matrix of \code{rbind(kmModel@X,simupoints)}, if \code{NULL} it is computed.
#' @param F.mat the evaluation of the trend function at \code{rbind(kmModel@X,simupoints)}, see \link[stats]{model.matrix}, if \code{NULL} it is computed.
#' @return the profile extrema functions at \code{options_approx$design} for the mean and variance function of the difference process \eqn{Z^\Delta = Z_x - \widetilde{Z}_x}.
#' @export
prof_mean_var_Delta<-function(kmModel,simupoints,allPsi=NULL,options_full_sims=NULL,options_approx=NULL,F.mat=NULL,T.mat=NULL){
  if(is.null(F.mat))
    F.mat <- model.matrix(object=kmModel@trend.formula, data = data.frame(rbind(kmModel@X,simupoints)))
  if(is.null(T.mat)){
    K <- covMatrix(object=kmModel@covariance,X=rbind(kmModel@X,simupoints))$C
    T.mat <- chol(K)
  }

  ## Set up functions for optimization
  # mean Delta
  #  g_mean_spec<-function(x){
  #    return(mean_Delta_T(x=x,kmModel=kmModel,simupoints=simupoints,T.mat = T.mat,F.mat = F.mat))
  #  }

  # grad mean Delta
  #  g_mean_der_spec<-function(x){
  #    return(grad_mean_Delta_T(x=x,kmModel=kmModel,simupoints=simupoints,T.mat = T.mat,F.mat = F.mat))
  #  }

  # var Delta
  g_var_spec<-function(x){
    temp<-var_Delta_T(x=x,kmModel=kmModel,simupoints=simupoints,T.mat = T.mat,F.mat = F.mat)
    return(temp$vv+temp$vv_E)
  }

  if(is.null(allPsi)){
    g_var_der_spec<-function(x){
      temp<-grad_var_Delta_T(x=x,kmModel=kmModel,simupoints=simupoints,T.mat = T.mat,F.mat = F.mat)
      return(matrix(temp$ggV+temp$ggRest,ncol=kmModel@d))
    }
  }else{
    g_var_der_spec<-function(x){
      temp<-grad_var_Delta_T(x=x,kmModel=kmModel,simupoints=simupoints,T.mat = T.mat,F.mat = F.mat)
      return(t(matrix(temp$ggV+temp$ggRest,ncol=kmModel@d)))
    }
  }

  d<-kmModel@d


  if(!is.null(options_full_sims)){
    #  timeIn<-get_nanotime()
    #  mean_delta<-getAllMaxMin(f = g_mean_spec,fprime = g_mean_der_spec,d = d,options = options_full_sims)
    #  tFull_mean<-(get_nanotime()-timeIn)*1e-9


    if(is.null(allPsi)){
      timeIn<-get_nanotime()
      var_delta<-getAllMaxMin(f = g_var_spec,fprime = g_var_der_spec,d = d,options = options_full_sims)
      time_var<-(get_nanotime()-timeIn)*1e-9
    }else{
      timeIn<-get_nanotime()
      var_delta<-getProfileExtrema(f = g_var_spec,fprime = g_var_der_spec,allPsi = allPsi,opts = options_full_sims,d = d)
      time_var<-(get_nanotime()-timeIn)*1e-9
    }

    #    times<-c(tFull_mean,tFull_var)

  }else{
    #   timeIn<-get_nanotime()
    #   mean_delta<-approxMaxMin(f = g_mean_spec,fprime = g_mean_der_spec,d = d,opts = options_approx)
    #   tApprox_mean<-(get_nanotime()-timeIn)*1e-9

    if(is.null(allPsi)){
      timeIn<-get_nanotime()
      var_delta<-approxMaxMin(f = g_var_spec,fprime = g_var_der_spec,d = d,opts = options_approx)
      time_var<-(get_nanotime()-timeIn)*1e-9
    }else{
      timeIn<-get_nanotime()
      var_delta<-approxProfileExtrema(f = g_var_spec,fprime = g_var_der_spec,d = d,opts = options_approx,allPsi = allPsi)
      time_var<-(get_nanotime()-timeIn)*1e-9
    }
  }

  # mean_delta is zero
  mean_delta<-list(res=list(min=matrix(0,ncol=d,nrow=nrow(var_delta$res$min)),
                            max=matrix(0,ncol=d,nrow=nrow(var_delta$res$min))))


  return(list(mean=mean_delta,var=var_delta,times=time_var))
}


# bound_profiles function
#' @author Dario Azzimonti
#' @name bound_profiles
#' @title Bound for profile extrema quantiles
#' @description The function bound_profiles computes the upper and lower bounds for the profile extrema quantiles of a Gaussian process model.
#' @param objectUQ an object returned by \link{coordProf_UQ} or the object saved in \code{obj$res_UQ}, if \code{obj} is the object returned by \link{coordinateProfiles}
#' @param mean_var_delta  the profile extrema functions at \code{options_approx$design} for the mean and variance function of the difference process \eqn{Z^\Delta = Z_x - \widetilde{Z}_x}. Object returned by \link{prof_mean_var_Delta}.
#' @param beta the level of confidence for the approximate simulations
#' @param alpha the level of confidence for the bound
#' @param allPsi optional list of matrices (dim \eqn{pxd}) for which to compute the profile extrema. If NULL coordinate profiles are computed.
#' @param options_approx an optional list of options for \link{approxMaxMin} (or \link{approxProfileExtrema} if \code{allPsi} not NULL).
#' @param options_full_sims an optional list of options for \link{getAllMaxMin} (or \link{getProfileExtrema} if \code{allPsi} not NULL). If NULL the full computations are not excuted. NOTE: this computations might be very expensive!
#' @return a list containing \itemize{
#' \item \code{bound: } a list containing the upper/lower bound for profile sup and inf
#' \item \code{approx: } a list containing the upper/lower approximate quantiles for profile sup and inf
#' }
#' @export
bound_profiles<-function(objectUQ,mean_var_delta=NULL,beta=0.0124,alpha=0.025,allPsi=NULL,options_approx=NULL,options_full_sims=NULL){

  if(is.null(mean_var_delta)){
    mean_var_delta<-prof_mean_var_Delta(kmModel=objectUQ$kmModel,simupoints = objectUQ$sPts$par,allPsi=allPsi,options_full_sims=options_full_sims,options_approx=options_approx,F.mat=NULL,T.mat=NULL)
  }

  d<-objectUQ$kmModel@d
  if(!is.null(allPsi)){
    d <- length(allPsi)
    p <- nrow(allPsi[[1]])
  }else{
    p <- 1
  }

  # get (empirical) quantiles from approximations
  approx_quant<-list(lower=list(res=list(min=matrix(NA,nrow = options_approx$fullDesignSize^p,ncol = d),
                                         max=matrix(NA,nrow = options_approx$fullDesignSize^p,ncol = d))),
                     upper=list(res=list(min=matrix(NA,nrow = options_approx$fullDesignSize^p,ncol = d),
                                         max=matrix(NA,nrow = options_approx$fullDesignSize^p,ncol = d))) )

  bound_quant<-list(lower=list(res=list(min=matrix(NA,nrow = options_approx$fullDesignSize^p,ncol = d),
                                        max=matrix(NA,nrow = options_approx$fullDesignSize^p,ncol = d))),
                    upper=list(res=list(min=matrix(NA,nrow = options_approx$fullDesignSize^p,ncol = d),
                                        max=matrix(NA,nrow = options_approx$fullDesignSize^p,ncol = d))) )

  correction_u<-log(2/(alpha-beta))
  correction_l<-log(2/(alpha-2*beta))

  # In the loop below mean_var_delta$mean is removed because we always assume that the mean is zero.
  for(coord in seq(d)){
    approx_quant$lower$res$max[,coord]<-apply(objectUQ$profSups[coord,,],1,function(x){return(quantile(x,beta))})
    bound_quant$lower$res$max[,coord]<-approx_quant$lower$res$max[,coord]-sqrt(2*pmax(mean_var_delta$var$res$max[,coord],rep(0,nrow(mean_var_delta$var$res$min)))*correction_l)
    #bound_quant$lower$res$max[,coord]<-approx_quant$lower$res$max[,coord]+mean_var_delta$mean$res$max[,coord]-sqrt(2*max(mean_var_delta$var$res$max[,coord],0)*correction)

    approx_quant$lower$res$min[,coord]<-apply(objectUQ$profInfs[coord,,],1,function(x){return(quantile(x,beta))})
    bound_quant$lower$res$min[,coord]<-approx_quant$lower$res$min[,coord]-sqrt(2*pmax(mean_var_delta$var$res$min[,coord],rep(0,nrow(mean_var_delta$var$res$min)))*correction_l)
    #bound_quant$lower$res$min[,coord]<-approx_quant$lower$res$min[,coord]+mean_var_delta$mean$res$min[,coord]-sqrt(2*max(mean_var_delta$var$res$min[,coord],0)*correction)

    approx_quant$upper$res$max[,coord]<-apply(objectUQ$profSups[coord,,],1,function(x){return(quantile(x,1-beta))})
    bound_quant$upper$res$max[,coord]<-approx_quant$upper$res$max[,coord]+sqrt(2*pmax(mean_var_delta$var$res$max[,coord],rep(0,nrow(mean_var_delta$var$res$min)))*correction_u)
    #    bound_quant$upper$res$max[,coord]<-approx_quant$upper$res$max[,coord]+mean_var_delta$mean$res$max[,coord]+sqrt(2*max(mean_var_delta$var$res$max[,coord],0)*correction)

    approx_quant$upper$res$min[,coord]<-apply(objectUQ$profInfs[coord,,],1,function(x){return(quantile(x,1-beta))})
    bound_quant$upper$res$min[,coord]<-approx_quant$upper$res$min[,coord]+sqrt(2*pmax(mean_var_delta$var$res$min[,coord],rep(0,nrow(mean_var_delta$var$res$min)))*correction_u)
    #    bound_quant$upper$res$min[,coord]<-approx_quant$upper$res$min[,coord]+mean_var_delta$mean$res$min[,coord]+sqrt(2*max(mean_var_delta$var$res$min[,coord],0)*correction)

  }

  if(p==2){
    temp  <- approx_quant$lower$res$max
    temp2 <- bound_quant$lower$res$max
    approx_quant$lower$res$max <- list()
    approx_quant$lower$res$max <- lapply(seq_len(ncol(temp)), function(i) temp[,i])
    bound_quant$lower$res$max <- list()
    bound_quant$lower$res$max <- lapply(seq_len(ncol(temp2)), function(i) temp2[,i])

    temp  <- approx_quant$lower$res$min
    temp2 <- bound_quant$lower$res$min
    approx_quant$lower$res$min <- list()
    approx_quant$lower$res$min <- lapply(seq_len(ncol(temp)), function(i) temp[,i])
    bound_quant$lower$res$min <- list()
    bound_quant$lower$res$min <- lapply(seq_len(ncol(temp2)), function(i) temp2[,i])

    temp  <- approx_quant$upper$res$max
    temp2 <- bound_quant$upper$res$max
    approx_quant$upper$res$max <- list()
    approx_quant$upper$res$max <- lapply(seq_len(ncol(temp)), function(i) temp[,i])
    bound_quant$upper$res$max <- list()
    bound_quant$upper$res$max <- lapply(seq_len(ncol(temp2)), function(i) temp2[,i])

    temp  <- approx_quant$upper$res$min
    temp2 <- bound_quant$upper$res$min
    approx_quant$upper$res$min <- list()
    approx_quant$upper$res$min <- lapply(seq_len(ncol(temp)), function(i) temp[,i])
    bound_quant$upper$res$min <- list()
    bound_quant$upper$res$min <- lapply(seq_len(ncol(temp2)), function(i) temp2[,i])
  }

  return(list(bound=bound_quant,approx=approx_quant,mean_var_D=mean_var_delta))


}

# obliqueProf_UQ function
#' @name obliqueProf_UQ
#' @author Dario Azzimonti
#' @title Oblique profiles UQ from a kriging model
#' @description The function obliqueProf_UQ computes the profile extrema functions for posterior realizations of a Gaussian process and its confidence bounds
#' @param object either a \link[DiceKriging]{km} model or a list containing partial results. If \code{object} is a km model then all computations are carried out. If \code{object} is a list, then the function carries out all computations to complete the results list.
#' @param allPsi a list containing the matrices Psi (dim \eqn{pxd}) for which to compute the profile extrema
#' @param threshold the threshold of interest
#' @param allResMean a list resulting from \code{getProfileExtrema} or \code{approxProfileExtrema} for the profile extrema on the mean. If NULL the median from the observations is plotted
#' @param quantiles_uq a vector containing the quantiles to be computed
#' @param options_approx an optional list of options for approxProfileExtrema, see \link{approxProfileExtrema} for details.
#' @param options_full_sims an optional list of options for getProfileExtrema, see \link{getProfileExtrema} for details. If NULL the full computations are not executed. NOTE: this computations might be very expensive!
#' @param options_sims an optional list of options for the posterior simulations.
#' \itemize{
#' \item \code{algorithm:} string choice of the algorithm to select the pilot points ("A" or "B", default "B");
#' \item \code{lower:} \eqn{d} dimensional vector with lower bounds for pilot points, default \code{rep(0,d)};
#' \item \code{upper:} \eqn{d} dimensional vector with upper bounds for pilot points, default \code{rep(1,d)};
#' \item \code{batchsize:} number of pilot points, default \code{120};
#' \item \code{optimcontrol:} list containing the options for optimization, see \link[pGPx]{optim_dist_measure};
#' \item \code{integcontrol:} list containing the options for numerical integration of the criterion, see \link[pGPx]{optim_dist_measure};
#' \item \code{integration.param:} list containing the integration design, obtained with the function \link[KrigInv]{integration_design};
#' \item \code{nsim:} number of approximate GP simulations, default \code{300}.
#' }
#' @param options_bound an optional list containing \code{beta} the confidence level for the approximation and \code{alpha} the confidence level for the bound. Note that \code{alpha > 2*beta}. If \code{NULL}, the bound is not computed.
#' @param plot_level an integer to select the plots to return (0=no plots, 1=basic plots, 2= all plots)
#' @param plot_options an optional list of parameters for plots. See \link{setPlotOptions} for currently available options.
#' @param return_level an integer to select the amount of details returned
#' @return If return_level=1 a list containing \itemize{
#' \item \code{profSups:} an array \code{dxfullDesignSizexnsims} containing the profile sup for each coordinate for each realization.
#' \item \code{profInfs:} an array \code{dxfullDesignSizexnsims} containing the profile inf for each coordinate for each realization.
#' \item \code{prof_quantiles_approx:} a list containing the quantiles (levels set by \code{quantiles_uq}) of the profile extrema functions.
#' } if return_level=2 the same list as above but also including \code{more:} a list containing \itemize{
#' \item \code{times:} a list containing
#' 	\itemize{
#' 	\item \code{tSpts:} computational time for selecting pilot points.
#' 	\item \code{tApprox1ord:} vector containing the computational time required for profile extrema computation for each realization
#' 	}
#' \item \code{simuls:} a matrix containing the value of the field simulated at the pilot points
#' \item \code{sPts:} the pilot points
#' }
#' @examples
#' if (!requireNamespace("DiceKriging", quietly = TRUE)) {
#' stop("DiceKriging needed for this example to work. Please install it.",
#'      call. = FALSE)
#' }
#' # Compute a kriging model from 50 evaluations of the Branin function
#' # Define the function
#' g<-function(x){
#'   return(-branin(x))
#' }
#' gp_des<-lhs::maximinLHS(20,2)
#' reals<-apply(gp_des,1,g)
#' kmModel<-km(design = gp_des,response = reals,covtype = "matern3_2")
#'
#' threshold=-10
#' d<-2
#'
#' # Compute oblique profiles UQ starting from GP model
#' # define simulation options
#' options_sims<-list(algorithm="B", lower=rep(0,d), upper=rep(1,d),
#'                    batchsize=80, optimcontrol = list(method="genoud",pop.size=100,print.level=0),
#'                    integcontrol = list(distrib="sobol",n.points=1000), nsim=150)
#' # define approximation options
#' options_approx<- list(multistart=4,heavyReturn=TRUE,
#'                       initDesign=NULL,fullDesignSize=100,
#'                       smoother=NULL)
#' # define plot options
#' options_plots<-list(save=FALSE, titleProf = "Coordinate profiles",
#'                     title2d = "Posterior mean",qq_fill=TRUE)
#'
#' # Define the oblique directions
#' # (for theta=0 it is equal to coordinateProfiles)
#' theta=pi/4
#' allPsi = list(Psi1=matrix(c(cos(theta),sin(theta)),ncol=2),
#'               Psi2=matrix(c(cos(theta+pi/2),sin(theta+pi/2)),ncol=2))
#' \dontrun{
#' # here we reduce the number of simulations to speed up the example
#' # a higher number should be used
#' options_sims$nsim <- 50
#'
#' # profile UQ on approximate oblique profiles
#' oProfiles_UQ<-obliqueProf_UQ(object = kmModel,threshold = threshold,allPsi=allPsi,
#'                              allResMean = NULL,quantiles_uq = c(0.05,0.95),
#'                              options_approx = options_approx, options_full_sims = NULL,
#'                              options_sims = options_sims,options_bound = NULL,
#'                              plot_level = 3, plot_options = options_plots,return_level = 3)
#' # profile UQ on full optim oblique profiles
#'
#' options_full_sims<-list(multistart=4,heavyReturn=TRUE)
#' oProfiles_UQ_full<- obliqueProf_UQ(object = oProfiles_UQ,threshold = threshold,allPsi=allPsi,
#'                              allResMean = NULL,quantiles_uq = c(0.05,0.95),
#'                              options_approx = options_approx, options_full_sims = options_full_sims,
#'                              options_sims = options_sims,options_bound = NULL,
#'                              plot_level = 3, plot_options = options_plots,return_level = 3)
#'
#'
#'
#' # profile UQ on full optim oblique profiles with bound
#' oProfiles_UQ_full_bound<-obliqueProf_UQ(object = oProfiles_UQ_full,threshold = threshold,
#'                                         allPsi=allPsi, allResMean = NULL,
#'                                         quantiles_uq = c(0.05,0.95),
#'                                         options_approx = options_approx,
#'                                         options_full_sims = options_full_sims,
#'                                       options_sims = options_sims,
#'                                       options_bound = list(beta=0.024,alpha=0.05),
#'                                       plot_level = 3, plot_options = options_plots,
#'                                       return_level = 3)
#' }
#' @export
obliqueProf_UQ = function(object,allPsi,threshold,allResMean=NULL,quantiles_uq=c(0.05,0.95),options_approx=NULL,options_full_sims=NULL,options_sims=NULL,options_bound=NULL,plot_level=0,plot_options=NULL,return_level=1){

  # number of thresholds
  num_T<-length(threshold)

  # Check object
  if(is(object,"km")){
    object<-list(kmModel=object)
  }else if(!is.list(object)){
    stop("object must be either a list or a km object")
  }

  # set up dimension
  d<-object$kmModel@d

  # set up Psi dimension
  p=nrow(matrix(allPsi[[1]],ncol=d))

  # Options setup
  # Initialize options_approx
  if(is.null(options_approx)){
    options_approx<- list(multistart=8,heavyReturn=TRUE,fullDesignSize=100,smoother=NULL)
  }

  # initialize limits
  if(is.null(options_approx$limits)){
    ll_b = rep(0,d)
    uu_b = rep(1,d)
  }else{
    ll_b = options_approx$limits$lower
    uu_b = options_approx$limits$upper
  }
  # Useful for choosing limits for eta
  cubeVertex<-matrix(c(ll_b[1],uu_b[1]),nrow=1)
  ii=1
  while(ii<d){
    ii=ii+1
    cubeVertex <- cbind(cubeVertex,cubeVertex)
    cubeVertex <-rbind(cubeVertex, c(rep(ll_b[ii],ncol(cubeVertex)/2),c(rep(uu_b[ii],ncol(cubeVertex)/2))))
  }

  # Initialize initial design for approximation
  num_Psi = length(allPsi)
  if(is.null(options_approx$initDesign)){
    init_des<-list()
    for(i in seq(num_Psi)){
      p<-nrow(allPsi[[i]])
      # Choose limits for etas for current Psi
      mmEtas<-min(crossprod(t(allPsi[[i]]),cubeVertex))
      MMetas<-max(crossprod(t(allPsi[[i]]),cubeVertex))

      # Get initial design
      if(length(options_approx$initDesign)<num_Psi){
        if(p==1){
          init_des[[i]]<-matrix(seq(from=mmEtas[1],to=MMetas[1],,ceiling(sqrt(d)*10)),ncol=1)
        }else{
          ### NEEDS TEST!
          init_des[[i]]<-mmEtas+maximinLHS(ceiling(sqrt(d*p)*10),d)*(MMetas-mmEtas)
          init_des[[i]]<- rbind(mmEtas,init_des[[i]][-2,],MMetas)
        }
      }
      #    init_des<-maximinLHS(ceiling(sqrt(object$kmModel@d)*10),object$kmModel@d)
    }
    options_approx$initDesign<-init_des
  }

  # Set up plot options
  plot_options<-setPlotOptions(plot_options = plot_options,d=d,num_T=num_T,kmModel=object$kmModel)

  # Set-up simulation options
  options_sims_base<-list(algorithm="B", lower=rep(0,d), upper=rep(1,d),
                     batchsize=120, optimcontrol = list(method="genoud",pop.size=100,print.level=0),
                     integcontrol = list(distrib="sobol",n.points=1000),
                     nsim=300)
  options_sims_base$integration.param = integration_design(options_sims_base$integcontrol,d,options_sims_base$lower,options_sims_base$upper,object$kmModel,threshold)
  options_sims_base$integration.param$alpha <- 0.5
  if(is.null(options_sims)){
    options_sims<-list()
  }
  options_sims<-modifyList(options_sims_base,options_sims)
  if(p==1){
    if(is.null(allResMean)){
      quantiles_uq<-c(quantiles_uq,0.5)
    }else{
      changePP <- getChangePoints(threshold = threshold,allRes = allResMean,Design = allResMean$Design)
    }
  }else{
      changePP <- NULL
  }

  if(is.null(object$more)){
    object$more <-list()
  }

  ##### Get the pilot points

  # If not already in object, obtain the pilot points
  if(is.null(object$sPts)){
    timeIn<-get_nanotime()
    object$sPts<-optim_dist_measure(model = object$kmModel,threshold = threshold[1],
                                    lower = options_sims$lower,upper = options_sims$upper,batchsize = options_sims$batchsize,
                                    algorithm = options_sims$algorithm,verb=1,optimcontrol = options_sims$optimcontrol,integration.param = options_sims$integration.param)
    timeMdist<-(get_nanotime()-timeIn)*1e-9
    object$more <- modifyList(object$more,list(times=list(tSpts=timeMdist)))
  }


  #cairo_pdf(paste(plot_options$folderPlots,"critVal_tt",index_exp,".pdf",sep=""),width = 14,height = 14)
  if(plot_level>0){
    oldpar<-par()
    if(plot_options$save)
      cairo_pdf(filename = paste(plot_options$folderPlots,"sPtsCritVal",plot_options$id_save,".pdf",sep=""),width = 12,height = 12)
    plot(object$sPts$value,type='o',main="Optimized criterion",ylab="y",xlab="Iteration")
    if(plot_options$save)
      dev.off()
  }

  ###
  # Prepare the functions for UQ profiles
  nugget.sim=1e-6
  type="UK"
  simu_points<-object$sPts$par

  if(is.null(object$more$simuls)){
    some.simu <- simulate(object=object$kmModel,nsim=options_sims$nsim,newdata=simu_points,nugget.sim=nugget.sim,
                             cond=TRUE,checkNames = FALSE)
  }else{
    some.simu<-object$more$simuls
  }

  g_uq<-function(x,realization,kmModel,simupoints,F.mat=NULL,T.mat=NULL){
    x<-matrix(x,ncol=kmModel@d)
    colnames(x)<-colnames(kmModel@X)
    obj <- krig_weight_GPsimu(object=kmModel,simu_points=simupoints,krig_points=x,T.mat = T.mat,F.mat = F.mat)
    krig.mean.init <- matrix(obj$krig.mean.init,ncol=1)
    weights <- t(obj$Lambda.end)

    return(krig.mean.init + tcrossprod(weights,matrix(realization,nrow=1)))
  }

  g_uq_deriv<-function(x,realization,kmModel,simupoints,T.mat=NULL,F.mat=NULL){
    x<-matrix(x,ncol=kmModel@d)
    colnames(x)<-colnames(kmModel@X)
    obj_deriv<-grad_kweights(object = kmModel,simu_points = simupoints,krig_points = matrix(x,ncol=kmModel@d),T.mat = T.mat,F.mat = F.mat)
    krig_mean_init <- matrix(obj_deriv$krig.mean.init,ncol=kmModel@d)
    weights <- t(obj_deriv$Lambda.end)

    return(krig_mean_init + tcrossprod(matrix(realization,nrow=1),weights))
  }

  # Useful one-time computations
  F.mat <- model.matrix(object=object$kmModel@trend.formula, data = data.frame(rbind(object$kmModel@X,simu_points)))
  K <- covMatrix(object=object$kmModel@covariance,X=rbind(object$kmModel@X,simu_points))$C
  T.mat <- chol(K)

  ### Lets compute the profile extrema for this realization
  # if the profSups and profInfs are not already there, compute them
  if(is.null(object$profSups) || is.null(object$profInfs)){
    # choose size of full design
    object$profSups<-array(NA,dim = c(num_Psi,options_approx$fullDesignSize^p,options_sims$nsim))
    object$profInfs<-array(NA,dim = c(num_Psi,options_approx$fullDesignSize^p,options_sims$nsim))
    object$more$times$tApprox1ord<-rep(NA,options_sims$nsim)
  }

  if(!is.null(options_full_sims) && is.null(object$profSups_full)){
    options_full_sims<-modifyList(list(multistart=4,heavyReturn=TRUE,discretization=options_approx$fullDesignSize),options_full_sims)
    object$profSups_full<-array(NA,dim = c(num_Psi,options_approx$fullDesignSize^p,options_sims$nsim))
    object$profInfs_full<-array(NA,dim = c(num_Psi,options_approx$fullDesignSize^p,options_sims$nsim))
    object$more$times$tFull<-rep(NA,options_sims$nsim)
  }

  for(i in seq(options_sims$nsim)){

    g_uq_spec<-function(x){
      return(g_uq(x=x,realization=some.simu[i,],kmModel = object$kmModel,simupoints = simu_points,F.mat = F.mat,T.mat = T.mat))
    }
    g_uq_der_spec<-function(x){
      return(t(g_uq_deriv(x=x,realization=some.simu[i,],kmModel = object$kmModel,simupoints = simu_points,T.mat = T.mat,F.mat = F.mat)))
    }


    if(!is.null(options_full_sims) && is.na(object$profSups_full[1,1,i])){
      if(i%%10==0){
        cat("Full_sims.Realization ",i,"\n")
      }
      timeIn<-get_nanotime()
      temp_full<-getProfileExtrema(f = g_uq_spec,fprime = g_uq_der_spec,allPsi = allPsi,d = d,opts = options_full_sims)
      object$more$times$tFull[i]<-(get_nanotime()-timeIn)*1e-9

      object$profSups_full[,,i]<-t(temp_full$res$max)
      object$profInfs_full[,,i]<-t(temp_full$res$min)

      # Save the design for plotting
      object$Design_full<-temp_full$Design
    }

    if(is.na(object$profSups[1,1,i]) || is.na(object$profInfs[1,1,i])){
      if(i%%10==0){
        cat("Approx_sims. Realization ",i,"\n")
      }
      timeIn<-get_nanotime()
      temp_1o<-approxProfileExtrema(f = g_uq_spec,fprime = g_uq_der_spec,allPsi = allPsi,d = d,opts = options_approx)
      object$more$times$tApprox1ord[i]<-(get_nanotime()-timeIn)*1e-9

      #  temp<-getProfileExtrema(f=g_uq_spec,fprime = NULL,d=2,options = list(multistart=2,heavyReturn=TRUE))
      object$profSups[,,i]<-t(temp_1o$res$max)
      object$profInfs[,,i]<-t(temp_1o$res$min)

      # Save the design for plotting
      object$Design_approx<-temp_1o$Design
    }
  }



  # save quantiles for approximations
  object$prof_quantiles_approx<-list()
  for(i in seq(length(quantiles_uq))){
    object$prof_quantiles_approx[[i]]<-list(res=list(min=matrix(NA,nrow = options_approx$fullDesignSize^p,ncol = num_Psi),
                                                     max=matrix(NA,nrow = options_approx$fullDesignSize^p,ncol = num_Psi)))
  }
  names(object$prof_quantiles_approx)<-quantiles_uq

  ccPP<-list()
  for(j in seq(length(quantiles_uq))){
    for(coord in seq(num_Psi)){
      object$prof_quantiles_approx[[j]]$res$max[,coord]<-apply(object$profSups[coord,,],1,function(x){return(quantile(x,quantiles_uq[j]))})
      object$prof_quantiles_approx[[j]]$res$min[,coord]<-apply(object$profInfs[coord,,],1,function(x){return(quantile(x,quantiles_uq[j]))})
    }
    if(p==1){
      ccPP[[j]]<-getChangePoints(threshold = threshold,allRes = object$prof_quantiles_approx[[j]],Design = allResMean$Design)
    }else{
      temp <- object$prof_quantiles_approx[[j]]$res$max
      object$prof_quantiles_approx[[j]]$res$max <- list()
      object$prof_quantiles_approx[[j]]$res$max <- lapply(seq_len(ncol(temp)), function(i) temp[,i])
      temp <- object$prof_quantiles_approx[[j]]$res$min
      object$prof_quantiles_approx[[j]]$res$min <- list()
      object$prof_quantiles_approx[[j]]$res$min <- lapply(seq_len(ncol(temp)), function(i) temp[,i])
    }
  }
  if(p==1)
    names(ccPP)<-quantiles_uq

  # save quantiles for full optim
  if(!is.null(options_full_sims)){
    object$prof_quantiles_full<-list()
    for(i in seq(length(quantiles_uq))){
      object$prof_quantiles_full[[i]]<-list(res=list(min=matrix(NA,nrow = options_approx$fullDesignSize^p,ncol = num_Psi),
                                                     max=matrix(NA,nrow = options_approx$fullDesignSize^p,ncol = num_Psi)))
    }
    names(object$prof_quantiles_full)<-quantiles_uq

    ccPP_full<-list()
    for(j in seq(length(quantiles_uq))){
      for(coord in seq(num_Psi)){
        object$prof_quantiles_full[[j]]$res$max[,coord]<-apply(object$profSups_full[coord,,],1,function(x){return(quantile(x,quantiles_uq[j]))})
        object$prof_quantiles_full[[j]]$res$min[,coord]<-apply(object$profInfs_full[coord,,],1,function(x){return(quantile(x,quantiles_uq[j]))})
      }
      if(p==1){
        ccPP_full[[j]]<-getChangePoints(threshold = threshold,allRes = object$prof_quantiles_full[[j]],Design = allResMean$Design)
      }else{
        temp <- object$prof_quantiles_full[[j]]$res$max
        object$prof_quantiles_full[[j]]$res$max <- list()
        object$prof_quantiles_full[[j]]$res$max <- lapply(seq_len(ncol(temp)), function(i) temp[,i])
        temp <- object$prof_quantiles_full[[j]]$res$min
        object$prof_quantiles_full[[j]]$res$min <- list()
        object$prof_quantiles_full[[j]]$res$min <- lapply(seq_len(ncol(temp)), function(i) temp[,i])
      }
    }
    if(p==1)
      names(ccPP_full)<-quantiles_uq

  }

  ## Plot profiles with Uncertainty
#  dd<-seq(0,1,,length.out = options_approx$fullDesignSize)

  if(is.null(allResMean) && p==1)
    changePP<-ccPP$`0.5`

  # Plot the posterior mean and visualize the actual excursion set and the regions of no-excursion according to the profile extrema functions.
  if(plot_level>=2 && num_Psi==2 &&  d==2){
    # since dimension==2 we can plot the posterior mean
    newdata<-expand.grid(seq(0,1,,100),seq(0,1,,100))
    colnames(newdata)<-colnames(object$kmModel@X)
    pred2d<-predict.km(object$kmModel,newdata = newdata,type = "UK",light.return = TRUE,se.compute = FALSE)

    if(plot_options$save)
      cairo_pdf(filename = paste(plot_options$folderPlots,"profMean_UQ",plot_options$id_save,".pdf",sep=""),width = 12,height = 12)
    par(mar = c(5, 5, 4, 2) + 0.1)
    image(matrix(pred2d$mean,nrow = 100),col=gray.colors(20), main=plot_options$title2d,xlab = "", ylab = "", #colnames(object$kmModel@X)[1],ylab= colnames(object$kmModel@X)[2],
          cex.main=3,cex.axis=1.8,cex.lab=2.8)
    contour(matrix(pred2d$mean,nrow = 100),add=T,nlevels = 10,lwd=1.5,labcex=1.2)
    contour(matrix(pred2d$mean,nrow = 100),add=T,levels = threshold,col=plot_options$col_thresh,lwd=3,labcex=1.5)
    for(tt in seq(num_T)){
      plotOblique(changePP$alwaysEx[[tt]][[1]],allPsi[[1]],col=plot_options$col_CCPthresh_alw[tt],lwd=2.5)
      plotOblique(changePP$alwaysEx[[tt]][[2]],allPsi[[2]],col=plot_options$col_CCPthresh_alw[tt],lwd=2.5)
      plotOblique(changePP$neverEx[[tt]][[1]],allPsi[[1]],col=plot_options$col_CCPthresh_nev[tt],lwd=2.5)
      plotOblique(changePP$neverEx[[tt]][[2]],allPsi[[2]],col=plot_options$col_CCPthresh_nev[tt],lwd=2.5)
    }
    if(!is.null(options_full_sims)){
      for(j in seq(length(quantiles_uq))){
        for(tt in seq(num_T)){
          plotOblique(ccPP_full[[j]]$alwaysEx[[tt]][[1]],allPsi[[1]],col=plot_options$col_CCPthresh_alw[tt],lwd=2,lty=2)
          plotOblique(ccPP_full[[j]]$alwaysEx[[tt]][[2]],allPsi[[2]],col=plot_options$col_CCPthresh_alw[tt],lwd=2,lty=2)
          plotOblique(ccPP_full[[j]]$neverEx[[tt]][[1]],allPsi[[1]],col=plot_options$col_CCPthresh_nev[tt],lwd=2,lty=2)
          plotOblique(ccPP_full[[j]]$neverEx[[tt]][[2]],allPsi[[2]],col=plot_options$col_CCPthresh_nev[tt],lwd=2,lty=2)
        }
      }
    }else{
      for(j in seq(length(quantiles_uq))){
        for(tt in seq(num_T)){
          plotOblique(ccPP[[j]]$alwaysEx[[tt]][[1]],allPsi[[1]],col=plot_options$col_CCPthresh_alw[tt],lwd=2,lty=2)
          plotOblique(ccPP[[j]]$alwaysEx[[tt]][[2]],allPsi[[2]],col=plot_options$col_CCPthresh_alw[tt],lwd=2,lty=2)
          plotOblique(ccPP[[j]]$neverEx[[tt]][[1]],allPsi[[1]],col=plot_options$col_CCPthresh_nev[tt],lwd=2,lty=2)
          plotOblique(ccPP[[j]]$neverEx[[tt]][[2]],allPsi[[2]],col=plot_options$col_CCPthresh_nev[tt],lwd=2,lty=2)
        }
      }
    }
    if(plot_options$fun_evals>0){
      points(object$kmModel@X,pch=17,cex=1.6)
    }
    if(plot_options$save)
      dev.off()
  }

  if(plot_level>=1){
#    plot_options$design<-allResMean$Design
    if(p==1){
      object$bound$bound=NULL
      plot_options$design <- object$Design_approx
      plot_univariate_profiles_UQ(objectUQ = object, plot_options = plot_options,nsims = options_sims$nsim,quantiles_uq=quantiles_uq,
                                  threshold = threshold,nameFile ="prof_UQ_approx", profMean = allResMean,typeProf = "approx")

      if(!is.null(options_full_sims)){
        plot_options$design <- object$Design_full
        plot_univariate_profiles_UQ(objectUQ = object, plot_options = plot_options,nsims = options_sims$nsim,quantiles_uq=quantiles_uq,
                                    threshold = threshold,nameFile ="prof_UQ_full", profMean = allResMean,typeProf = "full")
      }
    }else{
      for(j in seq(length(quantiles_uq))){
          if(plot_options$save)
            pdf(file = paste(plot_options$folderPlots,"prof_UQ_approx_q",names(object$prof_quantiles_approx)[j],plot_options$id_save,".pdf",sep=""),width = 18,height = 9)
          plotOneBivProfile(allRes = object$prof_quantiles_approx[[j]],allPsi = allPsi,Design=object$Design_approx,threshold=threshold,main_addendum=paste("(UQ quantile,",names(object$prof_quantiles_approx)[j],")"),xlab=expression(eta[1]),ylab=expression(eta[2]))
          if(plot_options$save)
            dev.off()
      }

    }
  }
  #  object$profSups=profSups
  #  object$profInfs=profInfs
  #  object$prof_quantiles_approx=prof_quantiles_approx
  #  object$sPts=m_dist

  # Compute the bound correction
  if(!is.null(options_bound)){
    object$bound<-bound_profiles(objectUQ = object,mean_var_delta = object$bound$mean_var_D,beta = options_bound$beta,alpha = options_bound$alpha,
                                 options_approx = options_approx,options_full_sims = options_full_sims,allPsi=allPsi)

    if(plot_level>=1){
      if(p==1){
      plot_univariate_profiles_UQ(objectUQ = object, plot_options = plot_options,nsims = options_sims$nsim,quantiles_uq=quantiles_uq,
                                  threshold = threshold,nameFile ="prof_UQ_bound_approx", profMean = allResMean,typeProf = "approx")

      if(!is.null(options_full_sims))
        plot_univariate_profiles_UQ(objectUQ = object, plot_options = plot_options,nsims = options_sims$nsim,quantiles_uq=quantiles_uq,
                                    threshold = threshold,nameFile ="prof_UQ_bound_full", profMean = allResMean,typeProf = "full")
      }else{
          if(plot_options$save)
            pdf(file = paste(plot_options$folderPlots,"prof_UQ_bound_approx_lower",plot_options$id_save,".pdf",sep=""),width = 18,height = 9)
          plotOneBivProfile(allRes = object$bound$bound$lower,allPsi = allPsi,Design=object$Design_approx,threshold=threshold,main_addendum=paste("(UQ bound lower, ",2*options_bound$alpha,")"),xlab=expression(eta[1]),ylab=expression(eta[2]))
          if(plot_options$save)
            dev.off()

          if(plot_options$save)
            pdf(file = paste(plot_options$folderPlots,"prof_UQ_bound_approx_upper",plot_options$id_save,".pdf",sep=""),width = 18,height = 9)
          plotOneBivProfile(allRes = object$bound$bound$upper,allPsi = allPsi,Design=object$Design_approx,threshold=threshold,main_addendum=paste("(UQ bound upper, ",2*options_bound$alpha,")"),xlab=expression(eta[1]),ylab=expression(eta[2]))
          if(plot_options$save)
            dev.off()

      }
    }
  }


  if(return_level==1){
    object$more <- NULL
    return(object)
  }else{
    return(object)
  }

}

#' Coastal flooding as function of offshore forcing conditions
#'
#' A dataset containing the results of a numerical simulation conducted with
#' the MARS model (Lazure and Dumas, 2008) for coastal flooding.
#' The numerical model was adapted to the Boucholeurs area (French Atlantic coast),
#' close to La Rochelle, and validated with data from the 2010 Xynthia storm event.
#' See Azzimonti et al. (2017+) and Rohmer et al. (2018) for more details.
#'
#' The data frame contains 5 input variables: \code{Tide}, \code{Surge}, \code{phi}, \code{t-}, \code{t+} detailing
#' the offshore forcing conditions for the model. All input variables are normalized
#' in \eqn{[0,1]}. The response is \code{Area}, the area flooded in m^2.
#'
#'
#' @format A data frame with 200 rows and 6 variables:
#' \describe{
#'   \item{Tide}{High tide level in meters;}
#'   \item{Surge}{Surge peak amplitude in meters;}
#'   \item{phi}{Phase difference between high tide and surge peak;}
#'   \item{t-}{Duration of the increasing part of the surge temporal signal (assumed to be triangular);}
#'   \item{t+}{Duration of the decreasing part of the surge temporal signal (assumed to be triangular);}
#'   \item{Area}{Flooded area in m^2.}
#' }
#'
#' @examples
#' # Define inputs
#' inputs<-data.frame(coastal_flooding[,-6])
#' colnames(inputs)<-colnames(coastal_flooding[,-6])
#' colnames(inputs)[4:5]<-c("tPlus","tMinus")
#'
#' # put response in areaFlooded variable
#' areaFlooded<-data.frame(coastal_flooding[,6])
#' colnames(areaFlooded)<-colnames(coastal_flooding)[6]
#' response = sqrt(areaFlooded)
#'
#' \donttest{
#' model <- km(formula=~Tide+Surge+I(phi^2)+tMinus+tPlus,
#'             design = inputs,response = response,covtype="matern3_2")
#' # Fix threshold
#' threshold<-sqrt(c(1.2e6,1.9e6,3.1e6,6.5e6))
#'
#' # use the coordinateProfile function
#' ## set up plot options
#' options_plots <- list(save=FALSE, folderPlots = "./" ,
#'                       titleProf = "Coordinate profiles",
#'                       title2d = "Posterior mean",qq_fill=TRUE)
#' # set up full profiles options
#' options_full<-list(multistart=15,heavyReturn=TRUE)
#' # set up approximation options
#' d <- model@@d
#' init_des<-lhs::maximinLHS(5*d , d )
#' options_approx<- list(multistart=2,heavyReturn=TRUE, initDesign=init_des,
#'                       fullDesignSize=100, smoother="quantSpline")
#'
#' # run the coordinate profile extrema on the mean
#' CF_CoordProf_mean<- coordinateProfiles(object = model, threshold = threshold,
#'                                        uq_computations = FALSE, options_approx = options_approx,
#'                                        plot_level=3, plot_options= options_plots, return_level=3,
#'                                        options_full=options_full)
#'}
#'\dontrun{
#' ## UQ computations might require a long time
#' # set up simulation options
#' ## reduce nsims and batchsize for faster/less accurate UQ
#' nsims=200
#' opts_sims<-list(algorithm="B", lower=rep(0,d ),
#'                 upper=rep(1,d ), batchsize=150,
#'                 optimcontrol=list(method="genoud", pop.size=100,print.level=0),
#'                 integcontrol = list(distrib="sobol",n.points=2000),nsim=nsims)
#'
#' opts_sims$integration.param <- integration_design(opts_sims$integcontrol,
#'                                                   d , opts_sims$lower,
#'                                                   opts_sims$upper,
#'                                                   model,threshold)
#' opts_sims$integration.param$alpha <- 0.5
#'
#' # run UQ computations
#' CF_CoordProf_UQ<- coordinateProfiles(object = CF_CoordProf_mean, threshold = threshold,
#'                                      uq_computations = TRUE, options_approx = options_approx,
#'                                      plot_level=3, plot_options= options_plots, return_level=3,
#'                                      options_sims=opts_sims,options_full=options_full,
#'                                      options_bound = list(beta=0.024,alpha=0.05))
#' }
#'
#' @references
#'
#' Azzimonti, D., Ginsbourger, D., Rohmer, J. and Idier, D. (2017+). \emph{Profile extrema for visualizing and quantifying uncertainties on excursion regions. Application to coastal flooding.} arXiv:1710.00688.
#'
#' Rohmer, J., Idier, D., Paris, F., Pedreros, R., and Louisor, J. (2018). \emph{Casting light on forcing and breaching scenarios that lead to marine inundation: Combining numerical simulations with a random-forest classification approach.} Environmental Modelling & Software, 104:64-80.
#'
#' Lazure, P. and Dumas, F. (2008). \emph{An external-internal mode coupling for a 3D hydrodynamical model for applications at regional scale (MARS)}. Advances in Water Resources, 31:233-250.
#'
"coastal_flooding"

# getProfileSup function with smarter algorithm!
#' @author Dario Azzimonti
#' @name getProfileSup_optim
#' @title Generic profile sup function computation with optim
#' @description Compute profile sup function for an arbitrary matrix \code{Psi} with the L-BFGS-B algorithm of \link[stats]{optim}.
#' @param eta \eqn{p} dimensional point where the function is to be evaluated
#' @param Psi projection matrix of dimensions \code{p x d}
#' @param f function to be optimized (takes a vector y of dimension d and returns a real number)
#' @param fprime derivative of f (same format, returning a \eqn{d} dimensional vector)
#' @param d dimension of the input for f
#' @param options a list containing the options to be passed to optim:
#' \itemize{
#' \item \code{par:} contains the starting point (a point in dimension d-1)
#' \item \code{lower:} the lower bounds for the optimization domain (see optim for details)
#' \item \code{upper:} the upper bounds for the optimization domain (see optim for details)
#' }
#' @return a real value corresponding to \eqn{max_{x \in D_Psi} f(x)}
#' @seealso \link{getProfileInf_optim}, \link{plotMaxMin}
#' @export
# requires library(MASS) and lpSolve
getProfileSup_optim = function(eta,Psi,f,fprime,d,options=NULL){

  p = dim(Psi)[1]

  # put to zero small values
  Psi[which(abs(Psi)<.Machine$double.neg.eps)]<-0

  if(!is.null(options$trace))
    trace=options$trace

  # find equations for boundary conditions on z
  ## TODO: add option to set box-constraints
  lower<-rep(0,d)
  upper<-rep(1,d)

  # Verify if the feasible set is not empty
  Hlarge<-makeH(a1 = rbind(diag(-1,nrow = d,ncol=d),diag(1,nrow = d,ncol=d)),
                b1 = rbind(matrix(lower,nrow = d),matrix(upper,nrow = d)),
                a2 = Psi,b2=matrix(eta))

  Hlarge<-d2q(Hlarge)
  solLP<-lpcdd(hrep = Hlarge,objgrd = d2q(rep(0,d)))

  if(solLP$solution.type=="Optimal"){
    # Find any solution of the linear system
    xiSol<-matrix(q2d(solLP$primal.solution),nrow=d) #solutionLinConstraint(eta=eta,Psi=Psi,low = lower,upp=upper)
  }else{
    warning("Cannot proceed with these constraints. A simple LP optimization problem does not have an optimal solution")
    optRes<-list(value=NA,solLP=solLP,message="The feasible set is either empty (solLP$solution.type==Inconsistent) or the problem is dual inconsistent. See solLP for the results of a LP problem with those contraints and objective function 0.")
    return(list(val=optRes$value,aux=optRes))
  }
  # rm(Hlarge,solLP)
  # gc()


  nullSpacePsi<-Null(t(Psi))



  Az <- rbind(-nullSpacePsi,nullSpacePsi)
  bz <- rbind(xiSol-lower,-xiSol+upper)
  #  vertices_poly<-matrix(enumerate.vertices(A=Az,b=bz),ncol=d-p)
  rcdd_Hrep0<-makeH(a1 = Az,b1 = bz)
  rcdd_Hrep<-d2q(rcdd_Hrep0)
  # rm(rcdd_Hrep0)
  # gc()
  vertices_poly<-matrix(q2d(scdd(rcdd_Hrep)$output[,-c(1,2)]),ncol=d-p)


  ff<-function(z){
    x_cand<-xiSol+nullSpacePsi%*%z
    g_cand<- sqrt(.Machine$double.xmax)
    if(all(x_cand>=rep(0,d))&all(x_cand<=rep(1,d))){
      return(-f(x_cand))
    }else{
      return(g_cand)
    }
  }

  ffprime<-function(z){
    x_cand<-xiSol+nullSpacePsi%*%z
    if(all(Az%*%z-bz<=0)){
      return(-crossprod(fprime(xiSol+nullSpacePsi%*%z),nullSpacePsi))
    }else{
      ww<-Az%*%z-bz
      dirs<-which(ww>0)
      ww<-matrix(ww[dirs]/sum(ww[dirs]))
      return(crossprod(ww,-Az[dirs,]))
      #return(-Az[dirs,]%*%ww)
    }
    return(-crossprod(fprime(xiSol+nullSpacePsi%*%z),nullSpacePsi))
  }

  if(!is.null(options$par)){
    #    startingPoint<-options$par
    #   if(Psi%*%startingPoint< eta)
    warning("Constraints check not implemented!")
  }else{
    #    startingPoint<-runif(min = lower,max = upper,d-p)
    startingPoint<-colMeans(vertices_poly)
    #      matrix(c(xiSol,1))
  }

  if(dim(vertices_poly)[1]==1){
    # The feasible region for the subproblem is one point, the starting point,
    # so it doesn't make sense to try different starting points
    options$multistart=NULL
  }

  if(!is.null(options$multistart)){
    optRess<-NULL
    optRes<-list(value=-Inf,aux=NA)
    for(i in seq(options$multistart)){
      if(i==1){
        startingPoint<-colMeans(vertices_poly)
      }else if(i==2){
        m_lp<-lpcdd(rcdd_Hrep,objgrd = d2q(rep(0,d-p)))
        if(m_lp$solution.type=="Optimal")
          startingPoint<-q2d(m_lp$primal.solution)
      }else if(i==3){
        m_lp<-lpcdd(rcdd_Hrep,objgrd = d2q(rep(1,d-p)),minimize = FALSE)
        if(m_lp$solution.type=="Optimal")
          startingPoint<-q2d(m_lp$primal.solution)
      }else{
        m_lp<-lpcdd(rcdd_Hrep,objgrd = d2q(runif(d-p)))
        if(m_lp$solution.type=="Optimal")
          startingPoint<-q2d(m_lp$primal.solution)
      }

      if(!is.null(options$par) && i==1)
        startingPoint<-options$par
      #      optRess<-optim(par = startingPoint,fn = ff,gr = ffprime,lower=lower,upper=upper,method="L-BFGS-B")
      optRess<-optim(par = startingPoint,fn = ff,gr = ffprime,method="BFGS")
      optRess$value= -optRess$value
      #      optRess<-optim(par = startingPoint,fn = fxFixed,gr = fprimexFixed,control = list(fnscale=-1,trace=trace,pgtol=1e-4),method = methd,lower = lower,upper = upper)
      if(optRess$value>=optRes$value){
        optRes=optRess
      }else{
        break
      }
    }
  }else{
    #    optRes<-optim(par = startingPoint,fn = ff,gr = ffprime,lower=lower,upper=upper,method="L-BFGS-B")#,method = mm)
    optRes<-optim(par = startingPoint,fn = ff,gr = ffprime,method="BFGS")
    optRes$value = -optRes$value
  }

  optRes$solution = xiSol+Null(t(Psi))%*%optRes$par

  rm(rcdd_Hrep0,Hlarge,solLP)
#  gc()

  return(list(val=optRes$value,aux=optRes))
}


# getProfileInf function
#' @author Dario Azzimonti
#' @name getProfileInf_optim
#' @title Generic profile inf function computation with optim
#' @description Compute profile inf function for an arbitrary matrix \code{Psi} with with the L-BFGS-B algorithm of \link[stats]{optim}. Here the linear equality constraint is eliminated by using the Null space of \code{Psi}.
#' @param eta \eqn{p} dimensional point where the function is to be evaluated
#' @param Psi projection matrix of dimension \code{pxd}
#' @param f function to be optimized (takes a vector y of dimension d and returns a real number)
#' @param fprime derivative of f (same format, returning a \eqn{d} dimensional vector)
#' @param d dimension of the input for f
#' @param options a list containing the options to be passed to optim:
#' \itemize{
#' \item \code{par:} contains the starting point (a point in dimension d)
#' \item \code{lower:} the lower bounds for the optimization domain (see optim for details)
#' \item \code{upper:} the upper bounds for the optimization domain (see optim for details)
#' }
#' @return a real value corresponding to \eqn{min_{x \in D_Psi} f(x)}
#' @seealso \link{getProfileSup_optim}, \link{plotMaxMin}
#' @export
getProfileInf_optim = function(eta,Psi,f,fprime,d,options=NULL){

  p = dim(Psi)[1]

  # put to zero small values
  Psi[which(abs(Psi)<.Machine$double.neg.eps)]<-0

  if(!is.null(options$trace))
    trace=options$trace


  # find equations for boundary conditions on z
  ## TODO: add option to set box-constraints
  lower<-rep(0,d)
  upper<-rep(1,d)

  # Verify if the feasible set is not empty
  Hlarge<-makeH(a1 = rbind(diag(-1,nrow = d,ncol=d),diag(1,nrow = d,ncol=d)),
                b1 = rbind(matrix(lower,nrow = d),matrix(upper,nrow = d)),
                a2 = Psi,b2=matrix(eta))
  Hlarge<-d2q(Hlarge)
  solLP<-lpcdd(hrep = Hlarge,objgrd = d2q(rep(0,d)))

  if(solLP$solution.type=="Optimal"){
    # Find any solution of the linear system
    xiSol<-matrix(q2d(solLP$primal.solution),nrow=d) #solutionLinConstraint(eta=eta,Psi=Psi,low = lower,upp=upper)
  }else{
    warning("Cannot proceed with these constraints. A simple LP optimization problem does not have an optimal solution")
    optRes<-list(value=NA,solLP=solLP,message="The feasible set is either empty (solLP$solution.type==Inconsistent) or the problem is dual inconsistent. See solLP for the results of a LP problem with those contraints and objective function 0.")
    return(list(val=optRes$value,aux=optRes))
  }
  # rm(Hlarge,solLP)
  # gc()


  nullSpacePsi<-Null(t(Psi))


  Az <- rbind(-nullSpacePsi,nullSpacePsi)
  bz <- rbind(xiSol-lower,-xiSol+upper)
  #  vertices_poly<-matrix(enumerate.vertices(A=Az,b=bz),ncol=d-p)
  rcdd_Hrep0<-makeH(a1 = Az,b1 = bz)
  rcdd_Hrep<-d2q(rcdd_Hrep0)
  vertices_poly<-matrix(q2d(scdd(rcdd_Hrep)$output[,-c(1,2)]),ncol=d-p)


  ff<-function(z){
    x_cand<-xiSol+nullSpacePsi%*%z
    g_cand<- sqrt(.Machine$double.xmax)
    if(all(x_cand>=rep(0,d))&all(x_cand<=rep(1,d))){
      return(f(x_cand))
    }else{
      return(g_cand)
    }
  }

  ffprime<-function(z){
    x_cand<-xiSol+nullSpacePsi%*%z
    if(all(Az%*%z-bz<=0)){
      return(crossprod(fprime(xiSol+nullSpacePsi%*%z),nullSpacePsi))
    }else{
      ww<-Az%*%z-bz
      dirs<-which(ww>0)
      ww<-matrix(ww[dirs]/sum(ww[dirs]))
      return(crossprod(ww,-Az[dirs,]))
      #return(-Az[dirs,]%*%ww)
    }
    return(crossprod(fprime(xiSol+nullSpacePsi%*%z),nullSpacePsi))
  }

  if(!is.null(options$par)){
    #    startingPoint<-options$par
    #   if(Psi%*%startingPoint< eta)
    warning("Constraints check not implemented!")
  }else{
    #    startingPoint<-runif(min = lower,max = upper,d-p)
    startingPoint<-colMeans(vertices_poly)#lp(direction = "max",objective.in = runif(d-p),const.mat = Az,const.dir = "<=",const.rhs = bz)$solution
    #      matrix(c(xiSol,1))
  }

  if(dim(vertices_poly)[1]==1){
    # The feasible region for the subproblem is one point, the starting point,
    # so it doesn't make sense to try different starting points
    options$multistart=NULL
  }


  if(!is.null(options$multistart)){
    #  cat("not implemented yet!")
    optRess<-NULL
    optRes<-list(value=Inf,aux=NA)
    for(i in seq(options$multistart)){
      if(i==1){
        startingPoint<-colMeans(vertices_poly)
      }else if(i==2){
        m_lp<-lpcdd(rcdd_Hrep,objgrd = d2q(rep(0,d-p)))
        if(m_lp$solution.type=="Optimal")
          startingPoint<-q2d(m_lp$primal.solution)
      }else if(i==3){
        m_lp<-lpcdd(rcdd_Hrep,objgrd = d2q(rep(1,d-p)),minimize = FALSE)
        if(m_lp$solution.type=="Optimal")
          startingPoint<-q2d(m_lp$primal.solution)
      }else{
        m_lp<-lpcdd(rcdd_Hrep,objgrd = d2q(runif(d-p)))
        if(m_lp$solution.type=="Optimal")
          startingPoint<-q2d(m_lp$primal.solution)
      }

      if(!is.null(options$par) && i==1)
        startingPoint<-options$par
      #      optRess<-optim(par = startingPoint,fn = ff,gr = ffprime,lower=lower,upper=upper,method="L-BFGS-B")
      optRess<-optim(par = startingPoint,fn = ff,gr = ffprime,method="BFGS")
      if(optRess$value<=optRes$value){
        optRes=optRess
      }else{
        break
      }
    }
  }else{
    #    optRes<-optim(par = startingPoint,fn = ff,gr = ffprime,lower=lower,upper=upper,method="L-BFGS-B")#,method = mm)
    optRes<-optim(par = startingPoint,fn = ff,gr = ffprime,method="BFGS")
  }


  optRes$solution = xiSol+Null(t(Psi))%*%optRes$par

  rm(rcdd_Hrep0,Hlarge,solLP)
#  gc()

  return(list(val=optRes$value,aux=optRes))

}

# cleanProfileResults function
#' @name cleanProfileResults
#' @author Dario Azzimonti
#' @title Clean a profile extrema object
#' @description The function cleanProfileResults cleans a profile extrema object to partially redo some computations.
#' @param object a list containing profile extrema results.
#' @param level an integer 1-4 denoting how much it should be removed from \code{object}. See Value for details.
#' @return returns \code{object} with the deleted parts as selected by \code{level}. In particular \itemize{
#' \item \code{1:} keep only \code{profMean_full}.
#' \item \code{2:} keep \code{profMean_full} and \code{profMean_approx}. Remove all UQ results.
#' \item \code{3:} keep \code{profMean_full} and \code{profMean_approx} and the pilot points. Remove all UQ simulations.
#' \item \code{4:} Remove only the bound computations.
#' }
#'
#' @examples
#' if (!requireNamespace("DiceKriging", quietly = TRUE)) {
#' stop("DiceKriging needed for this example to work. Please install it.",
#'      call. = FALSE)
#' }
#' # Compute a kriging model from 50 evaluations of the Branin function
#' # Define the function
#' g=function(x){
#'   return(-branin(x))
#' }
#' gp_des<-lhs::maximinLHS(20,2)
#' reals<-apply(gp_des,1,g)
#' kmModel<-km(design = gp_des,response = reals,covtype = "matern3_2")
#'
#' threshold=-10
#'
#' # Compute coordinate profiles on the posterior mean
#' # Increase multistart and size of designs for more precise results
#' options_full<-list(multistart=2,heavyReturn=TRUE, Design = replicate(2,seq(0,1,,50)))
#' init_des<-lhs::maximinLHS(12,2)
#' options_approx<- list(multistart=2,heavyReturn=TRUE,initDesign=init_des,fullDesignSize=50)
#' cProfilesMean<-coordinateProfiles(object=kmModel,threshold=threshold,options_full=options_full,
#'                                   options_approx=options_approx,uq_computations=FALSE,
#'                                   plot_level=3,plot_options=NULL,CI_const=NULL,return_level=2)
#'
#' # If we want to run again the computation of approximate coordinate profiles
#' # we delete that result and run again the coordinate profiles function
#' cProfiles_full <- cleanProfileResults(cProfilesMean,level=1)
#' \dontrun{
#' # Coordinate profiles with UQ with approximate profiles
#' plot_options<-list(save=FALSE, titleProf = "Coordinate profiles",
#'                    title2d = "Posterior mean",qq_fill=TRUE)
#' cProfilesUQ<-coordinateProfiles(object=cProfilesMean,threshold=threshold,options_full=options_full,
#'                                   options_approx=options_approx,uq_computations=TRUE,
#'                                   plot_level=3,plot_options=NULL,CI_const=NULL,return_level=2)
#' # If we would like to remove all UQ results
#' cProf_noUQ <- cleanProfileResults(cProfilesUQ,level=2)
#'
#' # If we would like to remove the simulations but keep the pilot points
#' cProf_noSims <- cleanProfileResults(cProfilesUQ,level=3)
#' # the line above is useful, for example, if we need a more accurate UQ. In that case
#' # we obtain more simulations with the same pilot points and then combine the results.
#'
#' }
#' @export
cleanProfileResults = function(object,level = 1){
  if(level <= 4){
    object$res_UQ$bound$bound <- NULL
    object$res_UQ$bound$approx <- NULL
    if(level <=3){
      object$res_UQ$profSups <- NULL
      object$res_UQ$profInfs <- NULL
      object$res_UQ$more <- NULL
      object$res_UQ$prof_quantiles_approx <- NULL
      if(level <= 2){
        object$res_UQ <- NULL
        if(level <=1){
          object$profMean_approx <- NULL
          object$more$times$approx <- NULL
          object$more$abs_err<- NULL
        }
      }
    }
  }
  return(object)
}

# plotBivariateProfiles function
#' @author Dario Azzimonti
#' @title Plot bivariate profiles
#' @name plotBivariateProfiles
#' @description Plot bivariate profiles, for dimension up to 6.
#' @param bivProf list returned by \code{obliqueProfiles}.
#' @param allPsi a list containing the matrices Psi (dim \eqn{2xd}) for which to compute the profile extrema
#' @param Design a matrix of dimension \eqn{(2d)x numPsi} encoding the first (\code{Design[1:d,]}) and the second ((\code{Design[(d+1):(2*d),]})) axis values.
#' @param threshold if not \code{NULL} plots the level as a contour.
#' @param whichIQR which quantiles to use for the inter-quantile range plot. If \code{NULL}, automatically selects the first and the last element of \code{bivProfres_UQ$prof_quantiles_approx}
#' @param plot_options list as returned by \code{setPlotOptions}.
#' @param ... additional parameters to be passed to the plot function
#' @return plots the 2d maps of the profile sup and inf of the function for each Psi in allPsi. If threshold is not NULL also contours the threshold level.
#' @export
plotBivariateProfiles<-function(bivProf,allPsi,Design=NULL,threshold=NULL,whichIQR=NULL,plot_options=NULL,...){

  if(is.null(plot_options))
    plot_options <- setPlotOptions(plot_options = plot_options,d = bivProf$kmModel@d,num_T = 1,kmModel = bivProf$kmModel)

  trueEvals <-plot_options$fun_evals

  num_Psi<-length(allPsi)

  if(is.null(Design)){
    warning("Design not selected, auto choice not implemented")
    return(NULL)
  }

  dd_eta<-length(Design[,1])/2

  bkg_col <- gray.colors(20)

  if(!is.null(threshold)){
    whichAbove = bivProf$kmModel@y>threshold
  }else{
    trueEvals = min(1,trueEvals)
  }


  ## Mean maps
  allRes <- bivProf$profMean_full

  if(plot_options$save)
    pdf(file = paste(plot_options$folderPlots,"profMean_full",plot_options$id_save,".pdf",sep=""),width = 18,height = 9)

  for(i in seq(num_Psi)){

#    whichVars<-as.logical(apply(allPsi[[i]]==1,2,sum))
#    pp <- bivProf$kmModel@X[,whichVars]

    wwVV<- apply(allPsi[[i]],2,function(x)return(x>0))

    pp<-bivProf$kmModel@X%*%t(allPsi[[i]])

    xlab_i <- paste(allPsi[[i]][1,wwVV[1,]],"*",plot_options$coord_names[wwVV[1,]],collapse = " + ")
    ylab_i <- paste(allPsi[[i]][2,wwVV[2,]],"*",plot_options$coord_names[wwVV[2,]],collapse = " + ")

    if(trueEvals<3){
      zzBackgroud_max<-matrix(allRes$res$max[[i]],ncol=dd_eta)
      zzBackgroud_min<-matrix(allRes$res$min[[i]],ncol=dd_eta)
      bck_name <- "mean"
    }else{
      propPoints<-getPointProportion(pp=pp,xBins = Design[1:(dd_eta),i],
                                     yBins = Design[(dd_eta+1):(2*dd_eta),i],whichAbove = whichAbove,plt = FALSE)

      zzBackgroud_max<-propPoints$freq
      zzBackgroud_min<-propPoints$freq
      bck_name <- "frequency"
    }
#    layout(matrix(1:2,nrow=1),widths=c(0.5,0.5))
    layout(matrix(1:4,nrow=1),widths=c(0.45,0.05,0.45,0.05))
    par(mar=c(4,4,4,0.2) + 0.2)
    image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackgroud_max,col=bkg_col,
          main=bquote(P[Psi[.(i)]]^sup ~" mean, "~  .(bck_name) ),xlab= xlab_i, ylab = ylab_i,...)
    contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$max[[i]],ncol=dd_eta),nlevels = 10,add=T)
    contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$max[[i]],ncol=dd_eta),levels = threshold,add=T,col=2,lwd=1.4)
    switch (as.character(trueEvals),
            "0" = {},
            "1" = {points(pp,pch=3)},
            {points(pp[-whichAbove,],pch=3,col="blue3");points(pp[whichAbove,],pch=3,col=3)})

    legend_info<-hist(zzBackgroud_max,plot=FALSE,breaks=20)
    which_col_leg <- which(legend_info$counts>0)
    par(mar=c(5,0.2,5,2.8))
    image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
    axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg),...)#,mgp=c(0,.5,0))

    par(mar=c(4,4,4,0.2) + 0.2)
    image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackgroud_min,col=bkg_col,
          main=bquote(P[Psi[.(i)]]^inf ~" mean, "~  .(bck_name) ),xlab= xlab_i, ylab = ylab_i,...)
    contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$min[[i]],ncol=dd_eta),nlevels = 10,add=T)
    contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$min[[i]],ncol=dd_eta),levels = threshold,add=T,col=2,lwd=1.4)
    switch (as.character(trueEvals),
            "0" = {},
            "1" = {points(pp,pch=3)},
            {points(pp[-whichAbove,],pch=3,col="blue3");points(pp[whichAbove,],pch=3,col=3)})

    legend_info<-hist(zzBackgroud_min,plot=FALSE,breaks=20)
    which_col_leg <- which(legend_info$counts>0)
    par(mar=c(5,0.2,5,2.8))
    image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
    axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg),...)#,mgp=c(0,.5,0))

  }

  if(plot_options$save)
    dev.off()


  ## quantiles maps
  if(!is.null(bivProf$res_UQ$prof_quantiles_approx)){
    for(j in seq(length(bivProf$res_UQ$prof_quantiles_approx))){
      allRes <- bivProf$res_UQ$prof_quantiles_approx[[j]]


      if(plot_options$save)
        pdf(file = paste(plot_options$folderPlots,"profQuant_",names(bivProf$res_UQ$prof_quantiles_approx)[j],plot_options$id_save,".pdf",sep=""),width = 18,height = 9)

      for(i in seq(num_Psi)){

#        whichVars<-as.logical(apply(allPsi[[i]]==1,2,sum))
#        pp <- bivProf$kmModel@X[,whichVars]

        wwVV<- apply(allPsi[[i]],2,function(x)return(x>0))

        pp<-bivProf$kmModel@X%*%t(allPsi[[i]])

        xlab_i <- paste(allPsi[[i]][1,wwVV[1,]],"*",plot_options$coord_names[wwVV[1,]],collapse = " + ")
        ylab_i <- paste(allPsi[[i]][2,wwVV[2,]],"*",plot_options$coord_names[wwVV[2,]],collapse = " + ")

        if(trueEvals<3){
          zzBackgroud_max<-matrix(allRes$res$max[[i]],ncol=dd_eta)
          zzBackgroud_min<-matrix(allRes$res$min[[i]],ncol=dd_eta)
          bck_name <- "mean"
        }else{
          propPoints<-getPointProportion(pp=pp,xBins = Design[1:(dd_eta),i],
                                         yBins = Design[(dd_eta+1):(2*dd_eta),i],whichAbove = whichAbove,plt = FALSE)

          zzBackgroud_max<-propPoints$freq
          zzBackgroud_min<-propPoints$freq
          bck_name <- "frequency"
        }
#        layout(matrix(1:2,nrow=1),widths=c(0.5,0.5))
        layout(matrix(1:4,nrow=1),widths=c(0.45,0.05,0.45,0.05))
        par(mar=c(4,4,4,0.2) + 0.2)
        image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackgroud_max,col=bkg_col,
              main=bquote(P[Psi[.(i)]]^sup ~" Z (" ~ .(names(bivProf$res_UQ$prof_quantiles_approx)[j]) ~ "), "~  .(bck_name) ),xlab= xlab_i, ylab = ylab_i,...)
        contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$max[[i]],ncol=dd_eta),nlevels = 10,add=T)
        contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$max[[i]],ncol=dd_eta),levels = threshold,add=T,col=2,lwd=1.4)
        switch (as.character(trueEvals),
                "0" = {},
                "1" = {points(pp,pch=3)},
                {points(pp[-whichAbove,],pch=3,col="blue3");points(pp[whichAbove,],pch=3,col=3)})


        legend_info<-hist(zzBackgroud_max,plot=FALSE,breaks=20)
        which_col_leg <- which(legend_info$counts>0)
        par(mar=c(5,0.2,5,2.8))
        image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
        axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg),...)#,mgp=c(0,.5,0))

        par(mar=c(4,4,4,0.2) + 0.2)
        image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackgroud_min,col=bkg_col,
              main=bquote(P[Psi[.(i)]]^inf ~" Z (" ~ .(names(bivProf$res_UQ$prof_quantiles_approx)[j]) ~ "), " ~ .(bck_name) ),xlab= xlab_i, ylab = ylab_i,...)
        contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$min[[i]],ncol=dd_eta),nlevels = 10,add=T)
        contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$min[[i]],ncol=dd_eta),levels = threshold,add=T,col=2,lwd=1.4)
        switch (as.character(trueEvals),
                "0" = {},
                "1" = {points(pp,pch=3)},
                {points(pp[-whichAbove,],pch=3,col="blue3");points(pp[whichAbove,],pch=3,col=3)})


        legend_info<-hist(zzBackgroud_min,plot=FALSE,breaks=20)
        which_col_leg <- which(legend_info$counts>0)
        par(mar=c(5,0.2,5,2.8))
        image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
        axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg))#,mgp=c(0,.5,0))
      }

      if(plot_options$save)
        dev.off()
    }

    if(is.null(whichIQR)){
      whichIQR <- c(1,length(bivProf$res_UQ$prof_quantiles_approx))
    }

    actualQQ <- round(as.numeric(names(bivProf$res_UQ$prof_quantiles_approx))[whichIQR],3)

    # Compute weighted interquantile range
    if(!is.null(bivProf$res_UQ$prof_quantiles_approx$`0.50`$res)){
      allRes<-bivProf$res_UQ$prof_quantiles_approx$`0.50`$res
    }else{
      allRes <- bivProf$profMean_full
    }
    d<-bivProf$kmModel@d
    #/(abs(z-tt)^d+1)^(1/d)

    qqRange<-bivProf$res_UQ$prof_quantiles_approx[[whichIQR[2]]]$res
    qqRange$min<-mapply(FUN=function(x,y,z,tt){(x - y)},x=qqRange$min,y=bivProf$res_UQ$prof_quantiles_approx[[whichIQR[1]]]$res$min,z=allRes$res$min,MoreArgs = list(tt=threshold))
    qqRange$max<-mapply(FUN=function(x,y,z,tt){(x - y)},x=qqRange$max,y=bivProf$res_UQ$prof_quantiles_approx[[whichIQR[1]]]$res$max,z=allRes$res$max,MoreArgs = list(tt=threshold))
    #qqRange$min<-qqRange$min-bivProf$res_UQ$prof_quantiles_approx[[whichIQR[1]]]$res$min
    #qqRange$max<-qqRange$max-bivProf$res_UQ$prof_quantiles_approx[[whichIQR[1]]]$res$max
    ddRange<-qqRange
    ddRange$max<-ddRange$max*sapply(bivProf$res_UQ$prof_quantiles_approx[[whichIQR[2]]]$res$max,FUN=function(x){x>threshold},simplify=TRUE)*sapply(bivProf$res_UQ$prof_quantiles_approx[[whichIQR[1]]]$res$max,FUN=function(x){x<threshold},simplify=TRUE)#*matrix(as.integer(qquantiles[[whichIQR[1]]]$res$max<threshold),ncol=2)

    ddRange$min<-ddRange$min*sapply(bivProf$res_UQ$prof_quantiles_approx[[whichIQR[2]]]$res$min,FUN=function(x){x>threshold},simplify=TRUE)*sapply(bivProf$res_UQ$prof_quantiles_approx[[whichIQR[1]]]$res$min,FUN=function(x){x<threshold},simplify=TRUE)#matrix(as.integer(qquantiles[[whichIQR[2]]]$res$min>threshold)*as.integer(qquantiles[[whichIQR[1]]]$res$min<threshold),ncol=2)

#    ddRange$min <- ddRange$min + mapply(FUN=function(x,y,z,tt){abs(2*x - y-z)/(abs(z-tt)^d+1)^(1/d)},x=allRes$res$min,y=bivProf$res_UQ$prof_quantiles_approx[[whichIQR[2]]]$res$min,z=bivProf$res_UQ$prof_quantiles_approx[[whichIQR[1]]]$res$min,MoreArgs = list(tt=threshold))
#    ddRange$max <- ddRange$max + mapply(FUN=function(x,y,z,tt){abs(2*x - y-z)/(abs(z-tt)^d+1)^(1/d)},x=allRes$res$max,y=bivProf$res_UQ$prof_quantiles_approx[[whichIQR[2]]]$res$max,z=bivProf$res_UQ$prof_quantiles_approx[[whichIQR[1]]]$res$max,MoreArgs = list(tt=threshold))

    if(plot_options$save)
      pdf(file = paste(plot_options$folderPlots,"profUQrange",plot_options$id_save,".pdf",sep=""),width = 18,height = 9,pointsize = 15)

    for(i in seq(num_Psi)){

#      whichVars<-as.logical(apply(allPsi[[i]]==1,2,sum))
#      pp <- bivProf$kmModel@X[,whichVars]

      wwVV<- apply(allPsi[[i]],2,function(x)return(x>0))

      pp<-bivProf$kmModel@X%*%t(allPsi[[i]])

      xlab_i <- paste(allPsi[[i]][1,wwVV[1,]],"*",plot_options$coord_names[wwVV[1,]],collapse = " + ")
      ylab_i <- paste(allPsi[[i]][2,wwVV[2,]],"*",plot_options$coord_names[wwVV[2,]],collapse = " + ")

      zzBackgroud_max<-matrix(ddRange$max[,i],ncol=dd_eta)
      zzBackgroud_min<-matrix(ddRange$min[,i],ncol=dd_eta)

#      layout(matrix(1:2,nrow=1),widths=c(0.5,0.5))
      layout(matrix(1:4,nrow=1),widths=c(0.45,0.05,0.45,0.05))
      par(mar=c(4,4,4,0.2) + 0.2)
      image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackgroud_max,col=bkg_col,
            main=bquote(P[Psi[.(i)]]^sup ~" mean, wIQR " ~ .(actualQQ[1]) ~ "-" ~ .(actualQQ[2]) ),xlab= xlab_i, ylab = ylab_i,...)
      contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$max[[i]],ncol=dd_eta),nlevels = 10,add=T,lwd = 1.5,labcex=1)
      contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$max[[i]],ncol=dd_eta),levels = threshold,add=T,col=2,lwd=2,labcex=1,lty=2)
      switch (as.character(trueEvals),
              "0" = {},
              "1" = {points(pp,pch=3)},
              {points(pp[-whichAbove,],pch=3,col="blue3");points(pp[whichAbove,],pch=3,col=3)})

      legend_info<-hist(zzBackgroud_max,plot=FALSE,breaks=20)
      which_col_leg <- which(legend_info$counts>0)
      par(mar=c(5,0.2,5,2.8))
      image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
      axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg),...)#,mgp=c(0,.5,0))

      par(mar=c(4,4,4,0.2) + 0.2)
      image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackgroud_min,col=bkg_col,
            main=bquote(P[Psi[.(i)]]^inf ~" mean, wIQR " ~ .(actualQQ[1]) ~ "-" ~ .(actualQQ[2]) ),xlab= xlab_i, ylab = ylab_i,...)
      contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$min[[i]],ncol=dd_eta),nlevels = 10,add=T,lwd = 1.5,labcex=1)
      contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$min[[i]],ncol=dd_eta),levels = threshold,add=T,col=2,lwd=2,labcex=1,lty=2)
      switch (as.character(trueEvals),
              "0" = {},
              "1" = {points(pp,pch=3)},
              {points(pp[-whichAbove,],pch=3,col="blue3");points(pp[whichAbove,],pch=3,col=3)})

      legend_info<-hist(zzBackgroud_min,plot=FALSE,breaks=20)
      which_col_leg <- which(legend_info$counts>0)
      par(mar=c(5,0.2,5,2.8))
      image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
      axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg),...)#,mgp=c(0,.5,0))
    }

    if(plot_options$save)
      dev.off()



  }

  ## bound maps
  if(!is.null(bivProf$res_UQ$bound$bound)){
    for(j in seq(length(bivProf$res_UQ$bound$bound))){
      allRes <- bivProf$res_UQ$bound$bound[[j]]

      if(plot_options$save)
        pdf(file = paste(plot_options$folderPlots,"profBound_",names(bivProf$res_UQ$bound$bound)[j],plot_options$id_save,".pdf",sep=""),width = 18,height = 9)


      for(i in seq(num_Psi)){

#        whichVars<-as.logical(apply(allPsi[[i]]==1,2,sum))
#        pp <- bivProf$kmModel@X[,whichVars]

        wwVV<- apply(allPsi[[i]],2,function(x)return(x>0))

        pp<-bivProf$kmModel@X%*%t(allPsi[[i]])

        xlab_i <- paste(allPsi[[i]][1,wwVV[1,]],"*",plot_options$coord_names[wwVV[1,]],collapse = " + ")
        ylab_i <- paste(allPsi[[i]][2,wwVV[2,]],"*",plot_options$coord_names[wwVV[2,]],collapse = " + ")

        bck_name <- "mean"
        if(trueEvals<3){
          zzBackgroud_max<-matrix(allRes$res$max[[i]],ncol=dd_eta)
          zzBackgroud_min<-matrix(allRes$res$min[[i]],ncol=dd_eta)
          bck_name <- "mean"
        }else{
          propPoints<-getPointProportion(pp=pp,xBins = Design[1:(dd_eta),i],
                                         yBins = Design[(dd_eta+1):(2*dd_eta),i],whichAbove = whichAbove,plt = FALSE)

          zzBackgroud_max<-propPoints$freq
          zzBackgroud_min<-propPoints$freq
          bck_name <- "frequency"
        }
#        layout(matrix(1:2,nrow=1),widths=c(0.5,0.5))
        layout(matrix(1:4,nrow=1),widths=c(0.45,0.05,0.45,0.05))
        par(mar=c(4,4,4,0.2) + 0.2)
        image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackgroud_max,col=bkg_col,
              main=bquote(P[Psi[.(i)]]^sup ~ .(names(bivProf$res_UQ$bound$bound)[j]) ~ " bound," ~ " "~ .(bck_name)),xlab= xlab_i, ylab = ylab_i,...)
        contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$max[[i]],ncol=dd_eta),nlevels = 10,add=T)
        contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$max[[i]],ncol=dd_eta),levels = threshold,add=T,col=2,lwd=1.4)
        switch (as.character(trueEvals),
                "0" = {},
                "1" = {points(pp,pch=3)},
                {points(pp[-whichAbove,],pch=3,col="blue3");points(pp[whichAbove,],pch=3,col=3)})

        legend_info<-hist(zzBackgroud_max,plot=FALSE,breaks=20)
        which_col_leg <- which(legend_info$counts>0)
        par(mar=c(5,0.2,5,2.8))
        image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
        axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg),...)#,mgp=c(0,.5,0))

        par(mar=c(4,4,4,0.2) + 0.2)
        image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackgroud_min,col=bkg_col,
              main=bquote(P[Psi[.(i)]]^inf ~ .(names(bivProf$res_UQ$bound$bound)[j]) ~ " bound," ~ " "~ .(bck_name)),xlab= xlab_i, ylab = ylab_i,...)
        contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$min[[i]],ncol=dd_eta),nlevels = 10,add=T)
        contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$min[[i]],ncol=dd_eta),levels = threshold,add=T,col=2,lwd=1.4)
        switch (as.character(trueEvals),
                "0" = {},
                "1" = {points(pp,pch=3)},
                {points(pp[-whichAbove,],pch=3,col="blue3");points(pp[whichAbove,],pch=3,col=3)})

        legend_info<-hist(zzBackgroud_min,plot=FALSE,breaks=20)
        which_col_leg <- which(legend_info$counts>0)
        par(mar=c(5,0.2,5,2.8))
        image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
        axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg),...)#,mgp=c(0,.5,0))
      }

      if(plot_options$save)
        dev.off()
    }


    # Compute weighted bound range
    if(!is.null(bivProf$res_UQ$prof_quantiles_approx$`0.50`$res)){
      allRes<-bivProf$res_UQ$prof_quantiles_approx$`0.50`$res
    }else{
      allRes <- bivProf$profMean_full
    }
    d<-bivProf$kmModel@d
    #/(abs(z-tt)^d+1)^(1/d)

    qqRange<-bivProf$res_UQ$bound$bound$upper$res
    qqRange$min<-mapply(FUN=function(x,y,z,tt){(x - y)},x=qqRange$min,y=bivProf$res_UQ$bound$bound$lower$res$min,z=allRes$res$min,MoreArgs = list(tt=threshold))
    qqRange$max<-mapply(FUN=function(x,y,z,tt){(x - y)},x=qqRange$max,y=bivProf$res_UQ$bound$bound$lower$res$max,z=allRes$res$max,MoreArgs = list(tt=threshold))
    #qqRange$min<-qqRange$min-bivProf$res_UQ$prof_quantiles_approx[[whichIQR[1]]]$res$min
    #qqRange$max<-qqRange$max-bivProf$res_UQ$prof_quantiles_approx[[whichIQR[1]]]$res$max
    ddRange<-qqRange
    ddRange$max<-ddRange$max*sapply(bivProf$res_UQ$bound$bound$upper$res$max,FUN=function(x){x>threshold},simplify=TRUE)*sapply(bivProf$res_UQ$bound$bound$lower$res$max,FUN=function(x){x<threshold},simplify=TRUE)

    ddRange$min<-ddRange$min*sapply(bivProf$res_UQ$bound$bound$upper$res$min,FUN=function(x){x>threshold},simplify=TRUE)*sapply(bivProf$res_UQ$bound$bound$lower$res$min,FUN=function(x){x<threshold},simplify=TRUE)

    #    ddRange$min <- ddRange$min + mapply(FUN=function(x,y,z,tt){abs(2*x - y-z)/(abs(z-tt)^d+1)^(1/d)},x=allRes$res$min,y=bivProf$res_UQ$bound$bound$upper$res$min,z=bivProf$res_UQ$bound$bound$lower$res$min,MoreArgs = list(tt=threshold))
    #    ddRange$max <- ddRange$max + mapply(FUN=function(x,y,z,tt){abs(2*x - y-z)/(abs(z-tt)^d+1)^(1/d)},x=allRes$res$max,y=bivProf$res_UQ$bound$bound$upper$res$max,z=bivProf$res_UQ$bound$bound$lower$res$max,MoreArgs = list(tt=threshold))

    if(plot_options$save)
      pdf(file = paste(plot_options$folderPlots,"profUBoundRange",plot_options$id_save,".pdf",sep=""),width = 18,height = 9,pointsize = 15)

    for(i in seq(num_Psi)){

#      whichVars<-as.logical(apply(allPsi[[i]]==1,2,sum))
#      pp <- bivProf$kmModel@X[,whichVars]

      wwVV<- apply(allPsi[[i]],2,function(x)return(x>0))

      pp<-bivProf$kmModel@X%*%t(allPsi[[i]])

      xlab_i <- paste(allPsi[[i]][1,wwVV[1,]],"*",plot_options$coord_names[wwVV[1,]],collapse = " + ")
      ylab_i <- paste(allPsi[[i]][2,wwVV[2,]],"*",plot_options$coord_names[wwVV[2,]],collapse = " + ")

      zzBackgroud_max<-matrix(ddRange$max[,i],ncol=dd_eta)
      zzBackgroud_min<-matrix(ddRange$min[,i],ncol=dd_eta)

#      layout(matrix(1:2,nrow=1),widths=c(0.5,0.5))
      layout(matrix(1:4,nrow=1),widths=c(0.45,0.05,0.45,0.05))
      par(mar=c(4,4,4,0.2) + 0.2)
      image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackgroud_max,col=bkg_col,
            main=bquote(P[Psi[.(i)]]^sup ~ "mean, bound range"),xlab= xlab_i, ylab = ylab_i,...)
      contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$max[[i]],ncol=dd_eta),nlevels = 10,add=T,lwd = 1.5,labcex=1)
      contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$max[[i]],ncol=dd_eta),levels = threshold,add=T,col=2,lwd=2,labcex=1,lty=2)
      switch (as.character(trueEvals),
              "0" = {},
              "1" = {points(pp,pch=3)},
              {points(pp[-whichAbove,],pch=3,col="blue3");points(pp[whichAbove,],pch=3,col=3)})


      legend_info<-hist(zzBackgroud_max,plot=FALSE,breaks=20)
      which_col_leg <- which(legend_info$counts>0)
      par(mar=c(5,0.2,5,2.8))
      image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
      axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg),...)#,mgp=c(0,.5,0))

      par(mar=c(4,4,4,0.2) + 0.2)
      image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackgroud_min,col=bkg_col,
            main=bquote(P[Psi[.(i)]]^inf ~ "mean, bound range"),xlab= xlab_i, ylab = ylab_i,...)
      contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$min[[i]],ncol=dd_eta),nlevels = 10,add=T,lwd = 1.5,labcex=1)
      contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$min[[i]],ncol=dd_eta),levels = threshold,add=T,col=2,lwd=2,labcex=1,lty=2)
      switch (as.character(trueEvals),
              "0" = {},
              "1" = {points(pp,pch=3)},
              {points(pp[-whichAbove,],pch=3,col="blue3");points(pp[whichAbove,],pch=3,col=3)})

      legend_info<-hist(zzBackgroud_min,plot=FALSE,breaks=20)
      which_col_leg <- which(legend_info$counts>0)
      par(mar=c(5,0.2,5,2.8))
      image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
      axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg),...)#,mgp=c(0,.5,0))
    }

    if(plot_options$save)
      dev.off()


    ### Plot the bivariate profile mean as contour and sigma_T^Delta as background
    if(plot_options$save)
      pdf(file = paste(plot_options$folderPlots,"profUQ_sigmaD",plot_options$id_save,".pdf",sep=""),width = 18,height = 9,pointsize = 15)

    for(i in seq(num_Psi)){

#      whichVars<-as.logical(apply(allPsi[[i]]==1,2,sum))
#      pp <- bivProf$kmModel@X[,whichVars]

      wwVV<- apply(allPsi[[i]],2,function(x)return(x>0))

      pp<-bivProf$kmModel@X%*%t(allPsi[[i]])

      xlab_i <- paste(allPsi[[i]][1,wwVV[1,]],"*",plot_options$coord_names[wwVV[1,]],collapse = " + ")
      ylab_i <- paste(allPsi[[i]][2,wwVV[2,]],"*",plot_options$coord_names[wwVV[2,]],collapse = " + ")

      zzBackgroud_max<-matrix(sqrt(pmax(bivProf$res_UQ$bound$mean_var_D$var$res$max[,i],0)),ncol=dd_eta)
      zzBackgroud_min<-matrix(sqrt(pmax(bivProf$res_UQ$bound$mean_var_D$var$res$min[,i],0)),ncol=dd_eta)


#      zzBackgroud_max<-matrix(ddRange$max[,i],ncol=dd_eta)
#      zzBackgroud_min<-matrix(ddRange$min[,i],ncol=dd_eta)

#      layout(matrix(1:2,nrow=1),widths=c(0.5,0.5))
      layout(matrix(1:4,nrow=1),widths=c(0.45,0.05,0.45,0.05))
      par(mar=c(4,4,4,0.2) + 0.2)
      image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackgroud_max,col=bkg_col,
            main=bquote(P[Psi[.(i)]]^sup ~ "mean, " ~(sigma[T]^tilde(Delta)) ),xlab= xlab_i, ylab = ylab_i,...)
      contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$max[[i]],ncol=dd_eta),nlevels = 10,add=T,lwd = 1.5,labcex=1)
      contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$max[[i]],ncol=dd_eta),levels = threshold,add=T,col=2,lwd=2,labcex=1,lty=2)
      switch (as.character(trueEvals),
              "0" = {},
              "1" = {points(pp,pch=3)},
              {points(pp[-whichAbove,],pch=3,col="blue3");points(pp[whichAbove,],pch=3,col=3)})


      legend_info<-hist(zzBackgroud_max,plot=FALSE,breaks=20)
      which_col_leg <- which(legend_info$counts>0)
      par(mar=c(5,0.2,5,2.8))
      image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
      axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg),...)#,mgp=c(0,.5,0))

      par(mar=c(4,4,4,0.2) + 0.2)
      image(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=zzBackgroud_min,col=bkg_col,
            main=bquote(P[Psi[.(i)]]^inf ~ "mean, " ~(sigma[T]^tilde(Delta)) ),xlab= xlab_i, ylab = ylab_i,...)
      contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$min[[i]],ncol=dd_eta),nlevels = 10,add=T,lwd = 1.5,labcex=1)
      contour(x=Design[1:(dd_eta),i],y=Design[(dd_eta+1):(2*dd_eta),i],z=matrix(allRes$res$min[[i]],ncol=dd_eta),levels = threshold,add=T,col=2,lwd=2,labcex=1,lty=2)
      switch (as.character(trueEvals),
              "0" = {},
              "1" = {points(pp,pch=3)},
              {points(pp[-whichAbove,],pch=3,col="blue3");points(pp[whichAbove,],pch=3,col=3)})

      legend_info<-hist(zzBackgroud_min,plot=FALSE,breaks=20)
      which_col_leg <- which(legend_info$counts>0)
      par(mar=c(5,0.2,5,2.8))
      image(y=1:length(which_col_leg),z=t(1:length(which_col_leg)), col=bkg_col[which_col_leg], axes=FALSE)#, main="Slope", cex.main=.8)
      axis(4,cex.axis=0.8,labels = legend_info$breaks[which_col_leg],at=1:length(which_col_leg),...)#,mgp=c(0,.5,0))
    }

    if(plot_options$save)
      dev.off()

  }


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profExtrema-package.R
\name{profExtrema}
\alias{profExtrema}
\title{profExtrema package}
\description{
Computation and plots of profile extrema functions. The package main functions are: \describe{
   \item{\strong{Computation:}}{ \itemize{
    \item \code{\link{coordinateProfiles}}: Given a \link[DiceKriging]{km} objects computes the coordinate profile extrema function for the posterior mean and its quantiles.
    \item \code{\link{coordProf_UQ}}: UQ part of \code{coordinateProfiles}.
    \item \code{\link{obliqueProfiles}}: Given a \link[DiceKriging]{km} objects computes the profile extrema functions for a generic list of matrices Psi for the posterior mean and its quantiles.
    \item \code{\link{obliqueProf_UQ}}: The UQ part of \code{obliqueProfiles}.
    \item \code{\link{getAllMaxMin}}: computes coordinate profile extrema with full optimization for a deterministic function.
    \item \code{\link{approxMaxMin}}: approximates coordinate profile extrema for a deterministic function.
    \item \code{\link{getProfileExtrema}}: computes profile extrema given a list of matrices Psi for a deterministic function.
    \item \code{\link{approxProfileExtrema}}: approximates profile extrema given a list of matrices Psi for a deterministic function.
   } }
   \item{\strong{Plotting:}}{ \itemize{
    \item \code{\link{plot_univariate_profiles_UQ}}: plots for the results of \code{coordProf_UQ} or \code{obliqueProf_UQ}. Note that this function only works for univariate profiles.
    \item \code{\link{plotBivariateProfiles}}: plots the bivariate maps results of a call to \code{obliqueProfiles} with a two dimensional projection matrix Psi.
    \item \code{\link{plotMaxMin}}: simple plotting function for univariate profile extrema.
    \item \code{\link{plotOneBivProfile}}: simple plotting function for bivariate profile extrema.
   }  }

}
}
\details{
Package: profExtrema \cr
Type: Package \cr
Version: 0.2.1 \cr
Date: 2020-03-20
}
\note{
This work was supported in part the Hasler Foundation, grant number 16065 and by the Swiss National Science Foundation, grant number 167199. The author warmly thanks David Ginsbourger, Jérémy Rohmer and Déborah Idier for fruitful discussions and accurate, thought provoking suggestions.
}
\references{
Azzimonti, D., Bect, J., Chevalier, C., and Ginsbourger, D. (2016). Quantifying uncertainties on excursion sets under a Gaussian random field prior. SIAM/ASA Journal on Uncertainty Quantification, 4(1):850-874.

Azzimonti, D., Ginsbourger, D., Rohmer, J. and Idier, D. (2017+). Profile extrema for visualizing and quantifying uncertainties on excursion regions. Application to coastal flooding. arXiv:1710.00688.

Chevalier, C. (2013). Fast uncertainty reduction strategies relying on Gaussian process models. PhD thesis, University of Bern.

Chevalier, C., Picheny, V., Ginsbourger, D. (2014). An efficient and user-friendly implementation of batch-sequential inversion strategies based on kriging. Computational Statistics & Data Analysis, 71: 1021-1034.

Johnson, S. G.  The NLopt nonlinear-optimization package, http://ab-initio.mit.edu/nlopt

Koenker, R. (2017). quantreg: Quantile Regression. R package version 5.33.

Nocedal, J. and Wright, S. J. (2006). Numerical Optimization, second edition. Springer- Verlag, New York.

Neuwirth, E. (2014). RColorBrewer: ColorBrewer Palettes. R package version 1.1-2.

Roustant, O., Ginsbourger, D., Deville, Y. (2012). DiceKriging, DiceOptim: Two R Packages for the Analysis of Computer Experiments by Kriging-Based Metamodeling and Optimization. Journal of Statistical Software, 51(1): 1-55.
}
\author{
Dario Azzimonti (dario.azzimonti@gmail.com) .
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_univariate_profiles_UQ.R
\name{setPlotOptions}
\alias{setPlotOptions}
\title{Set-up the plot options when NULL}
\usage{
setPlotOptions(plot_options = NULL, d, num_T, kmModel = NULL)
}
\arguments{
\item{plot_options}{the list of plot options to set-up}

\item{d}{number of coordinates}

\item{num_T}{number of thresholds of interest}

\item{kmModel}{a \link[DiceKriging]{km} model, used to obtain the coordinates names.}
}
\value{
the properly set-up list containing the following fields \itemize{
\item \code{save:} boolean, if TRUE saves the plots in \code{folderPlots}
\item \code{folderPlots:} a string containing the destination folder for plots, if \code{save==TRUE} default is \code{./}
\item \code{ylim:} a matrix \code{coord}x2 containing the ylim for each coordinate, if NULL in \code{plot_options} this is left NULL and automatically set at the plot time.
\item \code{titleProf:} a string containing the title for the coordinate profile plots, default is \code{"Coordinate profiles"}
\item \code{title2d:} a string containing the title for the 2d plots (if the input is 2d), default is \code{"Posterior mean"}
\item \code{design:} a \eqn{dxr} matrix where \eqn{d} is the input dimension and \eqn{r} is the size of the discretization for plots at each dimension
\item \code{coord_names:} a \eqn{d}-vector of characters naming the dimensions. If NULL and \code{kmModel} not NULL then it is the names of \code{kmModel@X} otherwise \code{x_1,...,x_d}
\item \code{id_save:} a string to be added to the plot file names, useful for serial computations on HPC, left as in \code{plot_options}.
\item \code{qq_fill:} if TRUE it fills the region between the first 2 quantiles in \code{quantiles_uq} and between the upper and lower bound in \code{objectUQ$bound$bound}, if \code{NULL}, it is set as \code{FALSE}.
\item \code{bound_cols:} a vector of two strings containing the names of the colors for upper and lower bound plots.
\item \code{qq_fill_colors:} a list containing the colors for qq_fill: \code{approx} for 2 quantiles, \code{bound_min} for bounds on the profile inf, \code{bound_max} for profile sup. Initialized only if \code{qq_fill==TRUE}.
\item \code{col_CCPthresh_nev:} Color palette of dimension \code{num_T} for the colors of the vertical lines delimiting the intersections between the profiles sup and the thresholds
\item \code{col_CCPthresh_alw:} Color palette of dimension \code{num_T} for the colors of the vertical lines delimiting the intersections between the profiles inf and the thresholds
\item \code{col_thresh:} Color palette of dimension \code{num_T} for the colors of the thresholds
\item \code{fun_evals:} integer denoting the level of plot for the true evaluations. \itemize{
\item 0: default, no plots for true evaluations;
\item 1: plot the true evaluations as points in 2d plots, no true evaluation plots in 1d;
\item 2: plot true evaluations, in 2d with different color for values above threshold;
\item 3: plot true evaluations, in 2d plots in color, with background of the image colored as proportion of points inside excursion; }
}
if all the fields are already filled then returns \code{plot_options}
}
\description{
Function to set-up plot options for \link{plot_univariate_profiles_UQ}, \link{plotBivariateProfiles}, \link{coordinateProfiles}, \link{coordProf_UQ}, \link{obliqueProfiles} and \link{obliqueProf_UQ}.
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getMaxNumOpt.R
\name{getMaxMinMC}
\alias{getMaxMinMC}
\title{Coordinate profile extrema with MC}
\usage{
getMaxMinMC(x, f, fprime, coord, d, options = NULL)
}
\arguments{
\item{x}{one dimensional point where the function is to be evaluated}

\item{f}{function to be optimized (takes a vector y of dimension d and returns a real number)}

\item{fprime}{derivative of f (same format)}

\item{coord}{integer selecting the dimension that is fixed, the other ones are optimized over}

\item{d}{dimension of the input for f}

\item{options}{a list containing the options to be passed to the MC optimizer:
\itemize{
\item \code{par:} contains the starting point (a point in dimension d-1)
\item \code{numMCsamples:} number of MC samples
\item \code{rand} string that chooses the type of randomness in MC: "unif" (uniform in [lower,upper]), "norm" (independent normal with mean 0 and variance 1)
\item \code{lower:} the lower bounds for the optimization domain (see optim for details)
\item \code{upper:} the upper bounds for the optimization domain (see optim for details)
}}
}
\value{
a real value corresponding to \eqn{max_{x_1,\dots, x_{coord-1},x_{coord+1}, \dots, x_d} f(x_1,\dots,x_d)}
}
\description{
Compute coordinate profile extrema with MC
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getMaxNumOpt.R
\name{approxMaxMin}
\alias{approxMaxMin}
\title{Approximate coordinate profile functions}
\usage{
approxMaxMin(f, fprime = NULL, d, opts = NULL)
}
\arguments{
\item{f}{the function to be evaluated}

\item{fprime}{derivative of the function}

\item{d}{dimension of the input domain}

\item{opts}{a list containing the options for this function and the subfunctions getMax, getMin or getMaxMinMC, see documentation of getMax, getMin, getMaxMinMC for details. The options only for approxMaxMin are
\itemize{
\item \code{limits:} an optional list with the upper and lower limits of each dimension, if NULL then for each dimension limits are 0,1
\item \code{smoother:} Select which smoother to use:a string that selects which smoother to use: \itemize{
      \item \code{"1order"}: first order interpolation with gradient
      \item \code{"splineSmooth"}: smoothing spline with default degrees of freedom (DEFAULT OPTION)
      \item \code{"quantSpline"}: profile inf and profile sup approximated with quantile spline regression at levels 0.1 and 0.9 respectively
}
\item \code{heavyReturn:} If TRUE returns also all minimizers, default is FALSE.
\item \code{initDesign:} The design of few points where the expensive sup is evaluated.
\item \code{fullDesignSize:} The full design where the function is approximated.
\item \code{multistart:} number of multistarts for optim procedure.
\item \code{MonteCarlo:} if TRUE, computes sup with Monte Carlo procedure.
\item \code{numMCsamples:} number of MC samples for the sup.
\item \code{plts:} If TRUE, plots the max/min functions at each coordinate, default is FALSE.
\item \code{verb:} If TRUE, outputs intermediate results, default is FALSE.
}}
}
\value{
a list of two data frames (min, max) of the evaluations of \eqn{f_sup(x_i) = sup_{x_j \neq i} f(x_1,\dots,x_d) } and \eqn{f_inf(x_i) = inf_{x_j \neq i} f(x_1,\dots,x_d) }
for each i at the design Design. By default Design is a 100 equally spaced points for each dimension. It can be changed by defining it in options$Design
}
\description{
Evaluate profile extrema over other variables with approximations at few values
}
\examples{
if (!requireNamespace("DiceKriging", quietly = TRUE)) {
stop("DiceKriging needed for this example to work. Please install it.",
     call. = FALSE)
}
# Compute the coordinate profile extrema with full optimization on 2d example

# Define the function
g=function(x){
  return(-branin(x))
}
# Define the gradient
gprime = function(x){
  x1 = x[1]*15-5
  x2 = x[2]*15
  f1prime = (15*25)/(4*pi^4)*x1^3 - (15*75)/(2*pi^3)*x1^2 +
  (80*15)/(pi^2)*x1 - (5*15)/(pi^2)*x2*x1 +
  10*15/pi*x2 - 60*15/pi-10*15* (1 - 1/(8*pi))*sin(x1)
  f2prime = 2*15*(x2-5/(4*pi^2)*x1^2 +5/pi*x1-6)
  return(matrix(c(-f1prime,-f2prime),nrow=1))
}

# generic approximation options
init_des<-lhs::maximinLHS(15,2)
options_approx<- list(multistart=4,heavyReturn=TRUE,initDesign=init_des,fullDesignSize=100)

# 1order approximation
options_approx$smoother<-"1order"
coordProf_approx_1order<-approxMaxMin(f = g,fprime = gprime,d=2,opts = options_approx)

# quantile regression
options_approx$smoother<-"quantSpline"
coordProf_approx_quantReg<-approxMaxMin(f = g,fprime = gprime,d=2,opts = options_approx)


\donttest{
# Consider threshold=-10
threshold<- -10
# obtain the points where the profiles take the threshold value
pp_change<-getChangePoints(threshold = threshold,allRes = coordProf_approx_quantReg)
# evaluate g at a grid and plot the image
x<-seq(0,1,,100)
grid<-expand.grid(x,x)
g_evals<- apply(X = grid,MARGIN = 1,FUN = g)
image(x = x,y = x,z = matrix(g_evals,nrow = 100),col = grey.colors(20))
contour(x=x,y=x,z=matrix(g_evals,nrow = 100), add=TRUE, nlevels = 20)
contour(x=x,y=x,z=matrix(g_evals,nrow = 100), add=TRUE, levels = threshold,col=2)
abline(h = pp_change$neverEx$`-10`[[2]],col="darkgreen",lwd=2)
abline(v = pp_change$neverEx$`-10`[[1]],col="darkgreen",lwd=2)
# Plot the coordinate profiles and a threshold
plotMaxMin(allRes = coordProf_approx_1order,threshold = threshold,changes = TRUE)
plotMaxMin(allRes = coordProf_approx_quantReg,threshold = threshold,changes = TRUE)
}
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bound_functions.R
\name{bound_profiles}
\alias{bound_profiles}
\title{Bound for profile extrema quantiles}
\usage{
bound_profiles(
  objectUQ,
  mean_var_delta = NULL,
  beta = 0.0124,
  alpha = 0.025,
  allPsi = NULL,
  options_approx = NULL,
  options_full_sims = NULL
)
}
\arguments{
\item{objectUQ}{an object returned by \link{coordProf_UQ} or the object saved in \code{obj$res_UQ}, if \code{obj} is the object returned by \link{coordinateProfiles}}

\item{mean_var_delta}{the profile extrema functions at \code{options_approx$design} for the mean and variance function of the difference process \eqn{Z^\Delta = Z_x - \widetilde{Z}_x}. Object returned by \link{prof_mean_var_Delta}.}

\item{beta}{the level of confidence for the approximate simulations}

\item{alpha}{the level of confidence for the bound}

\item{allPsi}{optional list of matrices (dim \eqn{pxd}) for which to compute the profile extrema. If NULL coordinate profiles are computed.}

\item{options_approx}{an optional list of options for \link{approxMaxMin} (or \link{approxProfileExtrema} if \code{allPsi} not NULL).}

\item{options_full_sims}{an optional list of options for \link{getAllMaxMin} (or \link{getProfileExtrema} if \code{allPsi} not NULL). If NULL the full computations are not excuted. NOTE: this computations might be very expensive!}
}
\value{
a list containing \itemize{
\item \code{bound: } a list containing the upper/lower bound for profile sup and inf
\item \code{approx: } a list containing the upper/lower approximate quantiles for profile sup and inf
}
}
\description{
The function bound_profiles computes the upper and lower bounds for the profile extrema quantiles of a Gaussian process model.
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_univariate_profiles_UQ.R
\name{plot_univariate_profiles_UQ}
\alias{plot_univariate_profiles_UQ}
\title{Univariate profile extrema with UQ}
\usage{
plot_univariate_profiles_UQ(
  objectUQ,
  plot_options,
  nsims,
  threshold,
  nameFile = "prof_UQ",
  quantiles_uq = c(0.05, 0.95),
  profMean = NULL,
  typeProf = "approx"
)
}
\arguments{
\item{objectUQ}{an object returned by \link{coordProf_UQ} or the object saved in \code{obj$res_UQ}, if \code{obj} is the object returned by \link{coordinateProfiles}.}

\item{plot_options}{a list containing the same elements as the one passed to \link{coordinateProfiles}}

\item{nsims}{number of simulations}

\item{threshold}{threshold of interest}

\item{nameFile}{the central name of the plot file}

\item{quantiles_uq}{a vector containing the quantiles to be computed}

\item{profMean}{the profile coordinate extrema functions for the mean. It is saved in \code{obj$profMean_full} or \code{obj$profMean_approx} if \code{obj} is an object returned by \link{coordinateProfiles}.}

\item{typeProf}{a string to choose with type of profile extrema for simulations to plot \itemize{
\item "approx": plots only the approximate profile extrema for simulations
\item "full": plots only the full profile extrema for simulations
\item "both": plots both the approximate and full profile extrema for simulations
}}
}
\value{
Plots either to the default graphical device or to pdf (according to the options passed in \code{plot_options})
}
\description{
Function to plot the univariate profile extrema functions with UQ
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profExtrema-package.R
\docType{package}
\name{profExtrema-package}
\alias{profExtrema-package}
\title{profExtrema: Compute and Visualize Profile Extrema Functions}
\description{
Computes profile extrema functions for arbitrary functions. If the function is expensive-to-evaluate it computes profile extrema by emulating the function with a Gaussian process (using package 'DiceKriging'). In this case uncertainty quantification on the profile extrema can also be computed. The different plotting functions for profile extrema give the user a tool to better locate excursion sets.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/dazzimonti/profExtrema}
  \item Report bugs at \url{https://github.com/dazzimonti/profExtrema/issues}
}

}
\author{
\strong{Maintainer}: Dario Azzimonti \email{dario.azzimonti@gmail.com} (\href{https://orcid.org/0000-0001-5080-3061}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Dario Azzimonti \email{dario.azzimonti@gmail.com} (\href{https://orcid.org/0000-0001-5080-3061}{ORCID}) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getMaxNumOpt.R
\name{getMax}
\alias{getMax}
\title{Coordinate profile sup function}
\usage{
getMax(x, f, fprime, coord, d, options = NULL)
}
\arguments{
\item{x}{one dimensional point where the function is to be evaluated}

\item{f}{function to be optimized (takes a vector y of dimension d and returns a real number)}

\item{fprime}{derivative of f (same format)}

\item{coord}{integer selecting the dimension that is fixed, the other ones are optimized over}

\item{d}{dimension of the input for f}

\item{options}{a list containing the options to be passed to optim:
\itemize{
\item \code{par:} contains the starting point (a point in dimension d-1)
\item \code{method:} is the string denoting the chosen method for the optimization (see optim for details)
\item \code{lower:} the lower bounds for the optimization domain (see optim for details)
\item \code{upper:} the upper bounds for the optimization domain (see optim for details)
}}
}
\value{
a real value corresponding to \eqn{max_{x_1,\dots, x_{coord-1},x_{coord+1}, \dots, x_d} f(x_1,\dots,x_d)}
}
\description{
Compute coordinate profile sup functions
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cleanProfiles.R
\name{cleanProfileResults}
\alias{cleanProfileResults}
\title{Clean a profile extrema object}
\usage{
cleanProfileResults(object, level = 1)
}
\arguments{
\item{object}{a list containing profile extrema results.}

\item{level}{an integer 1-4 denoting how much it should be removed from \code{object}. See Value for details.}
}
\value{
returns \code{object} with the deleted parts as selected by \code{level}. In particular \itemize{
\item \code{1:} keep only \code{profMean_full}.
\item \code{2:} keep \code{profMean_full} and \code{profMean_approx}. Remove all UQ results.
\item \code{3:} keep \code{profMean_full} and \code{profMean_approx} and the pilot points. Remove all UQ simulations.
\item \code{4:} Remove only the bound computations.
}
}
\description{
The function cleanProfileResults cleans a profile extrema object to partially redo some computations.
}
\examples{
if (!requireNamespace("DiceKriging", quietly = TRUE)) {
stop("DiceKriging needed for this example to work. Please install it.",
     call. = FALSE)
}
# Compute a kriging model from 50 evaluations of the Branin function
# Define the function
g=function(x){
  return(-branin(x))
}
gp_des<-lhs::maximinLHS(20,2)
reals<-apply(gp_des,1,g)
kmModel<-km(design = gp_des,response = reals,covtype = "matern3_2")

threshold=-10

# Compute coordinate profiles on the posterior mean
# Increase multistart and size of designs for more precise results
options_full<-list(multistart=2,heavyReturn=TRUE, Design = replicate(2,seq(0,1,,50)))
init_des<-lhs::maximinLHS(12,2)
options_approx<- list(multistart=2,heavyReturn=TRUE,initDesign=init_des,fullDesignSize=50)
cProfilesMean<-coordinateProfiles(object=kmModel,threshold=threshold,options_full=options_full,
                                  options_approx=options_approx,uq_computations=FALSE,
                                  plot_level=3,plot_options=NULL,CI_const=NULL,return_level=2)

# If we want to run again the computation of approximate coordinate profiles
# we delete that result and run again the coordinate profiles function
cProfiles_full <- cleanProfileResults(cProfilesMean,level=1)
\dontrun{
# Coordinate profiles with UQ with approximate profiles
plot_options<-list(save=FALSE, titleProf = "Coordinate profiles",
                   title2d = "Posterior mean",qq_fill=TRUE)
cProfilesUQ<-coordinateProfiles(object=cProfilesMean,threshold=threshold,options_full=options_full,
                                  options_approx=options_approx,uq_computations=TRUE,
                                  plot_level=3,plot_options=NULL,CI_const=NULL,return_level=2)
# If we would like to remove all UQ results
cProf_noUQ <- cleanProfileResults(cProfilesUQ,level=2)

# If we would like to remove the simulations but keep the pilot points
cProf_noSims <- cleanProfileResults(cProfilesUQ,level=3)
# the line above is useful, for example, if we need a more accurate UQ. In that case
# we obtain more simulations with the same pilot points and then combine the results.

}
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graphicsFuns.R
\name{plotMaxMin}
\alias{plotMaxMin}
\title{Plot coordinate profiles}
\usage{
plotMaxMin(
  allRes,
  Design = NULL,
  threshold = NULL,
  changes = FALSE,
  trueEvals = NULL,
  ...
)
}
\arguments{
\item{allRes}{list containing the list \code{res} which contains the computed minima and maxima. The object returned by the function \code{getAllMaxMin}.}

\item{Design}{a d dimensional design corresponding to the points}

\item{threshold}{if not \code{NULL} plots the level}

\item{changes}{boolean, if not \code{FALSE} plots the points where profile extrema take values near the threshold.}

\item{trueEvals}{if not \code{NULL} adds to each plot the data points and the observed value}

\item{...}{additional parameters to be passed to the plot function}
}
\value{
plots the sup and inf of the function for each dimension. If threshold is not NULL
}
\description{
Plot coordinate profiles, for dimension up to 6.
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profExtremaNumOpt.R
\name{approxProfileExtrema}
\alias{approxProfileExtrema}
\title{Approximate profile extrema functions}
\usage{
approxProfileExtrema(f, fprime = NULL, d, allPsi, opts = NULL)
}
\arguments{
\item{f}{the function to be evaluated}

\item{fprime}{derivative of the function}

\item{d}{dimension of the input domain}

\item{allPsi}{a list containing the matrices Psi (dim \eqn{pxd}) for which to compute the profile extrema}

\item{opts}{a list containing the options for this function and the subfunctions \link{getProfileSup_optim}, \link{getProfileInf_optim} or \link{getProfileExtrema}. The options only for approxProfileExtrema are
\itemize{
\item \code{limits:} an optional list with the upper and lower limits of input space dimension, if NULL then \code{limits=list(upper=rep(1,d),lower=rep(0,d))}
\item \code{smoother:} Select which smoother to use:a string that selects which smoother to use: \itemize{
      \item \code{"1order"}: first order interpolation with gradient
      \item \code{"splineSmooth"}: smoothing spline with default degrees of freedom (DEFAULT OPTION)
      \item \code{"quantSpline"}: profile inf and profile sup approximated with quantile spline regression at levels 0.1 and 0.9 respectively
}
\item \code{heavyReturn:} If TRUE returns also all minimizers, default is FALSE.
\item \code{initDesign:} A list of the same length as allPsi containing the designs of few points where the expensive sup is evaluated. If Null it is automatically initialized
\item \code{fullDesignSize:} The full design where the function is approximated.
\item \code{multistart:} number of multistarts for optim procedure.
\item \code{numMCsamples:} number of MC samples for the sup.
\item \code{plts:} If TRUE, plots the max/min functions at each coordinate, default is FALSE.
\item \code{verb:} If TRUE, outputs intermediate results, default is FALSE.
}}
}
\value{
a list of two data frames (min, max) of the evaluations of \eqn{f_sup(x_i) = sup_{x_j \neq i} f(x_1,\dots,x_d) } and \eqn{f_inf(x_i) = inf_{x_j \neq i} f(x_1,\dots,x_d) }
for each i at the design Design. By default Design is a 100 equally spaced points for each dimension. It can be changed by defining it in options$Design
}
\description{
Evaluate profile extrema for a set of Psi with approximations at few values
}
\examples{
# Compute the oblique profile extrema with approximate optimization on 2d example

# Define the function
testF <- function(x,params,v1=c(1,0),v2=c(0,1)){
return(sin(crossprod(v1,x)*params[1]+params[2])+cos(crossprod(v2,x)*params[3]+params[4])-1.5)
}

testFprime <- function(x,params,v1=c(1,0),v2=c(0,1)){
  return(matrix(c(params[1]*v1[1]*cos(crossprod(v1,x)*params[1]+params[2])-
                  params[3]*v2[1]*sin(crossprod(v2,x)*params[3]+params[4]),
                 params[1]*v1[2]*cos(crossprod(v1,x)*params[1]+params[2])-
                  params[3]*v2[2]*sin(crossprod(v2,x)*params[3]+params[4])),ncol=1))
}


# Define the main directions of the function
theta=pi/6
pparams<-c(1,0,10,0)
vv1<-c(cos(theta),sin(theta))
vv2<-c(cos(theta+pi/2),sin(theta+pi/2))

# Define optimizer friendly function
f <-function(x){
return(testF(x,pparams,vv1,vv2))
}
fprime <- function(x){
 return(testFprime(x,pparams,vv1,vv2))
}

# Define list of directions where to evaluate the profile extrema
all_Psi <- list(Psi1=vv1,Psi2=vv2)

# Evaluate profile extrema along directions of all_Psi
allOblique<-approxProfileExtrema(f=f,fprime = fprime,d = 2,allPsi = all_Psi,
                                 opts = list(plts=FALSE,heavyReturn=TRUE))

\donttest{
# Consider threshold=0
threshold <- 0

# Plot oblique profile extrema functions
plotMaxMin(allOblique,allOblique$Design,threshold = threshold)

## Since the example is two dimensional we can visualize the regions excluded by the profile extrema
# evaluate the function at a grid for plots
inDes<-seq(0,1,,100)
inputs<-expand.grid(inDes,inDes)
outs<-apply(X = inputs,MARGIN = 1,function(x){return(testF(x,pparams,v1=vv1,v2=vv2))})

# obtain the points where the profiles take the threshold value
cccObl<-getChangePoints(threshold = threshold,allRes = allOblique,Design = allOblique$Design)

# visualize the functions and the regions excluded

image(inDes,inDes,matrix(outs,ncol=100),col=grey.colors(20),main="Example and oblique profiles")
contour(inDes,inDes,matrix(outs,ncol=100),add=TRUE,nlevels = 20)
contour(inDes,inDes,matrix(outs,ncol=100),add=TRUE,levels = c(threshold),col=4,lwd=1.5)
plotOblique(cccObl$alwaysEx$`0`[[1]],all_Psi[[1]],col=3)
plotOblique(cccObl$alwaysEx$`0`[[2]],all_Psi[[2]],col=3)
plotOblique(cccObl$neverEx$`0`[[1]],all_Psi[[1]],col=2)
plotOblique(cccObl$neverEx$`0`[[2]],all_Psi[[2]],col=2)

}
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gradKm.R
\name{gradKm_dnewdata}
\alias{gradKm_dnewdata}
\title{Gradient of posterior mean and variance}
\usage{
gradKm_dnewdata(
  object,
  newdata,
  type,
  se.compute = TRUE,
  light.return = FALSE,
  bias.correct = FALSE
)
}
\arguments{
\item{object}{a \link[DiceKriging]{km} object}

\item{newdata}{a vector, matrix or data frame containing the points where to perform predictions.}

\item{type}{a character corresponding to the type of kriging family (\code{"SK"} or \code{"UK"}).}

\item{se.compute}{an optional boolean indicating whether to compute the posterior variance or not. Default is TRUE.}

\item{light.return}{an optional boolean indicating whether to return additional variables. Default is FALSE.}

\item{bias.correct}{an optional boolean to correct bias in the UK variance. Default is FALSE.}
}
\value{
Returns a list containing \itemize{
\item \code{mean:} the gradient of the posterior mean at \code{newdata}.
\item \code{trend:} the gradient of the trend at \code{newdata}.
\item \code{s2:} the gradient of the posterior variance at \code{newdata}.
}
}
\description{
Computes the gradient of the posterior mean and variance of the kriging model in \code{object} at the points \code{newdata}.
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getMaxNumOpt.R
\name{kGradSmooth}
\alias{kGradSmooth}
\title{First order approximation}
\usage{
kGradSmooth(newPoints, profPoints, profEvals, profGradient)
}
\arguments{
\item{newPoints}{vector of points where to approximate the function}

\item{profPoints}{locations where the function was evaluated}

\item{profEvals}{value of the evaluation at profPoints}

\item{profGradient}{value of the gradient at profPoints}
}
\value{
approximated values of the function at newPoints
}
\description{
Compute first order approximation of function from evaluations and gradient
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graphicsFuns.R
\name{plotOblique}
\alias{plotOblique}
\title{plotOblique}
\usage{
plotOblique(changePoints, direction, ...)
}
\arguments{
\item{changePoints}{Numerical vector with the change points (usually if \code{cp=getChangePoints(...)}, then this is cc$alwaysEx[[1]][[1]] for example)}

\item{direction}{The Psi vector used for the direction}

\item{...}{parameters to be passed to abline}
}
\value{
adds to the current plot the lines \eqn{x} s.t. \code{direction}^T \eqn{x} = \code{changePoints[i]} for all i
}
\description{
Auxiliary function for 2d plotting of excluded regions
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getProfiles_optim.R
\name{getProfileInf_optim}
\alias{getProfileInf_optim}
\title{Generic profile inf function computation with optim}
\usage{
getProfileInf_optim(eta, Psi, f, fprime, d, options = NULL)
}
\arguments{
\item{eta}{\eqn{p} dimensional point where the function is to be evaluated}

\item{Psi}{projection matrix of dimension \code{pxd}}

\item{f}{function to be optimized (takes a vector y of dimension d and returns a real number)}

\item{fprime}{derivative of f (same format, returning a \eqn{d} dimensional vector)}

\item{d}{dimension of the input for f}

\item{options}{a list containing the options to be passed to optim:
\itemize{
\item \code{par:} contains the starting point (a point in dimension d)
\item \code{lower:} the lower bounds for the optimization domain (see optim for details)
\item \code{upper:} the upper bounds for the optimization domain (see optim for details)
}}
}
\value{
a real value corresponding to \eqn{min_{x \in D_Psi} f(x)}
}
\description{
Compute profile inf function for an arbitrary matrix \code{Psi} with with the L-BFGS-B algorithm of \link[stats]{optim}. Here the linear equality constraint is eliminated by using the Null space of \code{Psi}.
}
\seealso{
\link{getProfileSup_optim}, \link{plotMaxMin}
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getMaxNumOpt.R
\name{getAllMaxMin}
\alias{getAllMaxMin}
\title{Coordinate profile extrema with BFGS optimization}
\usage{
getAllMaxMin(f, fprime = NULL, d, options = NULL)
}
\arguments{
\item{f}{the function to be evaluated}

\item{fprime}{derivative of the function}

\item{d}{dimension of the input domain}

\item{options}{a list containing the options for this function and the subfunctions getMax, getMin
see documentation of getMax, getMin for details. The options only for getAllMaxMin are
\itemize{
\item \code{Design:} an optional design matrix with the discretization of each dimension, if NULL then for each dimension Design[,coord] = seq(0,1,length.out=100)
\item \code{heavyReturn:} If TRUE returns also all minimizers, default is FALSE.
\item \code{plts:} If TRUE, plots the max/min functions at each coordinate, default is FALSE.
\item \code{verb:} If TRUE, outputs intermediate results, default is FALSE.
\item \code{MonteCarlo:} If TRUE, use the MC optimizer otherwise use standard optim.
}}
}
\value{
a list of two data frames (min, max) of the evaluations of \eqn{f_sup(x_i) = sup_{x_j \neq i} f(x_1,\dots,x_d) } and \eqn{f_inf(x_i) = inf_{x_j \neq i} f(x_1,\dots,x_d) }
for each i at the design Design. By default Design is a 100 equally spaced points for each dimension. It can be changed by defining it in options$Design
}
\description{
Evaluate coordinate profile extrema with full optimization.
}
\examples{
if (!requireNamespace("DiceKriging", quietly = TRUE)) {
stop("DiceKriging needed for this example to work. Please install it.",
     call. = FALSE)
}
# Compute the coordinate profile extrema with full optimization on 2d example

# Define the function
g=function(x){
  return(-branin(x))
}
# Define the gradient
gprime = function(x){
  x1 = x[1]*15-5
  x2 = x[2]*15
  f1prime = (15*25)/(4*pi^4)*x1^3 - (15*75)/(2*pi^3)*x1^2 +
  (80*15)/(pi^2)*x1 - (5*15)/(pi^2)*x2*x1 +
  10*15/pi*x2 - 60*15/pi-10*15* (1 - 1/(8*pi))*sin(x1)
  f2prime = 2*15*(x2-5/(4*pi^2)*x1^2 +5/pi*x1-6)
  return(c(-f1prime,-f2prime))
}
# set up dimension
coordProf<-getAllMaxMin(f = g,fprime = gprime,d=2,options = list(multistart=4,heavyReturn=TRUE))

\donttest{
# Consider threshold=-10
threshold<- -10
# obtain the points where the profiles take the threshold value
pp_change<-getChangePoints(threshold = threshold,allRes = coordProf)
# evaluate g at a grid and plot the image
x<-seq(0,1,,100)
grid<-expand.grid(x,x)
g_evals<- apply(X = grid,MARGIN = 1,FUN = g)
image(x = x,y = x,z = matrix(g_evals,nrow = 100),col = grey.colors(20))
contour(x=x,y=x,z=matrix(g_evals,nrow = 100), add=TRUE, nlevels = 20)
contour(x=x,y=x,z=matrix(g_evals,nrow = 100), add=TRUE, levels = threshold,col=2)
abline(h = pp_change$neverEx$`-10`[[2]],col="darkgreen",lwd=2)
abline(v = pp_change$neverEx$`-10`[[1]],col="darkgreen",lwd=2)
# Plot the coordinate profiles and a threshold
plotMaxMin(allRes = coordProf,threshold = threshold,changes = TRUE)
}
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graphicsFuns.R
\name{getChangePoints}
\alias{getChangePoints}
\title{Coordinate profiles crossing points}
\usage{
getChangePoints(threshold, Design = NULL, allRes)
}
\arguments{
\item{threshold}{if not null plots the level}

\item{Design}{a d dimensional design corresponding to the points}

\item{allRes}{list containing the list \code{res} which contains the computed minima and maxima. The object returned by the function \code{getAllMaxMin}.}
}
\value{
returns a list containing two lists with d components where
\itemize{
\item alwaysEx: each component is a numerical vector indicating the points \eqn{x_i} where \eqn{inf_{x^{-i}}f(x) >} \code{threshold};
\item neverEx: each component is a numerical vector indicating the points \eqn{x_i} where \eqn{sup_{x^{-i}}f(x) <} \code{threshold}.
}
}
\description{
Obtain the points where the coordinate profile extrema functions cross the threshold
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bound_functions.R
\name{mean_Delta_T}
\alias{mean_Delta_T}
\title{mean function of difference process}
\usage{
mean_Delta_T(x, kmModel, simupoints, T.mat, F.mat)
}
\arguments{
\item{x}{a matrix \eqn{r x d} containing the \eqn{r} points where the function is to be computed.}

\item{kmModel}{the \link[DiceKriging]{km} model of the Gaussian process \eqn{Z}.}

\item{simupoints}{the matrix \eqn{l x d} containing the pilot points \eqn{G}.}

\item{T.mat}{the upper triangular factor of the Choleski decomposition of the covariance matrix of \code{rbind(kmModel@X,simupoints)}}

\item{F.mat}{the evaluation of the trend function at \code{rbind(kmModel@X,simupoints)}, see \link[stats]{model.matrix}.}
}
\value{
the value of the mean function at \code{x} for the difference process \eqn{Z^\Delta = Z_x - \widetilde{Z}_x}.
}
\description{
The function mean_Delta_T computes the mean function of the difference process \eqn{Z_x - \widetilde{Z}_x} at \code{x}.
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bound_functions.R
\name{prof_mean_var_Delta}
\alias{prof_mean_var_Delta}
\title{Profile extrema for the mean and variance functions of difference process}
\usage{
prof_mean_var_Delta(
  kmModel,
  simupoints,
  allPsi = NULL,
  options_full_sims = NULL,
  options_approx = NULL,
  F.mat = NULL,
  T.mat = NULL
)
}
\arguments{
\item{kmModel}{the \link[DiceKriging]{km} model of the Gaussian process \eqn{Z}.}

\item{simupoints}{the matrix \eqn{l x d} containing the pilot points \eqn{G}.}

\item{allPsi}{optional list of matrices (dim \eqn{pxd}) for which to compute the profile extrema. If NULL coordinate profiles are computed.}

\item{options_full_sims}{an optional list of options for \link{getAllMaxMin}(or \link{approxProfileExtrema} if \code{allPsi} not NULL). If NULL the full computations are not excuted. NOTE: this computations might be very expensive!}

\item{options_approx}{an optional list of options for \link{approxMaxMin} (or \link{approxProfileExtrema} if \code{allPsi} not NULL).}

\item{F.mat}{the evaluation of the trend function at \code{rbind(kmModel@X,simupoints)}, see \link[stats]{model.matrix}, if \code{NULL} it is computed.}

\item{T.mat}{the upper triangular factor of the Choleski decomposition of the covariance matrix of \code{rbind(kmModel@X,simupoints)}, if \code{NULL} it is computed.}
}
\value{
the profile extrema functions at \code{options_approx$design} for the mean and variance function of the difference process \eqn{Z^\Delta = Z_x - \widetilde{Z}_x}.
}
\description{
The function prof_mean_var_Delta computes the profile extrema functions for the mean and variance functions of the difference process \eqn{Z_x - \widetilde{Z}_x} at \code{x}.
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coordProf_UQ.R
\name{coordProf_UQ}
\alias{coordProf_UQ}
\title{Coordinate profiles UQ from a kriging model}
\usage{
coordProf_UQ(
  object,
  threshold,
  allResMean = NULL,
  quantiles_uq = c(0.05, 0.95),
  options_approx = NULL,
  options_full_sims = NULL,
  options_sims = NULL,
  options_bound = NULL,
  plot_level = 0,
  plot_options = NULL,
  return_level = 1
)
}
\arguments{
\item{object}{either a \link[DiceKriging]{km} model or a list containing partial results. If \code{object} is a km model then all computations are carried out. If \code{object} is a list, then the function carries out all computations to complete the results list.}

\item{threshold}{the threshold of interest}

\item{allResMean}{a list resulting from \code{getAllMaxMin} or \code{approxMaxMin} for the profile extrema on the mean. If NULL the median from the observations is plotted}

\item{quantiles_uq}{a vector containing the quantiles to be computed}

\item{options_approx}{an optional list of options for approxMaxMin, see \link{approxMaxMin} for details.}

\item{options_full_sims}{an optional list of options for getAllMaxMin, see \link{getAllMaxMin} for details. If NULL the full computations are not excuted. NOTE: this computations might be very expensive!}

\item{options_sims}{an optional list of options for the posterior simulations.
\itemize{
\item \code{algorithm:} string choice of the algorithm to select the pilot points ("A" or "B", default "B");
\item \code{lower:} \eqn{d} dimensional vector with lower bounds for pilot points, default \code{rep(0,d)};
\item \code{upper:} \eqn{d} dimensional vector with upper bounds for pilot points, default \code{rep(1,d)};
\item \code{batchsize:} number of pilot points, default \code{120};
\item \code{optimcontrol:} list containing the options for optimization, see \link[pGPx]{optim_dist_measure};
\item \code{integcontrol:} list containing the options for numerical integration of the criterion, see \link[pGPx]{optim_dist_measure};
\item \code{integration.param:} list containing the integration design, obtained with the function \link[KrigInv]{integration_design};
\item \code{nsim:} number of approximate GP simulations, default \code{300}.
}}

\item{options_bound}{an optional list containing \code{beta} the confidence level for the approximation and \code{alpha} the confidence level for the bound. Note that \code{alpha > 2*beta}. If \code{NULL}, the bound is not computed.}

\item{plot_level}{an integer to select the plots to return (0=no plots, 1=basic plots, 2= all plots)}

\item{plot_options}{an optional list of parameters for plots. See \link{setPlotOptions} for currently available options.}

\item{return_level}{an integer to select the amount of details returned}
}
\value{
If return_level=1 a list containing \itemize{
\item \code{profSups:} an array \code{dxfullDesignSizexnsims} containing the profile sup for each coordinate for each realization.
\item \code{profInfs:} an array \code{dxfullDesignSizexnsims} containing the profile inf for each coordinate for each realization.
\item \code{prof_quantiles_approx:} a list containing the quantiles (levels set by \code{quantiles_uq}) of the profile extrema functions.
} if return_level=2 the same list as above but also including \code{more:} a list containing \itemize{
\item \code{times:} a list containing
	\itemize{
	\item \code{tSpts:} computational time for selecting pilot points.
	\item \code{tApprox1ord:} vector containing the computational time required for profile extrema computation for each realization
	}
\item \code{simuls:} a matrix containing the value of the field simulated at the pilot points
\item \code{sPts:} the pilot points
}
}
\description{
The function coordProf_UQ computes the profile extrema functions for posterior realizations of a Gaussian process and its confidence bounds
}
\examples{
if (!requireNamespace("DiceKriging", quietly = TRUE)) {
stop("DiceKriging needed for this example to work. Please install it.",
     call. = FALSE)
}
# Compute a kriging model from 50 evaluations of the Branin function
# Define the function
g<-function(x){
  return(-branin(x))
}
gp_des<-lhs::maximinLHS(20,2)
reals<-apply(gp_des,1,g)
kmModel<-km(design = gp_des,response = reals,covtype = "matern3_2")

threshold=-10
d<-2

# Compute coordinate profiles UQ starting from GP model
# define simulation options
options_sims<-list(algorithm="B", lower=rep(0,d), upper=rep(1,d),
                   batchsize=80, optimcontrol = list(method="genoud",pop.size=100,print.level=0),
                   integcontrol = list(distrib="sobol",n.points=1000), nsim=150)
# define 1 order approximation options
init_des<-lhs::maximinLHS(15,d)
options_approx<- list(multistart=4,heavyReturn=TRUE,
                      initDesign=init_des,fullDesignSize=100,
                      smoother="1order")
# define plot options
options_plots<-list(save=FALSE, titleProf = "Coordinate profiles",
                    title2d = "Posterior mean",qq_fill=TRUE)
\dontrun{
# profile UQ on approximate coordinate profiles
cProfiles_UQ<-coordProf_UQ(object = kmModel,threshold = threshold,allResMean = NULL,
                            quantiles_uq = c(0.05,0.95),options_approx = options_approx,
                            options_full_sims = NULL,options_sims = options_sims,
                            options_bound = NULL,plot_level = 3,
                            plot_options = options_plots,return_level = 3)
# profile UQ on full optim coordinate profiles
options_full_sims<-list(multistart=4,heavyReturn=TRUE)
cProfiles_UQ_full<-coordProf_UQ(object = cProfiles_UQ,threshold = threshold,allResMean = NULL,
                            quantiles_uq = c(0.05,0.95),options_approx = options_approx,
                            options_full_sims = options_full_sims,options_sims = options_sims,
                            options_bound = NULL,plot_level = 3,
                            plot_options = options_plots,return_level = 3)

# profile UQ on full optim coordinate profiles with bound
cProfiles_UQ_full_bound<-coordProf_UQ(object = cProfiles_UQ_full,threshold = threshold,
                                      allResMean = NULL, quantiles_uq = c(0.05,0.95),
                                      options_approx = options_approx,
                                      options_full_sims = options_full_sims,
                                      options_sims = options_sims,
                                      options_bound = list(beta=0.024,alpha=0.05),
                                      plot_level = 3, plot_options = options_plots,
                                      return_level = 3)
}
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getPointProportion.R
\name{getPointProportion}
\alias{getPointProportion}
\title{Obtain proportion of true observations in excursion set}
\usage{
getPointProportion(pp, xBins, yBins, whichAbove, plt = FALSE)
}
\arguments{
\item{pp}{a matrix of dimension nPts x 2 with the true points locations in 2 dimensions.}

\item{xBins}{numerical vector with the ordered breaks of the grid along the x axis}

\item{yBins}{numerical vector with the ordered breaks of the grid along the y axis}

\item{whichAbove}{boolean vector of dimension nPts, selects the points above}

\item{plt}{if not \code{TRUE} plots the grid, the points and the counts for each cell.}
}
\value{
a list containing \code{above}, the counts of points in excursion, \code{full} the counts per cell of all points,
 \code{freq}, the relative frequence.
}
\description{
Computes the proportion of observations in the excursion set from true function evaluations,
 binned by the grid determined with \code{xBins}, \code{yBins}.
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getMaxNumOpt.R
\name{getClosePoints}
\alias{getClosePoints}
\title{Find close points}
\usage{
getClosePoints(x, allPts, whichDim)
}
\arguments{
\item{x}{one dimensional point}

\item{allPts}{dataframe containing a list of d dimensional points}

\item{whichDim}{integer defining the dimension of x}
}
\value{
the index in allPts (row number) of the closest point in allPts to x along the whichDim dimension
}
\description{
Obtain points close in one specific dimension
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coordinateProfiles.R
\name{coordinateProfiles}
\alias{coordinateProfiles}
\title{Coordinate profiles starting from a kriging model}
\usage{
coordinateProfiles(
  object,
  threshold,
  options_full = NULL,
  options_approx = NULL,
  uq_computations = FALSE,
  plot_level = 0,
  plot_options = NULL,
  CI_const = NULL,
  return_level = 1,
  ...
)
}
\arguments{
\item{object}{either a \link[DiceKriging]{km} model or a list containing partial results. If \code{object} is a km model then all computations are carried out. If \code{object} is a list, then the function carries out all computations to complete the list results.}

\item{threshold}{the threshold of interest}

\item{options_full}{an optional list of options for getAllMaxMin, see \link{getAllMaxMin} for details.}

\item{options_approx}{an optional list of options for approxMaxMin, see \link{approxMaxMin} for details.}

\item{uq_computations}{boolean, if TRUE the uq computations for the profile mean are computed.}

\item{plot_level}{an integer to select the plots to return (0=no plots, 1=basic plots, 2= all plots)}

\item{plot_options}{an optional list of parameters for plots. See \link{setPlotOptions} for currently available options.}

\item{CI_const}{an optional vector containing the constants for the CI. If not NULL, then profiles extrema for \eqn{m_n(x) \pm CI_const[i]*s_n(x,x)} are computed.}

\item{return_level}{an integer to select the amount of details returned}

\item{...}{additional parameters to be passed to \link{coordProf_UQ}.}
}
\value{
If return_level=1 a list containing
\itemize{
\item \code{profMean_full:} the results of \code{getAllMaxMin} for the posterior mean
\item \code{profMean_approx:} the results of \code{approxMaxMin} for the posterior mean
\item \code{res_UQ:} the results of \code{coordProf_UQ} for the posterior mean
}
if return_level=2 the same list as above but also including
\itemize{
\item \code{abs_err:} the vector of maximum absolute approximation errors for the profile inf /sup on posterior mean for the chosen approximation
\item \code{times:}  a list containing
	\itemize{
	\item \code{full:} computational time for the full computation of profile extrema
	\item \code{approx:} computational time for the approximate computation of profile extrema
	}
}
}
\description{
The function coordinateProfiles computes the profile extrema functions for the posterior mean of a Gaussian process and its confidence bounds
}
\examples{
if (!requireNamespace("DiceKriging", quietly = TRUE)) {
stop("DiceKriging needed for this example to work. Please install it.",
     call. = FALSE)
}
# Compute a kriging model from 50 evaluations of the Branin function
# Define the function
g=function(x){
  return(-branin(x))
}
gp_des<-lhs::maximinLHS(20,2)
reals<-apply(gp_des,1,g)
kmModel<-km(design = gp_des,response = reals,covtype = "matern3_2")

threshold=-10

# Compute coordinate profiles on the posterior mean
# Increase multistart and size of designs for more precise results
options_full<-list(multistart=2,heavyReturn=TRUE, Design = replicate(2,seq(0,1,,50)))
init_des<-lhs::maximinLHS(12,2)
options_approx<- list(multistart=2,heavyReturn=TRUE,initDesign=init_des,fullDesignSize=50)
cProfilesMean<-coordinateProfiles(object=kmModel,threshold=threshold,options_full=options_full,
                                  options_approx=options_approx,uq_computations=FALSE,
                                  plot_level=3,plot_options=NULL,CI_const=NULL,return_level=2)
\dontrun{
# Coordinate profiles with UQ with approximate profiles
plot_options<-list(save=FALSE, titleProf = "Coordinate profiles",
                   title2d = "Posterior mean",qq_fill=TRUE)
cProfilesUQ<-coordinateProfiles(object=cProfilesMean,threshold=threshold,options_full=options_full,
                                  options_approx=options_approx,uq_computations=TRUE,
                                  plot_level=3,plot_options=NULL,CI_const=NULL,return_level=2)

# Coordinate profiles with UQ with fully optim profiles
options_full_sims<-list(multistart=4,heavyReturn=TRUE, Design = replicate(2,seq(0,1,,60)))
cProfilesUQ<-coordinateProfiles(object=cProfilesMean,threshold=threshold,options_full=options_full,
                                  options_approx=options_approx,uq_computations=TRUE,
                                  plot_level=3,plot_options=NULL,CI_const=NULL,return_level=2,
                                  options_full_sims=options_full_sims)
}
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graphicsFuns.R
\name{getSegments}
\alias{getSegments}
\title{getSegments}
\usage{
getSegments(y)
}
\arguments{
\item{y}{a vector}
}
\value{
plots the sup and inf of the function for each dimension. If threshold is not NULL
}
\description{
Auxiliary function for \code{getChangePoints}
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/obliqueProfiles.R
\name{obliqueProfiles}
\alias{obliqueProfiles}
\title{Oblique coordinate profiles starting from a kriging model}
\usage{
obliqueProfiles(
  object,
  allPsi,
  threshold,
  options_full = NULL,
  options_approx = NULL,
  uq_computations = FALSE,
  plot_level = 0,
  plot_options = NULL,
  CI_const = NULL,
  return_level = 1,
  ...
)
}
\arguments{
\item{object}{either a \link[DiceKriging]{km} model or a list containing partial results. If \code{object} is a km model then all computations are carried out. If \code{object} is a list, then the function carries out all computations to complete the list results.}

\item{allPsi}{a list containing the matrices Psi (dim \eqn{pxd}) for which to compute the profile extrema}

\item{threshold}{the threshold of interest}

\item{options_full}{an optional list of options for getProfileExtrema, see \link{getProfileExtrema} for details.}

\item{options_approx}{an optional list of options for approxProfileExtrema, see \link{approxProfileExtrema} for details.}

\item{uq_computations}{boolean, if TRUE the uq computations for the profile mean are computed.}

\item{plot_level}{an integer to select the plots to return (0=no plots, 1=basic plots, 2= all plots)}

\item{plot_options}{an optional list of parameters for plots. See \link{setPlotOptions} for currently available options.}

\item{CI_const}{an optional vector containing the constants for the CI. If not NULL, then profiles extrema for \eqn{m_n(x) \pm CI_const[i]*s_n(x,x)} are computed.}

\item{return_level}{an integer to select the amount of details returned}

\item{...}{additional parameters to be passed to \link{obliqueProf_UQ}.}
}
\value{
If return_level=1 a list containing
\itemize{
\item \code{profMean_full:} the results of \code{getProfileExtrema} for the posterior mean
\item \code{profMean_approx:} the results of \code{approxProfileExtrema} for the posterior mean
\item \code{res_UQ:} the results of \code{obliqueProf_UQ} for the posterior mean
}
if return_level=2 the same list as above but also including
\itemize{
\item \code{abs_err:} the vector of maximum absolute approximation errors for the profile inf /sup on posterior mean for the chosen approximation
\item \code{times:}  a list containing
	\itemize{
	\item \code{full:} computational time for the full computation of profile extrema
	\item \code{approx:} computational time for the approximate computation of profile extrema
	}
}
}
\description{
The function obliqueProfiles computes the (oblique) profile extrema functions for the posterior mean of a Gaussian process and its confidence bounds
}
\examples{
if (!requireNamespace("DiceKriging", quietly = TRUE)) {
stop("DiceKriging needed for this example to work. Please install it.",
     call. = FALSE)
}
# Compute a kriging model from 50 evaluations of the Branin function
# Define the function
g=function(x){
  return(-branin(x))
}
gp_des<-lhs::maximinLHS(20,2)
reals<-apply(gp_des,1,g)
kmModel<-km(design = gp_des,response = reals,covtype = "matern3_2")

threshold=-10

# Compute oblique profiles on the posterior mean
# (for theta=0 it is equal to coordinateProfiles)
options_full<-list(multistart=4,heavyReturn=TRUE,discretization=100)
options_approx<- list(multistart=4,heavyReturn=TRUE,initDesign=NULL,fullDesignSize=100)
theta=pi/4
allPsi = list(Psi1=matrix(c(cos(theta),sin(theta)),ncol=2),
Psi2=matrix(c(cos(theta+pi/2),sin(theta+pi/2)),ncol=2))

\dontrun{
profMeans<-obliqueProfiles(object = kmModel,allPsi = allPsi,threshold = threshold,
                           options_full = options_full,options_approx = options_approx,
                           uq_computations = FALSE,plot_level = 3,plot_options = NULL,
                           CI_const = NULL,return_level = 2)


# Approximate oblique profiles with UQ
plot_options<-list(save=FALSE, titleProf = "Coordinate profiles",
                   title2d = "Posterior mean",qq_fill=TRUE)
options_sims<-list(nsim=150)
obProfUQ<-obliqueProfiles(object=profMeans,threshold=threshold,allPsi = allPsi,
                           options_full=options_full, options_approx=options_approx,
                           uq_computations=TRUE, plot_level=3,plot_options=NULL,
                           CI_const=NULL,return_level=2,options_sims=options_sims)
}
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotBivariateProfiles.R
\name{plotBivariateProfiles}
\alias{plotBivariateProfiles}
\title{Plot bivariate profiles}
\usage{
plotBivariateProfiles(
  bivProf,
  allPsi,
  Design = NULL,
  threshold = NULL,
  whichIQR = NULL,
  plot_options = NULL,
  ...
)
}
\arguments{
\item{bivProf}{list returned by \code{obliqueProfiles}.}

\item{allPsi}{a list containing the matrices Psi (dim \eqn{2xd}) for which to compute the profile extrema}

\item{Design}{a matrix of dimension \eqn{(2d)x numPsi} encoding the first (\code{Design[1:d,]}) and the second ((\code{Design[(d+1):(2*d),]})) axis values.}

\item{threshold}{if not \code{NULL} plots the level as a contour.}

\item{whichIQR}{which quantiles to use for the inter-quantile range plot. If \code{NULL}, automatically selects the first and the last element of \code{bivProfres_UQ$prof_quantiles_approx}}

\item{plot_options}{list as returned by \code{setPlotOptions}.}

\item{...}{additional parameters to be passed to the plot function}
}
\value{
plots the 2d maps of the profile sup and inf of the function for each Psi in allPsi. If threshold is not NULL also contours the threshold level.
}
\description{
Plot bivariate profiles, for dimension up to 6.
}
\author{
Dario Azzimonti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{coastal_flooding}
\alias{coastal_flooding}
\title{Coastal flooding as function of offshore forcing conditions}
\format{
A data frame with 200 rows and 6 variables:
\describe{
  \item{Tide}{High tide level in meters;}
  \item{Surge}{Surge peak amplitude in meters;}
  \item{phi}{Phase difference between high tide and surge peak;}
  \item{t-}{Duration of the increasing part of the surge temporal signal (assumed to be triangular);}
  \item{t+}{Duration of the decreasing part of the surge temporal signal (assumed to be triangular);}
  \item{Area}{Flooded area in m^2.}
}
}
\usage{
coastal_flooding
}
\description{
A dataset containing the results of a numerical simulation conducted with
the MARS model (Lazure and Dumas, 2008) for coastal flooding.
The numerical model was adapted to the Boucholeurs area (French Atlantic coast),
close to La Rochelle, and validated with data from the 2010 Xynthia storm event.
See Azzimonti et al. (2017+) and Rohmer et al. (2018) for more details.
}
\details{
The data frame contains 5 input variables: \code{Tide}, \code{Surge}, \code{phi}, \code{t-}, \code{t+} detailing
the offshore forcing conditions for the model. All input variables are normalized
in \eqn{[0,1]}. The response is \code{Area}, the area flooded in m^2.
}
\examples{
# Define inputs
inputs<-data.frame(coastal_flooding[,-6])
colnames(inputs)<-colnames(coastal_flooding[,-6])
colnames(inputs)[4:5]<-c("tPlus","tMinus")

# put response in areaFlooded variable
areaFlooded<-data.frame(coastal_flooding[,6])
colnames(areaFlooded)<-colnames(coastal_flooding)[6]
response = sqrt(areaFlooded)

\donttest{
model <- km(formula=~Tide+Surge+I(phi^2)+tMinus+tPlus,
            design = inputs,response = response,covtype="matern3_2")
# Fix threshold
threshold<-sqrt(c(1.2e6,1.9e6,3.1e6,6.5e6))

# use the coordinateProfile function
## set up plot options
options_plots <- list(save=FALSE, folderPlots = "./" ,
                      titleProf = "Coordinate profiles",
                      title2d = "Posterior mean",qq_fill=TRUE)
# set up full profiles options
options_full<-list(multistart=15,heavyReturn=TRUE)
# set up approximation options
d <- model@d
init_des<-lhs::maximinLHS(5*d , d )
options_approx<- list(multistart=2,heavyReturn=TRUE, initDesign=init_des,
                      fullDesignSize=100, smoother="quantSpline")

# run the coordinate profile extrema on the mean
CF_CoordProf_mean<- coordinateProfiles(object = model, threshold = threshold,
                                       uq_computations = FALSE, options_approx = options_approx,
                                       plot_level=3, plot_options= options_plots, return_level=3,
                                       options_full=options_full)
}
\dontrun{
## UQ computations might require a long time
# set up simulation options
## reduce nsims and batchsize for faster/less accurate UQ
nsims=200
opts_sims<-list(algorithm="B", lower=rep(0,d ),
                upper=rep(1,d ), batchsize=150,
                optimcontrol=list(method="genoud", pop.size=100,print.level=0),
                integcontrol = list(distrib="sobol",n.points=2000),nsim=nsims)

opts_sims$integration.param <- integration_design(opts_sims$integcontrol,
                                                  d , opts_sims$lower,
                                                  opts_sims$upper,
                                                  model,threshold)
opts_sims$integration.param$alpha <- 0.5

# run UQ computations
CF_CoordProf_UQ<- coordinateProfiles(object = CF_CoordProf_mean, threshold = threshold,
                                     uq_computations = TRUE, options_approx = options_approx,
                                     plot_level=3, plot_options= options_plots, return_level=3,
                                     options_sims=opts_sims,options_full=options_full,
                                     options_bound = list(beta=0.024,alpha=0.05))
}

}
\references{
Azzimonti, D., Ginsbourger, D., Rohmer, J. and Idier, D. (2017+). \emph{Profile extrema for visualizing and quantifying uncertainties on excursion regions. Application to coastal flooding.} arXiv:1710.00688.

Rohmer, J., Idier, D., Paris, F., Pedreros, R., and Louisor, J. (2018). \emph{Casting light on forcing and breaching scenarios that lead to marine inundation: Combining numerical simulations with a random-forest classification approach.} Environmental Modelling & Software, 104:64-80.

Lazure, P. and Dumas, F. (2008). \emph{An external-internal mode coupling for a 3D hydrodynamical model for applications at regional scale (MARS)}. Advances in Water Resources, 31:233-250.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graphicsFuns.R
\name{plotOneBivProfile}
\alias{plotOneBivProfile}
\title{Plot bivariate profiles}
\usage{
plotOneBivProfile(
  allRes,
  allPsi,
  Design = NULL,
  threshold = NULL,
  trueEvals = NULL,
  main_addendum = "",
  ...
)
}
\arguments{
\item{allRes}{list containing the list \code{res} which contains the computed minima and maxima. The object returned by the function \code{getProfileExtrema}.}

\item{allPsi}{a list containing the matrices Psi (dim \eqn{2xd}) for which to compute the profile extrema}

\item{Design}{a matrix of dimension \eqn{(2d)x numPsi} encoding the first (\code{Design[1:d,]}) and the second ((\code{Design[(d+1):(2*d),]})) axis values.}

\item{threshold}{if not \code{NULL} plots the level as a contour.}

\item{trueEvals}{if not \code{NULL} adds to each plot the data points and the observed value}

\item{main_addendum}{additional string to add to image title. Default is empty string.}

\item{...}{additional parameters to be passed to the plot function}
}
\value{
plots the 2d maps of the profile sup and inf in \code{allRes} for each Psi in \code{allPsi}. If threshold is not NULL also contours the threshold level.
}
\description{
Plots the bivariate profiles stored in \code{allRes} for each Psi in \code{allPsi}.
}
\seealso{
plotBivariateProfiles
}
\author{
Dario Azzimonti
}
